--- Guest data controller for processing Guest and Key actions
-- 
-- @file	UserController.lua 
-- @author	Vyacheslav Khudan
-- @created 2015/07/02

--- @module UserController
local UserController = {}

--- Load additional modules
local wmsldap = require"core.wmsldap"

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
--- Make validation and conversion of the user attributes
--TODO: extend attributes list
local function __processUserAttrs(name, value, mode)
	if not name or "" == name then
		return false, "Invalid user attribute name"
	end
	mode = "w" == mode and "wrtite" or "read"
	
	-- processing attribute names
	local attrName = ({
		["read"] = {
    		 name	 = "cn"
    		,lang    = "preferredLanguage" 
    		,exten   = "telephoneNumber"
    		,uid	 = "uid"
    	},
    	["write"] = {
    		 name	 = "cn"
    		,lang    = "preferredLanguage"
    	}
	})[mode][name]
	
	-- processing attribute values
	local attrVal = value
	if name and value then
		if "lang" == name then
			attrVal = ({
				 en = "en"
        		,ea = "en"
        		,it = "it"
        		,fr = "fr"
        		,de = "de"
        		,ge = "de"
        		,bg = "bg"
        		,ru = "ru"
        		,es = "es"
        		,sp = "es"
			})[value:lower()] or "en"
		elseif "name" == name then
			attrVal = value:gsub("[%^%/\\%@%|%+%=%}%{%]%[%)%(%*%&%?%%%$%#%@%!%,\"%:%;%<%>]", " "):sub(1, 50)
		elseif "uid" == name or "exten" == name then
			attrVal = tonumber(value)
		end
	end
	
	return attrName, attrVal
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Retreive user information from LDAP
-- @param attr	A search attribute name
-- @param val	A value of the attribute for searching
-- @return list of the found user attribures and user distinguished name
function UserController:getAttr(attr, val)
	if not attr or not val then
		return false, "Invalid arguments"
	end
	
	local attrName = __processUserAttrs(attr)
	if not attrName then
		return false, string.format("Invalid user attribute: %s", attr)
	end
	
	local ldapData, err = wmsldap:get(attrName, val)
	if err then
		return false, err
	end
	
	local userData = ldapData[1]
	if "table" ~= type(userData) then
		return false, string.format("Not found user data by attribute (%s = %s)", attr, val)
	end
	
	local userDN   = userData[1]
	local userAttr = userData[2]
	
	return userAttr, userDN
end

--- Retreive user information by UID value
function UserController:getAttrByUID(val)
	return self:getAttr("uid", val)
end

--- Retreive user information by User Exten value
function UserController:getAttrByExten(val)
	return self:getAttr("exten", val)
end

--- Retreive user information by User Name value
function UserController:getAttrByName(val)
	return self:getAttr("name", val)
end

--- Update user attributes by new values
-- @param attr		An attribute name used for user entry detection
-- @param val		A value of the described above atrribute name
-- @param newAttrs	A list of the new attributes in format [name]=[value]
-- @return boolean TRUE value otherwise FALSE followed by error message
function UserController:setAttr(attr, val, newAttrs)
	if not attr or not val then
		return false, "Invalid search user values"
	end
	if "table" ~= type(newAttrs) then
		return false, "Invalid new attributes format"
	end
	
	local rv, err = self:getAttr(attr, val)
	if not rv then
		return false, err
	end
	local userDn = err
	
	local rs  = nil
	local err = nil
	for name, val in pairs(newAttrs) do
		if name and val then
    		local attrName, attrVal = __processUserAttrs(name, val)
    		if not attrName then
    			err = string.format("Invalid attribute name %q", name or "")
    			break
    		end
    		if not attrVal then
    			err = string.format("Invalid attribute value %q for %q", val or "", name or "")
    			break
    		end
    		rs, err = wmsldap:modify(userDn, attrName, attrVal, "set")
    		if not rs then break end
    	end
	end
	
	return rs, err
end

--- Update user attributes for user with specified UID
function UserController:setAttrByUID(val, newAttrs)
	return self:setAttr("uid", val, newAttrs)
end

--- Update user attributes for user with specified Exten
function UserController:setAttrByExten(val, newAttrs)
	return self:setAttr("exten", val, newAttrs)
end

--- Update user attributes for user with specified Name
function UserController:setAttrByName(val, newAttrs)
	return self:setAttr("name", val, newAttrs)
end

return UserController