--- Processing call class by number
--
-- @file    callClassImpl.lua
-- @author  Vyacheslav Khudan
-- @created 2015/12/03

package.path = package.path .. ";/usr/local/share/lua/5.1/common/?.lua"

--- @module callClassImpl
local callClassImpl = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
local classList            = require"callClass.callClassDefines".classList
local regionCallClasses    = require"callClass.callClassDefines".continentsCallClasses
local countriesCallClasses = require"callClass.callClassDefines".countriesCallClasses
local number2Class         = require"callClass.callClassDefines".numberType2CallClassMap

--- Retrieving data about call class
-- @param sign  A sting with sign search
-- @return table with class ident and name
local function __getCallClassData(sign, value)
    local rv = {}
    if not sign or not value or "" == value then
        return rv
    end
    
    if "numberingPlans" == sign then                
        local number      = value
        for countryCode, countryClasses in pairs(countriesCallClasses) do   -- searching in the static regexp
            if countryCode and number:match("^"..countryCode) then
                for obj, regexpList in pairs(countryClasses) do
                    for _, regexp in ipairs(regexpList) do
                        if number:match(regexp) then
                            rv.name  = obj.n
                            rv.ident = obj.p
                            break
                        end
                    end
                    if rv.ident then break end
                end
                break
            end
        end 
        if not rv.ident then    -- search by number via numbering plans    
            local numberObj = require"numberDetect.numberDetectImpl":new()
            local searchedData, errMsg = numberObj:search(number)
            if searchedData then
                local numberType = numberObj:getTypeIdent()
                local classData  = number2Class[numberType]
                if classData then
                    rv.name  = classData.n
                    rv.ident = classData.p
                end
            else
                print("Number detect is not performed:", errMsg)
                logger:error("Number detect is not performed:", errMsg)
            end
        end
    elseif "staticRegexp" == sign then             -- search by number via static continent table
        local number    = value
        for obj, regexpList in pairs(regionCallClasses) do
            for _, regexp in ipairs(regexpList) do
                if number:match(regexp) then
                    rv.name  = obj.n
                    rv.ident = obj.p
                    break
                end
            end
            if rv.ident then break end    
        end
    end
    
    return rv
end

--- Building list of the call classes
-- @param option    An option that set up result data (full, only id -> name, only name -> id)
-- @return list with call classes data
local function __buildCallClassMap(option)
    local rv     = {}
    local option = option or "fullData" 
    
    for _, data in pairs(classList) do
        if "table" == type(data) and data.p and data.n then
            if "fullData" == option or "onlyIndex" == option then
                rv[data.p] = data.n
            end
            if "fullData" == option or "onlyName" == option then
                rv[data.n] = data.p
            end 
        end
    end
    
    return rv
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------

--- Assembling classes names to the list
-- @return table with classes names
-- @note used by WMS interface
function callClassImpl:getNamesList()
    local rv = __buildCallClassMap("onlyIndex")
    
    rv[0] = nil -- remove unknown item
    
    return rv
end

--- Retrieving call class ident by name
-- @param val   A call class name
-- @return ident value otherwise nil
function callClassImpl:getIdByName(val)
    if not val or "" == val then
        return nil
    end
    local map = __buildCallClassMap("onlyName")
    return map[val]
end

--- Retrieving call class name by ident
-- @param val   A call class ident
-- @return name value otherwise nil
function callClassImpl:getNameById(val)
    if not val then
        return nil
    end
    local map = __buildCallClassMap("onlyIndex")
    return map[val]
end

--- Processing class identificator
-- @param number           A normilized number  (+XXXX...)
-- @param callCountryCode  A coutnry code uses for this call
-- @return class identificator
-- @note Not supported natioal number format
function callClassImpl:getClassByNumber(number, localCountryCode)
    local rv = 0
    number = number and tostring(number) or ""
    if "" == number then
        return rv
    end
    
    local numberObj = require"numberDetect.numberDetectImpl":new()
    if not numberObj:isNormalized(number) then
        return rv
    end
    
    localCountryCode = localCountryCode and tostring(localCountryCode) or ""
    if not localCountryCode or "" == localCountryCode then
        return rv
    end
    if "+" ~= localCountryCode:sub(1,1) then 
        localCountryCode = "+"..localCountryCode
    end
    if 2 > localCountryCode:len() then
        return rv
    end   
    
    local searchSign = "staticRegexp"   -- retrieve call class for other countries
    if number:match("^"..localCountryCode) then 
        logger:debug("Call to %s. Retrieving call class for local country from wnpdb (local country code is %s)", number, localCountryCode)
        searchSign = "numberingPlans"   -- retrieve call class for local country
    else
        logger:debug("Call to %s. Retrieving call class for other countries from callClassDefines (local country code is %s)", number, localCountryCode)
    end
    
    local callClassData = __getCallClassData(searchSign, number)
    rv = callClassData.ident or rv
    
    return rv
end

return callClassImpl