
local log = require"core.logger":init("cdrconnection", 1)
local helper =  require"common.helper"

cdrconnection = {
	engine = {},
	
	init = function(self)
		-- slite3 => backend, table, dbdir, dbfile
		-- mysql && tds => backend, table, dbname, hostname, port, user, pass
		local parameters = self:get_backend()
		
		local cdr_table   = parameters.table
		local cdr_backend = parameters.backend
		self.cdr_bachend  = cdr_backend 
		
		if cdr_backend == "sqlite3" then
			local dbdir = parameters.dbdir or ""
			local dbfile = parameters.dbfile or ""
			if dbdir == "" or dbfile == "" then
				log:error("Wrong parameters for sqlite3 connector")
				return false
			end
			local luasql = require"luasql.sqlite3"
			
			self.engine.env = luasql.sqlite3()
			self.engine.con = self.engine.env:connect(string.format("%s/%s", dbdir, dbfile))

		elseif cdr_backend == "mysql" then
			local database = parameters.dbname or ""
			local hostname = parameters.hostname or "" 
			local port = parameters.port or "3306"
			local user = parameters.user or ""
			local pass = parameters.password or ""

			if database == "" or hostname == "" or user == "" or pass == "" then
				log:error("Wrong parameters for MySQL connector")
				return false
			end
	
			local luasql = require"luasql.mysql"
			self.engine.env = assert(luasql.mysql())
			self.engine.con = self.engine.env:connect(database, user, pass, hostname, port)		
			
		elseif cdr_backend == "tds" then

			local database = parameters.dbname or ""
			local hostname = parameters.hostname or "" 
			local port = parameters.port
			local user = parameters.user or ""
			local pass = parameters.password or ""
			
			if database == "" or hostname == "" or user == "" or pass == "" then
				log:error(string.format("Wrong parameters for %s connector", cdr_backend))
				return false
			end
			
			local luasql = require "luasql.odbc"
			self.engine.env = assert (luasql.odbc())
			self.engine.con = self.engine.env:connect (database, user, pass, hostname, port)
			
		elseif cdr_backend == "csv" then
			log:error("Does not support CSV backend")
			return false
		end
		if not self.engine.con then
			log:error("Сonnection error")
			return false
		end
		
		return cdr_backend, cdr_table
	end;
	
	get_backend = function ()
		local iniParser =  require "inilazy"
		
		local cdr_params = {}
		local conf = iniParser.get('/etc/callweaver/cdr.conf')
		if conf and conf.general and conf.general.ctibackend and conf.general.ctibackend ~= "" then
			cdr_params.backend = helper.trim(conf.general.ctibackend)
		end
		if not cdr_params.backend then
			log:error("CDR backend not found")
			return {}
		end
		
		if string.lower(cdr_params.backend) == 'sqlite3' then
			local confSqlite = iniParser.get('/etc/callweaver/cdr_sqlite3_custom.conf')
			if confSqlite and confSqlite.cdrdb and confSqlite.cdrdb.table ~= "" then
		  		cdr_params.dbdir = helper.trim(confSqlite.cdrdb.dbdir)
		  		cdr_params.table = helper.trim(confSqlite.cdrdb.table)
		  		cdr_params.dbfile = helper.trim(confSqlite.cdrdb.dbfile)
		  		cdr_params.enabled = helper.trim(confSqlite.cdrdb.enabled)
		  	end
		elseif string.lower(cdr_params.backend) == 'mysql' then
			local confMySQL = iniParser.get('/etc/callweaver/cdr_mysql.conf')
			if confMySQL and confMySQL.global and confMySQL.global.table ~= "" then
				cdr_params.enabled = helper.trim(confMySQL.global.enabled)
				cdr_params.hostname = helper.trim(confMySQL.global.hostname)
				cdr_params.dbname = helper.trim(confMySQL.global.dbname)
				cdr_params.table = helper.trim(confMySQL.global.table)
				cdr_params.password = helper.trim(confMySQL.global.password)
				cdr_params.user = helper.trim(confMySQL.global.user)
				cdr_params.port = helper.trim(confMySQL.global.port)
				cdr_params.timeout = helper.trim(confMySQL.global.timeout)
		  	end
		elseif string.lower(cdr_params.backend) == 'tds' then
			local confTDS = iniParser.get('/etc/callweaver/cdr_tds.conf')
			if confTDS and confTDS.global and confTDS.global.table ~= "" then
				cdr_params.enabled = helper.trim(confTDS.global.enabled)
				cdr_params.hostname = helper.trim(confTDS.global.hostname)
				cdr_params.port = confTDS.global.port and helper.trim(confTDS.global.port)
				cdr_params.dbname = helper.trim(confTDS.global.dbname)
				cdr_params.table = helper.trim(confTDS.global.table)
				cdr_params.user = helper.trim(confTDS.global.user)
				cdr_params.password = helper.trim(confTDS.global.password)
		  	end
		end

		return cdr_params;
	end;
	
	execute = function (self, sql, ...)
		local query = string.format(sql, ...)
		local times = 30
		repeat
			local result, err = self.engine.con:execute(query)
			if type(result) ~= "userdata" and type(result) ~= "number" then
				-- handle sql errors
				local tmpstr = ""
				if err ~= nil then
					local errstr = err:match(".*\:%s*(.*)%s*\:.*") 
					if errstr == "no such table" then
						tmpstr = " (set limit to 1)"
						times = 1
					end
				end
				
				local t = {}
				t[#t+1] = "cdr backend: " .. self.cdr_bachend
				t[#t+1] = "iteration: " .. times .. tmpstr
				t[#t+1] = "timestamp: " .. os.time()
				t[#t+1] = "error: " .. err
				t[#t+1] = debug.traceback()
				log:error(t)
			else
				return result
			end

			-- we must wait a time
			require"socket".sleep(math.random(10,50)/10000)

			times = times - 1
		until times == 0
		return self:close()
	end;
	
	close = function (self)
		if self.engine.con then
		  self.engine.con:close()
		  self.engine.con = nil
		  
		  if self.engine.env then
			 self.engine.env:close()
			 self.engine.env = nil
		  end
	   end
	end;
}
