module( ... or "core", package.seeall)

package.path = package.path .. ";/var/www/provision/?.lua"

-- This code doubles functionality of wsqlite.lua and used only in provisioning/xml browser applications
connection = {
	-- Collection of databases
	databases = {
         ["wmsdb"]      = "/rw2/var/lib/wmsdb.d/wmsdb"
        ,["contacts"]	= ""
		,["devdb"]		= "/rw2/var/lib/wmsdb.d/devdb"
		,["htables"]    = "/rw2/var/kamailio/htables.db"
	},

	database = "devdb",     -- default database name
	
	logger = require"core.logger":init("luasqlite", 3),

	-- Switch connection to a database from database collections
	-- @param name A naem of databse
	switch = function (self, name)
		if not self.databases[name] then
			return false
		end

		if require"common.helper".filesize(self.databases[name]) == 0 then
			self.logger:error("Database %q is empty/not found", self.databases[name])
			return false
		end

        return self:connect(name)
	end;

    --- Make database connection
    -- @param name  The database name
    connect = function (self, name)
        if not self.databases[name] then
            return false
        end

        if not self.conn or not self.env then -- lazy database connection
            local luasql = require("luasql.sqlite3")
            self.env = assert(luasql.sqlite3())
        end

        if not self.conn or name ~= self.database then
            self.conn     = assert(self.env:connect(self.databases[name]))
            self.database = name
        end

	    return nil ~= self.conn
   end;

	--- Execute an sql and return iterator function
	-- @param sql_statement A string with sql statment
	-- @param mixed Values to be passed to string.format funciton
	-- @return funciton A function with iterator to results or false for bad request
	rows = function (self, sql_statement, ...)
		local cursor = self:execute (sql_statement, ...)
		if not cursor or type(cursor) ~= "userdata" then
			return false
		end
		local colnames = cursor:getcolnames()
		if not colnames or "table" ~= type(colnames) then
			return false
		end
		local colnum   = #colnames

		return function ()
			local result = {cursor:fetch()}
			if not result[1] then
				cursor:close()
			end

			return unpack(result, 1, colnum)
		end
	end;

	--- Execute an sql and return boolean state of execution
	-- @param sql_statement A string with sql statment
	-- @param mixed Values to be passed to string.format funciton
	-- @return boolean
	execute = function (self, sql_statement, ...)
		if not self.conn then
			self:switch("devdb")
		end

		local errmsg = ""
		do local times = 30
		repeat
			local query = string.format(sql_statement, ...)
			local result, err = self.conn:execute (query)

			if type(result) ~= "userdata" and type(result) ~= "number" then
				-- handle sql errors
				local tmpstr = ""
				if err ~= nil then
					local errstr = err:match(".*\:%s*(.*)%s*\:.*") 
					if errstr == "no such table" then
						tmpstr = " (set limit to 1)"
						times = 1
					end
				end
				
				local t = {}
				t[#t+1] = "database: " .. self.database
				t[#t+1] = "error: " .. err
				t[#t+1] = "query: " .. query .. "\n"
				self.logger:debug(t)
				errmsg = t
			else
				return result
			end

			-- we must wait a time
			require"socket".sleep(math.random (10,50) / 1000)

			times = times - 1
			until times == 0
		end
		self.logger:error(errmsg)
		return self:close(), "Database error"
	end;
	
	--- Close openned database connection
	close = function (self)
		if self.conn then
			self.conn:close()
			if self.env then
				self.env:close()
			end
			self.env = nil
			self.conn = nil
		end
	end;
}

--- Retrieving user funciton keys
-- @param user_features		An user features
-- @param user_number		An extension of the requested user
-- @return table with function keys of the user
function fkeys_get(user_features, user_number)
	local result = {
		["lineKeys"] = {},	-- I think not used !!
		["stdKeys"]  = {},	-- user function keys (Switch, Line, Colleague, etc.)
		["expKeys"]  = {}	-- keys for expansion module
	}
	
	if user_features and user_features["FKEYS"] then
		local keys = require"core.pbxutils".split(user_features["FKEYS"], ",")
		
		for i=1, (tonumber(user_features["FLINES"]) or 2) do
			table.insert(result["lineKeys"], 'LINE')
		end

		local exp_num, is_exp = 0, false
		for i=1, #keys do
			( function ()
				local keys   = keys
				local result = result

				if keys[i] == "" then
					return true
				end
				
				-- update self BLF keys to LINE
				if user_number and keys[i]:find(string.format("^EXT%%s+%s ", user_number)) then
					return true
				end

				if string.find(keys[i],"^---") then
					is_exp, exp_num = true, exp_num + 1
					return true
				end

				if not is_exp then
					table.insert (result["stdKeys"], keys[i])
				else
					local serial = "expMod" .. exp_num

					if not result["expKeys"][serial] then
						result["expKeys"][serial] = {};
					end

					table.insert (result["expKeys"][serial], keys[i])
				end
			end )()
		end
	end

	return result
end

function load_resource (module)
	local base = _G
	local s = "res_" .. module

	if not base[s] then
		base[s] = require("resources." .. module)
	end

	return base[s]
end

--- Read firmwares.json package info file
-- @param agent A table with general device informaion
-- @return table with firmwares info
function get_firmware_pack(agent)
	if "table" ~= type(agent) then
		return false
	end
	
	-- FIx agent for Wildix phone of the new series (R2)
	if "Wildix" == agent["vendor"] and "WPR2" == agent["vendor:additional"] then
		agent["vendor"] = "Wildix2"
	end
	local dir = string.format("/var/www/firmwares/%s/", agent["vendor"]:lower())
	
	local result = nil
	if require"common.helper".is_exists (dir .. "firmwares.json") then
		local fp = io.open(dir .. "firmwares.json")
		result = require"dkjson".decode(fp:read ("*a"))
		io.close (fp)
	end

	return result
end

--- Reading translations.json file for searching location of the translation file
-- @param agent		A table with general device information
-- @return list of the translation files for available languages and files common directory
-- @note used only for WP devices. In case using for other device you have to update this function
function get_translation_pack(agent)
	if "table" ~= type(agent) or not agent.vendor or not agent.model then
		return false
	end
	
	local rv 	   = nil
	local filesDir = nil
	if "Wildix" == agent["vendor"] then
		local translationDir  = "/var/www/firmwares/wildix/"
		local translationFile = translationDir.."translations.json"
		local fp = io.open(translationFile)
		if fp then
			local content = fp:read("*a")
			fp:close()
			rv = require"dkjson".decode(content)
			if rv then
				filesDir = translationDir
    			if agent.model then
    				rv = rv[agent.model]
    			end
    		end
		end
	end
	
	return rv, filesDir
end

--- Retrieve PBX time zone include daylight saving
function getPBXTimeZone(withDaylightSaving)
    local UTCTime   = os.date("!*t")
    local localTime = os.date("*t")
    
    if withDaylightSaving then
        localTime.isdst = false
    end
    
    return os.difftime(os.time(localTime), os.time(UTCTime))
end

tzones = {"-1100","-1000","-0900","-0800","-0700","-0600","-0500","-0430","-0400","-0330","-0300","-0200","-0100","+0000","+0100","+0200","+0300","+0330","+0400","+0430","+0500","+0530","+0545","+0600","+0630","+0700","+0800","+0900","+0930","+1000","+1100","+1200","+1300"}

function get_tzoffset()
	local timezone = getPBXTimeZone()
	local h, m = math.modf(timezone / 3600)
	local tz = string.format("%+.4d", 100 * h + 60 * m)
	if not tz:find("%-") and not tz:find("%+") then
		tz = "+" .. tz
	end
	return tz
end

----------------------------------------------------------------------
-- unknown managmnet

--- Register a device with an unknown account
-- @param mac A string with mac address
-- @param account A string with account
-- @param fxs_ports A table with free fxs ports
function register (mac, account, fxs_ports)
	local connection = connection
	local sql
	if fxs_ports then
		if fxs_ports[1] then	-- registered account set at device page or service login
			account = get_free_unknown_account()
			sql = [[
				insert or replace into autoprov (id,param,value,mac,flag) values(
					(select id from autoprov where param=%q and mac=%q and flag=%q),%q,'%s',%q, %q)
			]]
			for _, port in ipairs (fxs_ports) do
				connection:execute (sql, "username", mac, port, "username", account, mac, port)
			end
		else
			return false
		end
	else
		if is_unknown (mac) and not account then
			return false
		end
		if not account then
			account = get_free_unknown_account()
		end
		sql = [[
			 insert or replace into autoprov (id,param,value,mac) values(
				 (select id from autoprov where param=%q and mac=%q),%q,'%s',%q)
	    ]]
		connection:execute (sql, "username", mac, "username", account, mac)
	end
	return account
end

--- Return list of unknown users
-- @return A table username = mac
-- @return A table mac = username
function get_unknown_users ()
	local connection = connection
	local t, r

	local sql = "select value, mac from autoprov where param=%q and value like %q"

	for username, mac in connection:rows (sql, "username", "unknown%%") do
		if type(t) ~= "table" and type(r) ~= "table" then
			t, r = {}, {}
		end

		t[username], r[mac] = mac, username
	end

	return t, r
end

--- Get a new free unknown account name
-- @return A string with an unknown account
function get_free_unknown_account ()
	local connection = connection
	local users = get_unknown_users ()
	local account = "unknown1"

	if users then
		for i=1,100000 do
			account = "unknown" .. i
			if not users[account] then break  end
		end
	end

	return account
end

-- Check if unknown username already set
function is_unknown (mac)
	local connection = connection
	local sql = "select value from autoprov where mac=%q and param=%q and value like %q"
	local cur = connection:execute (sql, mac, "username", "unknown%%")
	local res = cur:fetch(); cur:close()
	return res ~= nil
end

----------------------------------------------------------------------
--- Device information processing
----------------------------------------------------------------------
--- Retrieve information about device and update by actual data
-- @param actualInfo   Device actual information
-- @return table with information about device
function updDeviceInfo(actualInfo)
    local rv = {}
    if "table" ~= type(actualInfo) then
        return rv
    end
    
    local deviceMac = actualInfo.mac or ""
    if "" == deviceMac then
        return rv
    end
    
    local json       = require"dkjson"
    local connection = connection
    
    -- Retrieve device information from database
    local sql               = "SELECT value, token FROM autoprov WHERE param='device' AND mac='%s'"
    local cursor            = connection:execute(sql, deviceMac)
    local dbInfoJSON, token = cursor:fetch(); cursor:close()
    connection:close()
    local dbInfo            = json.decode(dbInfoJSON or "") or {}

    local needUpdate = false

    -- Proceed reboot flag
    if 1 == dbInfo.reboot then
        rv.reboot = 1
        needUpdate = true
    end

    -- Comparing device info
    for name, val in pairs(actualInfo) do
        if val ~= dbInfo[name] then
            needUpdate = true
        end
        if not dbInfo[name] then
            dbInfo[name] = val
        end
    end
    
    -- Build device information for output
    for name, val in pairs(dbInfo) do
        if needUpdate and actualInfo[name] then
            val = actualInfo[name]
        end
        rv[name] = val
    end
    
    rv.token       = token
    rv.needUpgrade = needUpdate
    
    return rv
end

--- Update device information in database
-- @param devInfo   Device information that needs for updating
-- @return table with information about device
function saveDeviceInfo(devInfo)
    local rv = false
    
    repeat 
        if "table" ~= type(devInfo) then
            break
        end
    
        if not devInfo.mac or not devInfo.token then
            break
        end
        
        -- check upgrade flag
        if not devInfo.needUpgrade then
            rv = true
            break
        end
        
        local connection = connection

        -- build database information
        local dbInfo = {}
        for name, val in pairs(devInfo) do
            dbInfo[name] = val
        end
        
        -- clear variables for upgrading
        dbInfo.needUpgrade = nil
        dbInfo.token       = nil
        dbInfo.reboot      = nil
        
        -- processes device values
        local devToken  = devInfo.token or ""
        local deviceMac = devInfo.mac
        
        -- save device information 
        local dbInfoJSON = require"dkjson".encode(dbInfo) or ""
        local sql        = "INSERT OR REPLACE INTO autoprov (id, param, value, mac, token) VALUES ((SELECT id FROM autoprov WHERE param='device' AND mac='%s'), 'device', '%s', '%s', '%s')"
        
        local cursor = connection:execute(sql, deviceMac, dbInfoJSON, deviceMac, devToken or "")

        if "userdata" == type(cursor) then
            cursor:close()
        end

        connection:close()
        
        rv = true
    until true
            
    return rv
end

--- Processing device information. Merge database data with recieved info from
-- device and update result to database. 
-- @see updDeviceInfo
-- @see saveDeviceInfo
function proceedDeviceInfo(actualInfo)
    local rv = updDeviceInfo(actualInfo)
    saveDeviceInfo(rv)
    return rv
end

----------------------------------------------------------------------
function get_phonebook_names(ids)
	local connection = connection
	local tab = {}

	local tconnection = {
		["cticontacts"] = {
			["dbfile"]      = "/mnt/cticonnect/cticonnectdb",
			["table_cont"]  = "cti_phonebooks"
		}
	}
    local rids = {}

	for _, pbid in ipairs(ids) do
		if pbid == "Directory" then
			tab["Directory"] = "Directory"
		elseif pbid == "searchall" then
			tab["searchall"] = "searchall"
		else
		    table.insert(rids, tonumber(pbid))
		end
	end

	-- Get phonebook names
	for _, tset in pairs(tconnection) do
		connection.databases["contacts"] = tset["dbfile"]
		if connection:switch("contacts") then
			local sql = "SELECT rid, name FROM %s WHERE rid IN (%s)"
			local rids_string = table.concat(rids, ",")
			if rids_string ~= "" then
				local res_sql = connection:rows(sql, tset["table_cont"], rids_string)
				if res_sql then
					for rid, name in res_sql  do
						tab[rid] = name or ""
					end
				end
			end
		end
		connection:close()
	end
	
	return tab
end

--- Retrieving users phonebook identificators
-- @note The same functionality in services.lua:SHORTNUM
function get_list_phonebook(uid, withName)
	local rv 	 = {}
	if not uid then
		return rv
	end
		
	-- get user LDAP info
	local person = {}
	local res 	 = require"core.wmsldap":get("uid", uid)
	if 0 ~= #res then
		person = res[1][2]
	end
	if not person or not person.uid then
		return rv
	end
	
	-- loading user phonebooks
	local userFeatures = require"common.userFeatures.userFeatures":new(person.uid)
	if not userFeatures then
		return rv
	end
	userFeatures:load()

	-- init database connection
	local connection 				 = connection
	connection.databases["contacts"] = "/mnt/cticonnect/cticonnectdb"
	connection:switch("contacts")
	
	-- processing phonebook id list
    local userPb = userFeatures:get("PHONEBOOK-ID") or ""
	local isAllowAll = "yes" == require"common.permissions":check(person["owner"] or "", "canset", "PHONEBOOK", nil, person["licenseType"]) or "" == userPb
	local sql	     = string.format("SELECT rid, name FROM cti_phonebooks AS pb WHERE (pb.owner_id=%d and pb.type=1)", uid)
	if isAllowAll then
		-- select all phonebooks (exception LDAP users)
		sql = sql.." OR pb.type=2"
	else
		-- only user private phonebooks and selected public phonebooks
		sql = string.format("%s OR rid IN (%s)", sql, userPb)
	end
	
	local rSql = connection:rows(sql)
	if rSql then
		for rid, name in rSql do
			if withName then
				rv[name] = rid
			else
				rv[#rv + 1] = rid
			end
		end
	end
	connection:close()
	
	return rv
end

-- Adds sipmle_extension, if only one number in tab
local function __get_simple_ext(tab)
    local tab        = tab or {}
	local simple_ext = nil
    
    if "directory" == tab.pbookname then
        -- colleague (LDAP) entry
        simple_ext = tab["extension"]
    else
        -- phonebook entry
        local entryPriority = {"phone", "office", "home", "extension", "mobile", "home_mobile"}
        for _, tag in pairs(entryPriority) do
            if tab[tag] and "" ~= tab[tag] then
                simple_ext = tab[tag]
                break
            end
        end
    end
    
    if simple_ext then
        tab["simple_ext"] = simple_ext
    end
	return tab
end

-- Retrieve search unique extension
local function __get_search_ext(tab, searchVal)
    local tab      = tab or {}
    local uniq_ext = nil
    
    if not searchVal then
        return tab
    end
    
    local entryList = {"phone", "office", "home", "extension", "mobile", "home_mobile"}
    for _, tag in pairs(entryList) do
        if tab[tag] and tab[tag]:match(searchVal) then
            uniq_ext = tab[tag]
            break
        end
    end
    
    if not uniq_ext then
        -- try to search by user name
        if tab["username"] and tab["username"]:match(searchVal) then
            tab               = __get_simple_ext(tab)
            uniq_ext          = tab["simple_ext"]
            tab["simple_ext"] = nil
        end
    end
    
    if uniq_ext then
        tab["unique_ext"] = uniq_ext
    end
     
    return tab
end

function get_phonebook_entries(pbook, search_filter, limit)
	local connection = connection
	local tab = {}

	local tconnection = {
		["cticontacts"] = {
			["dbfile"]      = "/mnt/cticonnect/cticonnectdb",
			["table_cont"]  = "cti_contacts"
		}
	}

	local phonebook_table_name = "cti_phonebooks"
	local tset         = tconnection["cticontacts"]
	local filter_param = search_filter[1]
	local filter_value = search_filter[2]
    local filter_type  = search_filter[3]
    local order_value  = search_filter[4]

	connection.databases["contacts"] = tset["dbfile"]

	if not connection:switch("contacts") then
		return tab
	end

    if "table" == type(pbook) then
        pbook = table.concat(pbook, ",")
    end

	local sql = "SELECT count(*) as cnt FROM %s WHERE rid IN (%s)"
	local is_phonebook_present = false
	if  "entryuid" == filter_param then
		is_phonebook_present = true
	else
		for phonebookCnt in connection:rows(sql, phonebook_table_name, pbook) do
			if tonumber(phonebookCnt) > 0 then
				is_phonebook_present = true
			end
		end
	end
	if not is_phonebook_present then
		return nil
	end

	local SELECT = "SELECT rid, username, extension, phone, office, mobile, home, home_mobile, user_passport_id, phonebook_id"
	local FROM   = string.format(" FROM %s", tset.table_cont)	
    local WHERE  = string.format(" WHERE phonebook_id IN (%s)", pbook)
    local AND    = ""
    local ORDER  = order_value and " ORDER BY " .. order_value or ""
    local LIMIT  = limit and string.format(" LIMIT %d", limit) or ""

	if filter_value and "" ~= filter_value then
		filter_value = string.gsub(filter_value, "\\", "\\\\")
		filter_value = string.gsub(filter_value, "_", "\\_")
	end
	if  "entryuid" == filter_param and filter_value and "" ~= filter_value then
		AND = [[ AND (rid IS "]] .. filter_value .. [[")]]
	elseif filter_value and "" ~= filter_value then
		if string.match(filter_value, "%D") then -- fulltext search for filter contains non digit symbols
            AND = [[ AND `rowid` IN (SELECT `rowid` FROM `%s_fts4` WHERE `%s_fts4` MATCH "username:%s* OR email:%s* OR organization:%s*")]]
            AND = string.format(AND, tset["table_cont"], tset["table_cont"], filter_value, filter_value, filter_value)
        else
    		AND = [[ AND (username LIKE "%%]] .. 
                  filter_value .. [[%%" ESCAPE "\" OR extension LIKE "%%]] .. 
                  filter_value .. [[%%" ESCAPE "\" OR phone LIKE "%%]] ..
                  filter_value .. [[%%" ESCAPE "\" OR office LIKE "%%]] ..
                  filter_value .. [[%%" ESCAPE "\" OR mobile LIKE "%%]] .. 
                  filter_value .. [[%%" ESCAPE "\" OR home LIKE "%%]] ..
                  filter_value .. [[%%" ESCAPE "\" OR home_mobile LIKE "%%]] ..                   
                  filter_value .. [[%%" ESCAPE "\")]]
        end
	end

    sql = SELECT..FROM..WHERE..AND..ORDER..LIMIT
    
	local sql_res = connection:rows(sql)
	if sql_res then
		for rid, name, extension, phone, office, mobile, home, home_mobile, document_id, pbookId in sql_res do
			tab[#tab+1] = {
				["pbookname"]	= pbookId, 
				["entryuid"]	= rid,
				["username"]    = name,
				["extension"]   = extension or "",
				["phone"]       = phone or "",
				["office"]      = office or "",
				["mobile"]      = mobile or "",
				["home"]        = home or "",
				["home_mobile"] = home_mobile or "",
				["document_id"] = document_id or ""
			}
			if "unique" == filter_type then
                tab[#tab] = __get_search_ext(tab[#tab], filter_value)
            else
                --adds sipmle_extension, if only one number in record
                tab[#tab] = __get_simple_ext(tab[#tab])
            end
		end
	end
	connection:close()

	return tab
end

function get_callable_ldap_entries (user_info, search_filter, limit)
	local wmsldap = require"core.wmsldap"
	local h       = require"common.helper"
	local wmsperm = require"common.permissions"

	local extenNumber = h.get_extenNumber_str()

	local uid			= user_info[1]
	local userKind		= user_info[2]
	local filter_param	= search_filter[1]
	local filter_value	= search_filter[2]
    local filter_type   = search_filter[3]
	local result, attrs	= wmsldap:get ("uid", uid), {}

	if #result ~= 0 then
		_, attrs = unpack (result[1])
	end

	local filter = ""

	local function get_filter (tsrc, action)
		local t = {}
		local pat = action ~= "not" and "(owner=%s)" or "(!(owner=%s))"

		for _, gid in ipairs (tsrc) do
			t[#t+1] = string.format (pat, gid)
		end

		return t
	end

    local wmsldap_connection, wmsldap_basedn = wmsldap:getConection()
    
	local groups_dns         = {}
	local gr_search          = { base = wmsldap_basedn, scope = "subtree", filter = "(objectClass=wGroup)" }
	for gr_dn, gr_attrs in wmsldap_connection:search(gr_search) do
		groups_dns[#groups_dns+1] = gr_dn
	end
	
	local callable_groups_dns = {}
	for _, gr_dn in ipairs(groups_dns) do
		if wmsperm:check(attrs["owner"] or "", "yesno", "CALL", gr_dn, attrs["licenseType"]) == "yes" then
			if wmsperm:check(attrs["owner"] or "", "yesno", "SEE", gr_dn, attrs["licenseType"]) == "yes" then
				callable_groups_dns[#callable_groups_dns+1] = gr_dn
			end
		end
	end

	local tab = {}
   
	if #callable_groups_dns == 0 then
		return tab
	end
	local str = table.concat (get_filter (callable_groups_dns))
	if str ~= "" then
		filter = "(|" .. str .. ")"
	end
	if "entryuid" == filter_param and filter_value and "" ~= filter_value then
		filter = filter .. string.format("(|(%s=*%s*))", "uid", filter_value)
	elseif filter_value and "" ~= filter_value then
		filter = filter .. string.format("(|(%s=*%s*)(cn=*%s*)(mobile=*%s*))", extenNumber, filter_value, filter_value, filter_value)
	end

	filter = string.format ("(&(objectClass=wPerson)%s)", filter)

	local search = { base = wmsldap_basedn, scope = "subtree", filter = filter }
	if limit ~= nil then
		search["sizelimit"] = limit
	end

	for dn, attrs in wmsldap_connection:search (search) do
		if attrs[extenNumber] and "" ~= attrs[extenNumber] then
			local tname = attrs.cn
			local texts = attrs[extenNumber] or ""
			local toffs = attrs.wOfficePhone or ""
			local tmobs = attrs.mobile or ""
			local tkind = attrs.userKind
			local teuid = attrs.uid or ""

			if "room" ~= userKind or ("room" ~= tkind and "fax" ~= tkind) then --Hide room and fax user type for hotel phones
				tab[#tab+1] = {["pbookname"] = "directory", ["entryuid"] = teuid, ["username"] = tname, ["extension"] = texts,
							   ["office"] = toffs, ["mobile"] = tmobs}
			end

			if "unique" == filter_type then
				tab[#tab] = __get_search_ext(tab[#tab], filter_value)
			else
				--adds sipmle_extension, if only one number in record
				tab[#tab] = __get_simple_ext(tab[#tab])
			end
		end
	end
	wmsldap_connection:close()

	return tab
end

function get_ldap_entries_from_ctidb (user_info, search_filter, limit)
    local wmsldap = require"core.wmsldap"
    local wmsperm = require"common.permissions"

    local uid = user_info[1]
    local owner = user_info[2]
    local licenseType = user_info[3]
    local filter_param = search_filter[1]
    local filter_value = search_filter[2]
    local filter_type  = search_filter[3]
    local order_value  = search_filter[4]

    local tab = {}
    local connection = connection
    local tconnection = {
        ["cticontacts"] = {
            ["dbfile"]      = "/mnt/cticonnect/cticonnectdb",
            ["table_cont"]  = "cti_contacts"
        }
    }
    local tset = tconnection["cticontacts"]
    connection.databases["contacts"] = tset["dbfile"]
    if not connection:switch("contacts") then
        return tab
    end

    local SELECT = "SELECT username, source_id, extension, mobile, office, group_id"
    local FROM   = string.format(" FROM %s", tset.table_cont)
    local WHERE  = string.format(" WHERE phonebook_id IN (SELECT rid FROM cti_phonebooks WHERE type='3')")
    local AND    = ""    
    local ORDER  = order_value and " ORDER BY " .. order_value or ""
    local LIMIT  = limit and string.format(" LIMIT %d", limit) or ""

    if filter_value and "" ~= filter_value then
        filter_value = string.gsub(filter_value, "\\", "\\\\")
        filter_value = string.gsub(filter_value, "_", "\\_")
    end

    if "entryuid" == filter_param and filter_value and "" ~= filter_value then
        AND = string.format(" AND source_id='uid=%s,o=com0,dc=wildix'", filter_value)
    elseif filter_value and "" ~= filter_value then
        if string.match(filter_value, "%D") then -- fulltext search for filter contains non digit symbols
            AND = [[ AND `rowid` IN (SELECT `rowid` FROM `%s_fts4` WHERE `%s_fts4` MATCH "username:%s*")]]
            AND = string.format(AND, tset["table_cont"], tset["table_cont"], filter_value)
        else
            AND = [[ AND (username LIKE "%%]] .. 
                  filter_value .. [[%%" ESCAPE "\" OR extension LIKE "%%]] .. 
                  filter_value .. [[%%" ESCAPE "\" OR phone LIKE "%%]] ..
                  filter_value .. [[%%" ESCAPE "\" OR office LIKE "%%]] ..
                  filter_value .. [[%%" ESCAPE "\" OR mobile LIKE "%%]] ..                
                  filter_value .. [[%%" ESCAPE "\")]]
        end
    end

    sql = SELECT .. FROM .. WHERE .. AND .. ORDER .. LIMIT

    local permission_cache = {}

    function check_acl(gr_dn)
        if gr_dn and "" ~= gr_dn then
            if "nil" == type(permission_cache[gr_dn]) then
                if wmsperm:check(owner or "", "yesno", "CALL", gr_dn, licenseType) == "yes" then
                    if wmsperm:check(owner or "", "yesno", "SEE", gr_dn, licenseType) == "yes" then
                        permission_cache[gr_dn] = true
                        return true
                    end
                end
                permission_cache[gr_dn] = false
            end
        end 
        return permission_cache[gr_dn]
    end

    local sql_res = connection:rows(sql)
    if sql_res then
        local permission_cache, permit = {}, false
        for username, source_id, extension, mobile, office, group_id in sql_res do
            if check_acl(group_id) then 
                tab[#tab+1] = {
                    ["pbookname"]   = "directory", 
                    ["entryuid"]    = string.match(source_id, "uid=(%d+),.*"),
                    ["username"]    = username,
                    ["extension"]   = extension,
                    ["office"]      = office or "",
                    ["mobile"]      = mobile or ""
                }
                if "unique" == filter_type then
                    tab[#tab] = __get_search_ext(tab[#tab], filter_value)
                else
                    --adds sipmle_extension, if only one number in record
                    tab[#tab] = __get_simple_ext(tab[#tab])
                end
            end
        end
    end
    connection:close()

    return tab
end

---  XML PARSER, please remove it after switch to XML-RPC library
function parseargs(s)
	local arg = {}
	string.gsub(s, "(%w+)=([\"'])(.-)%2", function (w, _, a)
		arg[w] = a
	end)
	return arg
end

function collect(s)
	local stack = {}
	local top = {}
	table.insert(stack, top)
	local ni,c,label,xarg, empty
	local i, j = 1, 1
	while true do
		ni,j,c,label,xarg, empty = string.find(s, "<(%/?)([%w:]+)(.-)(%/?)>", i)
		if not ni then break end
		local text = string.sub(s, i, ni-1)
		if not string.find(text, "^%s*$") then
			table.insert(top, text)
		end
		if empty == "/" then  -- empty element tag
			table.insert(top, {label=label, xarg=parseargs(xarg), empty=1})
		elseif c == "" then   -- start tag
			top = {label=label, xarg=parseargs(xarg)}
			table.insert(stack, top)   -- new level
		else  -- end tag
			local toclose = table.remove(stack)  -- remove top
			top = stack[#stack]
			if #stack < 1 then
				error("nothing to close with "..label)
			end
			if toclose.label ~= label then
				error("trying to close "..toclose.label.." with "..label)
			end
			table.insert(top, toclose)
		end
		i = j+1
	end
	local text = string.sub(s, i)
	if not string.find(text, "^%s*$") then
		table.insert(stack[#stack], text)
	end
	if #stack > 1 then
		error("unclosed "..stack[#stack].label)
	end
	return stack[1]
end

--- Retrieving and notmalizing user phonebook records by jabber statuses
function get_jabber_presence_statuses(records)
    local users = {}                            -- list of the local users to retrieve presence
    for _, data in ipairs(records) do
        if data and data.extension and "directory" == data.pbookname then
            users[#users + 1] = data.extension
        end
    end
    
    local statusAlias = {
	    dnd     = "donotdistrub",
	    ringing = "Ringing",
	    talking = "inuse",
	    rt      = "Ringinginuse"
    }
    
    local statusList = require"common.userFeatures.userStatus":getDataFromJabberbyPresence(users)
    for idx, data in ipairs(records) do
        if data and data.extension then
            local status = statusList[data.extension]
            if status then
                records[idx]["presence"] = statusAlias[status] or status
            end
        end
    end
    
    return records
end

--- Encrypt the cipher according to mode end using the key
-- @param cipher A string to be encrypted
-- @param mode an encrypting mode (supports aes)
-- @param key an encrypting key
-- @return Encrypted cipher, error message.
function encrypt_cipher(cipher, mode, key)
local command	= nil
local output	= nil
local err		= nil
	if not cipher or "" == cipher then
		err = "Error, cipher value is empty."
	elseif not mode or "" == mode then
		err = "Error, mode value is empty." 
	elseif not key or "" == key then
		err = "Error, key value is empty."
	elseif "aes" == mode then
		command = "echo %s | nice -n 19 openssl enc -e -aes-128-ecb -nosalt -a -A -md md5 -pass pass:%s"
		output = tostring(require"core.pbxutils".run(command, cipher, key))			
		if not output or "" == output then err = "Error encrypt cipher." end
	else
		err = "Error encrypt cipher, " .. mode .. " mode not supported." 
	end
	return output, err
end

--- Get random bytes from system device
-- @param number - number of bytes
function get_random_bytes(number)
	local file = io.open("/dev/urandom")
	local random = file:read(number)
	file:close()
	return random
end
