--- Module for handling failover configuration
--
-- @file    faoliver.lua
-- @author  Gennadiy Ganchev
-- @created 2017/04/27

local logger = require"core.logger":init("failover", 2)

--- @module deviceFailover
local failover = {}
----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------

local function __getLdapData(param, value, class)
    local rv  = nil
    local res = require"core.wmsldap":get(param, value, class)
    if res and res[1] and res[1][2] then
        rv = res [1][2]
    end
    
    return rv
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local failoverImpl = { -- meta Class
     status = nil   -- failover status
    ,cfg    = nil   -- failover configuration
--        ,pbxOwnerDn  = nil   -- PBX owner dn 
--        ,pbxFailover = nil   -- the failover server public IP or hostname which will be failover for current PBX
--        ,sipPort     = nil   -- failover SIP server port
}

function failoverImpl:__loadData(dataSrc)
    if "wmsldap" == dataSrc then
        local pbxOwnerDn = self.cfg.pbxOwnerDn or self.cfg.pbxLocalDn -- set PBX owner as local PBX dn if PBX owner dn is not set 
        local wFailover = __getLdapData("pbxowner", pbxOwnerDn, "wFailover")
        if wFailover and "1" == wFailover.enabled then
            self.status                = "1"
            local pbxFailoverDn        = wFailover and wFailover.pbxFailover
            local pbxFailover          = __getLdapData("dn", pbxFailoverDn, "wPbx")
            self.cfg.pbxFailoverDn     = pbxFailoverDn
            self.cfg.pbxFailoverDomain = pbxFailover and pbxFailover.pbxDomain
            self.cfg.sipPort           = pbxFailover and pbxFailover.pbxPublicPort or "443"
        end
    end
end

---Retreive device failover global status
-- @return 1 or 0 (enabled or disabled)
function failoverImpl:isEnabled()
     return self.status
end

--- Retreive device failover configuration
-- @return configuration table
function failoverImpl:getConfig()
    local rv = {
               pbxFailoverDn     = self.cfg.pbxFailoverDn
              ,pbxFailoverDomain = self.cfg.pbxFailoverDomain
              ,sipPort           = self.cfg.sipPort
          }
    return rv
end

--- Check if current pbx can be failover for requested pbxDn
-- @return true or false
function failoverImpl:localPbxIsFailover() --isUserOnLocalPBX
    local rv = false
    local remoteOwner = self.cfg.pbxOwnerDn and self.cfg.pbxOwnerDn ~= self.cfg.pbxLocalDn -- PBX owner is not local PBX
    local canFailover = self.cfg.pbxLocalDn == self.cfg.pbxFailoverDn and remoteOwner      -- local PBX can failover PBX owner
    if "1" == self.status and canFailover then
        rv = true
    end
    return rv
end

--- Retreive failover PBX's dn
-- @return dn or false
function failoverImpl:getPbxDN() --isUserOnLocalPBX
    local rv = false
    if "1" == self.status and self.cfg.pbxFailoverDn then
        rv = self.cfg.pbxFailoverDn
    end
    return rv
end

-- TODO: make new methods for check if source or destination user is serving by local pbx
-- failover:is4Caller() --> check cid for local PBX + status for failover configuration
-- failover:is4Called() --> check SIP header + status for failover configuration

--- Object constructor
function failover:new(pbxOwnerDn)
    local obj = setmetatable({}, {__index = failoverImpl})
    
    -- initialize object
    obj.status = "0"
    obj.cfg = {
         pbxLocalDn  = require"core.pbxutils".getPbxDNBySerial()
        ,pbxOwnerDn  = pbxOwnerDn
        ,pbxFailover = nil
        ,sipPort     = nil
    }

    -- WMS-5356 do not use failover configuration from ldap if wms network disabled
    local wms_net_conf = require"common.helper".parse_simple_ini("/rw2/etc/pbx/net.conf")
    if not (wms_net_conf and "0" == wms_net_conf["STARTUP"]) then
        obj:__loadData("wmsldap")
    end

    return obj
end

return failover
