module( ... or "helper", package.seeall)

pbxutils = require"core.pbxutils"

--- ----------------------
-- Some useful functions
--- ----------------------

function parse_query_string(str)
	if not str and #str > 0 then
		return nil
	end
	local params = {}
	for k,v in str:gmatch("([^?&]+)=([^&]+)") do
	   params[k] = v
	end

	return params
end

-- TODO: PRESENT at pbxutils.lua, remove from this file and update all link functions
function split(str, pat)
	local str = str or ""

	local t = {}  -- NOTE: use {n = 0} in Lua-5.0
	local fpat = "(.-)" .. pat
	local last_end = 1
	local s, e, cap = tostring(str):find(fpat, 1)
	while s do
		if s ~= 1 or cap ~= "" then
			table.insert(t,cap)
		end
		last_end = e+1
		s, e, cap = str:find(fpat, last_end)
	end
	if last_end <= #str then
		cap = str:sub(last_end)
		table.insert(t, cap)
	end
	return t
end

--- Deep copy of a given table
-- @param object A table to be copied
-- @return object A copy of a given table
-- TODO: PRESENT at pbxutils.lua, remove from this file and update all link functions
function tcopy (object)
	local lookup_table = {}
	local function _copy(object)
		if type(object) ~= "table" then
			return object
		elseif lookup_table[object] then
			return lookup_table[object]
		end
		local new_table = {}
		lookup_table[object] = new_table
		for index, value in pairs(object) do
			new_table[_copy(index)] = _copy(value)
		end
		return setmetatable(new_table, getmetatable(object))
	end
	return _copy(object)
end

-- TODO: PRESENT at pbxutils.lua, remove from this file and update all link functions
local dstring
function tprint (tt, indent, done)
	if tt == nil then return "" end
	if done == nil then dstring = "" end

	dstring = dstring or ""
	done = done or {}
	indent = indent or 0
	if type(tt) == "table" then
		for key, value in pairs (tt) do
			dstring = dstring .. string.rep (" ", indent) -- indent it
			if type (value) == "table" and not done [value] then
				done [value] = true
				dstring = dstring .. string.format("[\"%s\"] = {\n", tostring (key));
				tprint (value, indent + 7, done)
				dstring = dstring .. string.rep (" ", indent+4) -- indent it
				dstring = dstring .. "},\n";
			else
				dstring = dstring .. string.format("[\"%s\"] = \"%s\",\n", tostring (key), tostring(value))
			end
		end
		if string.find(dstring, "\n$") then dstring = string.sub(dstring,1,-2) end -- remove last new enter
	else
		dstring = dstring .. tt .. "\n"
	end
	return dstring
end

function print_r ( t )
 	local print_r_cache={}
	local function sub_print_r(t,indent)
		if (print_r_cache[tostring(t)]) then
			io.write(indent.."*"..tostring(t))
		else
			print_r_cache[tostring(t)]=true
			if (type(t)=="table") then
				for pos,val in pairs(t) do
					if (type(val)=="table") then
						io.write(indent.."["..pos.."] => "..tostring(t).." {\n")
						sub_print_r(val,indent..string.rep(" ",string.len(pos)+8))
						io.write(indent..string.rep(" ",string.len(pos)+6).."}\n")
					else
						io.write(indent.."["..pos.."] => "..tostring(val).."\n")
					end
				end
			else
				io.write(indent..tostring(t))
			end
		end
	end
	sub_print_r(t,"  ")
end

function print(msg)
	return io.stderr:write( msg .. "\n" )
end

--- Check value in a plain table
-- @param e A value to lookup
-- @paema t Source table
-- @return boolean
-- TODO: PRESENT at pbxutils.lua, remove from this file and update all link functions
function in_table (e, t)
	for i=1,#t do
		if (t[i] == e) then return true end
	end
	return false
end

--- Check file presence on the storage disk
-- @param filename	A string with absolute path to file
-- @return boolean value
function is_exists(filename)
	if not filename then
		return false
	end
	local file = io.open(filename)
	return file ~= nil and io.close(file)
end

--- Get file size
-- @param file A string with absolute path to file
-- @return integer
function filesize(filename)
	local file, size = io.open(filename), 0
	if file ~= nil then
		local current = file:seek() -- get current position
		size = file:seek("end")     -- get file size
		file:seek("set", current)   -- restore position
		io.close(file)
	end
	return size
end

--- Convert string to XML format (using some encoding)
-- @param value		A string value
-- @return XML string
function toxmlstring(value)
	if not value or "" == value then
		return ""
	end

	value = string.gsub (value, "&", "&amp;");			-- '&' -> "&amp;"
	value = string.gsub (value, "<", "&lt;");			-- '<' -> "&lt;"
	value = string.gsub (value, ">", "&gt;");			-- '>' -> "&gt;"
	value = string.gsub (value, "'", "&apos;");			-- '\'' -> "&apos;"
	value = string.gsub (value, "\"", "&quot;");		-- '"' -> "&quot;"

	-- replace non printable char -> "&#xD;"
	value = string.gsub(value, "([^%w%&%;%p%\t% ])",
		function (c)
			return string.format("&#x%X;", string.byte(c))
--			return string.format("&#x%02X;", string.byte(c))
--			return string.format("&#%02d;", string.byte(c))
		end);
	return value;
end

function is_addr_leased (address)
	return pbxutils.run([[ fgrep %q /etc/dnsmasq.leases ]], address):find(address)
end

--- Get default gateway set via DHCP
-- @return string 	DHCP interface
-- @return string	default getaway ip
function get_dhcp_gateway()
	local iface = trim(pbxutils.run[[ sed -ne 's/^DMASQINTERFACE=\(.*\)/\1/p' /etc/dnsmasq.d/config 2>/dev/null ]])
	local gw = pbxutils.run[[ sed -ne 's/.*\:router,\(.*\)/\1/p' /etc/dnsmasq.d/dhcp.conf 2>/dev/null ]]
	gw = gw:match("^%d+%.%d+%.%d+%.%d+")
	return iface, gw
end

-- Get pbx network configurations
function get_network(address)
	local storage = nmap and nmap.registry or _G
	if storage.networkSettings then
		return storage.networkSettings
	end

	local t = {}

	-- get interfaces
	local n = pbxutils.run[[ (echo -n "return {"; ip addr | awk '/inet[^6]/&&$0!~/host lo|tun/{if($5=="secondary"){$7=$6};printf"[\""$7"\"]=\""$2"\","}'; echo -n "}") ]]
	local net = assert (loadstring(n))()

	-- get default route
	local g = pbxutils.run[[ (echo -n "return {"; ip route | awk '/default/{printf"[\""$5"\"]=\""$3"\","}';  echo -n "}") ]]
	local gws = assert (loadstring(g))()

	local is_leased = address and is_addr_leased(address) or false
	if is_leased then
		local dhcp_iface, dhcp_gw = get_dhcp_gateway()
		gws[dhcp_iface] = dhcp_gw
	end

	for iface, _ in pairs (net) do
		if not t[iface] then t[iface] = {} end
		if net[iface] then t[iface]["address"] = net[iface] end
		if gws[iface] then t[iface]["gateway"] = gws[iface] end
	end

	storage.networkSettings = t
	return t
end

function set_addr_static (address, mac)
	local t = {}

	-- convert 9c7514654321 -> 9c:75:14:65:43:2
	for i=1,12,2 do
		t[#t+1] = string.sub (mac,i,i+1)
	end

	local mac_str = table.concat (t, ":")
	local str = string.format ("%s %s", mac_str, address)

	-- it is neccessary to restart dnsmasq
	return pbxutils.run ([[
	 fethers=$(mktemp -p/dev/shm/wmsdb.d/);
	 (sed '/%s/d;/%s/d' /etc/ethers; echo %q) > $fethers;
	 cp $fethers /etc/ethers; rm $fethers; sudo /etc/init.d/dnsmasq restart & ]], mac_str, address, str)
end

function is_addr_present (address, mac)
	local json = require"dkjson"
	local core = require"common.core"
	local connection = core.connection
	connection:switch ("devdb")

	local sql = [[
		 select count(value) from autoprov
			where param=%q and value like '%%"address":"%s"%%' and mac<>%q
	]]

	local cursor = connection:execute (sql,"device", address, mac)
	local count = cursor:fetch(); cursor:close()
	connection:close()

	return tonumber (count) > 0
end

function is_vlan (address)
	if not address or "" == address then
		return false, "0"
	end

	local interface = trim(pbxutils.run([[
		ip addr show up scope global label "eth*.*" | awk '/inet %s/{print$7}'
	]], address))

	return interface ~= "", string.match(interface, "%.(%d+)") or "0"
end

function trim (str)
	return string.match (str, "^%s*(.-)%s*$") or ""
end


-- calculate netmask and network
-- TODO use pure lua
local function __ip_calculator(addr, mask)
	local cmd = "ipcalc -nb " .. addr .. "/" .. mask
	local out = pbxutils.run(cmd) or ""
	local netmask = string.match(out, "Netmask:%s+([0-9.]+)")
	local network = string.match(out, "Network:%s+([0-9.]+)")
	return network, netmask
end

function host_settings (address, ...)
	--TODO: use bit operations or ipOp library
	local network = get_network(address)
	local subject = select (1, ...) -- can be an ipaddr or interface

	-- calculate network and mask from where @subject
	if subject then
		if string.find(subject, "(%d+)%.(%d+)%.(%d+)%.(%d+)") then
			local addr, bitmask, netmask, cur
			for interface, info in pairs (network) do
				addr, bitmask = info["address"]:match("([0-9.]*)/(%d+)")
				if addr == subject then
					cur, netmask = __ip_calculator(addr, bitmask)
					local net = __ip_calculator(address, bitmask)
					if cur == net then
						return addr, netmask, info["gateway"] or addr, bitmask, interface, net
					end
				end
			end
		elseif network[subject] then
			local interface = subject
			local info = network[interface]
			local addr, bitmask = info["address"]:match("([0-9.]*)/(%d+)")
			local net, netmask = __ip_calculator(addr, bitmask)
			return addr, netmask, info["gateway"] or addr, bitmask, interface, net
		end
	end

	-- common algorithm
	local addr, bitmask, netmask, cur, net
	for interface, info in pairs (network) do
		addr, bitmask = info["address"]:match("([0-9.]*)/(%d+)")
		cur, netmask = __ip_calculator(addr, bitmask)
		net = __ip_calculator(address, bitmask)
		if cur == net then
			return addr, netmask, info["gateway"] or addr, bitmask, interface, net
		end
	end

	return false
end

-- Check @ip is local device
function is_local_route (ip)
	local g = pbxutils.run [[ echo -n "return {"; route -n | awk '{if($2=="0.0.0.0") printf"[\""$1"\"]=\""$3"\","}';  echo -n "}" ]]
	local local_routes, res_net = assert(loadstring(g))()
	for net, mask in pairs(local_routes) do
		if mask == "255.255.255.255" then
			if ip == net then return true end
		else
			res_net = __ip_calculator(ip, mask)
			if res_net == net then return true end
		end
	end
	return false
end

function get_admin_lang()
	local lang = "en"
	local person = {}
	local res = require"core.wmsldap":get("cn", "admin")
	if #res ~= 0 then
		_, person = unpack(res[1])
		lang = person["preferredLanguage"] or "en"
	end

	return lang
end

function get_dialtone(vendor, country)
    local dialtone_t = {
        welltech  = {},
        welltech04 = {
            ["au"] = "0,400,13,500,0,10,425,13,0,0,10,0,0,0,0,0,1,425,13,100,40,10,0,13,0,0,10,0,0,0,0,0,1,400,13,400,200,10,450,13,400,2000,10,0,0,0,0,0,1,425,13,375,375,10,0,13,0,0,10,0,0,0,0,0,1,425,13,200,200,10,0,13,200,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,375,375,10,0,13,0,0,10,0,0,0,0,0,1,400,13,375,375,10,0,13,0,0,10,0,0,0,0,0,"
           ,["be"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,30,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,3000,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,1400,13,170,170,10,0,13,170,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,170,170,10,0,13,0,0,10,0,0,0,0,0,"
           ,["br"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,30,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,4000,10,0,13,0,0,10,0,0,0,0,0,1,425,13,250,250,10,0,13,0,0,10,0,0,0,0,0,1,425,13,50,1000,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,250,250,10,0,13,0,0,10,0,0,0,0,0,1,0,13,0,0,10,0,13,0,0,10,0,0,0,0,0,"
           ,["cn"] = "0,450,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,450,13,350,350,10,0,13,0,0,10,0,0,0,0,0,1,450,13,1000,4000,10,0,13,0,0,10,0,0,0,0,0,1,450,13,350,350,10,0,13,0,0,10,0,0,0,0,0,1,450,13,400,4000,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,450,13,350,350,10,0,13,0,0,10,0,0,0,0,0,1,450,13,700,700,10,0,13,0,0,10,0,0,0,0,0,"
           ,["cz"] = "1,425,13,330,330,10,0,13,660,660,10,0,0,0,0,0,1,425,13,1800,200,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1500,3000,10,0,13,0,0,10,0,0,0,0,0,1,425,13,330,330,10,0,13,0,0,10,0,0,0,0,0,1,425,13,330,9000,10,0,13,0,0,10,0,0,0,0,0,1,1440,13,500,500,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,330,330,10,0,13,0,0,10,0,0,0,0,0,1,425,13,165,165,10,0,13,0,0,10,0,0,0,0,0,"
           ,["fi"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,650,250,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,4000,10,0,13,0,0,10,0,0,0,0,0,1,425,13,300,300,10,0,13,0,0,10,0,0,0,0,0,1,425,13,20,20,10,0,13,20,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,300,300,10,0,13,0,0,10,0,0,0,0,0,1,425,13,200,200,10,0,13,0,0,10,0,0,0,0,0,"
           ,["fr"] = "0,440,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,440,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,440,13,1500,3500,10,0,13,0,0,10,0,0,0,0,0,1,440,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,440,13,300,0,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,440,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,0,13,0,0,10,0,13,0,0,10,0,0,0,0,0,"
           ,["de"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,0,400,13,500,0,10,425,13,0,0,10,0,0,0,0,0,1,425,13,1000,4000,10,0,13,0,0,10,0,0,0,0,0,1,425,13,480,480,10,0,13,0,0,10,0,0,0,0,0,1,425,13,200,200,10,0,13,200,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,425,13,30,30,10,0,13,30,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,480,480,10,0,13,0,0,10,0,0,0,0,0,1,425,13,240,240,10,0,13,0,0,10,0,0,0,0,0,"
           ,["gr"] = "1,425,13,200,300,10,0,13,700,800,10,0,0,0,0,0,1,400,13,200,300,10,425,13,700,800,10,0,0,0,0,0,1,425,13,1000,4000,10,0,13,0,0,10,0,0,0,0,0,1,425,13,300,300,10,0,13,0,0,10,0,0,0,0,0,1,425,13,300,0,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,300,300,10,0,13,0,0,10,0,0,0,0,0,1,425,13,150,150,10,0,13,0,0,10,0,0,0,0,0,"
           ,["hk"] = "0,350,13,500,0,10,440,13,0,0,10,0,0,0,0,0,1,425,13,650,250,10,0,13,0,0,10,0,0,0,0,0,1,440,13,400,200,10,480,13,400,300,10,0,0,0,0,0,1,480,13,500,500,10,620,13,0,0,10,0,0,0,0,0,1,425,13,40,0,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,480,13,500,500,10,620,13,0,0,10,0,0,0,0,0,1,480,13,240,250,10,620,13,0,0,10,0,0,0,0,0,"
           ,["hu"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,0,350,13,500,0,10,400,13,0,0,10,0,0,0,0,0,1,425,13,1250,3750,10,0,13,0,300,10,0,0,0,0,0,1,425,13,300,300,10,0,13,0,0,10,0,0,0,0,0,1,425,13,40,0,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,425,13,300,300,10,0,13,300,10,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,300,300,10,0,13,0,0,10,0,0,0,0,0,1,425,13,300,300,10,0,13,0,0,10,0,0,0,0,0,"
           ,["is"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,400,40,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,4000,10,0,13,0,300,10,0,0,0,0,0,1,425,13,250,250,10,0,13,0,0,10,0,0,0,0,0,1,425,13,20,20,10,0,13,20,20,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,425,13,400,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,250,250,10,0,13,0,0,10,0,0,0,0,0,1,425,13,250,250,10,0,13,0,0,10,0,0,0,0,0,"
           ,["in"] = "0,400,13,500,0,10,0,13,0,0,10,0,0,0,0,0,0,400,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,400,13,400,200,10,0,13,400,200,10,0,0,0,0,0,1,400,13,750,750,10,0,13,0,0,10,0,0,0,0,0,1,400,13,200,100,10,0,13,200,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,400,13,250,250,10,0,13,250,3250,10,0,0,0,0,0,1,400,13,750,750,10,0,13,0,0,10,0,0,0,0,0,1,400,13,250,250,10,0,13,0,0,10,0,0,0,0,0,"
           ,["id"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,400,13,400,200,10,0,13,400,200,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,150,150,10,0,13,200,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,250,250,10,0,13,0,0,10,0,0,0,0,0,"
           ,["kw"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,4000,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,0,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,250,250,10,0,13,0,0,10,0,0,0,0,0,"
           ,["kr"] = "0,350,13,500,0,10,440,13,0,0,10,0,0,0,0,0,0,350,13,500,0,10,440,13,0,0,10,0,0,0,0,0,1,440,13,1000,2000,10,480,13,0,0,10,0,0,0,0,0,1,480,13,500,500,10,620,13,0,0,10,0,0,0,0,0,1,350,13,250,250,10,440,13,250,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,480,13,500,500,10,620,13,0,0,10,0,0,0,0,0,1,480,13,300,200,10,620,13,0,0,10,0,0,0,0,0,"
           ,["jp"] = "0,400,13,500,0,10,0,13,0,0,10,0,0,0,0,0,0,400,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,400,13,1000,2000,10,0,13,0,0,10,0,0,0,0,0,1,400,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,250,250,10,0,13,250,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,400,13,130,130,10,0,13,130,640,10,0,0,0,0,0,1,400,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,0,13,0,0,10,0,13,0,0,10,0,0,0,0,0,"
           ,["it"] = "1,425,13,200,200,10,0,13,600,1000,10,0,0,0,0,0,0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,4000,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,500,500,10,0,0,0,0,0,1,425,13,400,100,10,0,13,250,100,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,500,500,10,0,0,0,0,0,1,425,13,200,200,10,0,13,200,200,10,0,0,0,0,0,"
           ,["il"] = "0,400,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,400,13,1000,250,10,0,13,0,0,10,0,0,0,0,0,1,400,13,1000,3000,10,0,13,0,0,10,0,0,0,0,0,1,400,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,500,500,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,400,13,50,2000,10,0,13,0,0,10,0,0,0,0,0,1,400,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,0,13,0,0,10,0,13,0,0,10,0,0,0,0,0,"
           ,["ie"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,400,13,400,200,10,450,13,400,2000,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,180,200,10,0,13,200,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,0,13,0,0,10,0,13,0,0,10,0,0,0,0,0,"
           ,["ir"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,4000,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,200,200,10,0,13,200,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,250,250,10,0,13,0,0,10,0,0,0,0,0,"
           ,["iq"] = "0,400,13,500,0,10,0,13,0,0,10,0,0,0,0,0,0,400,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,400,13,1000,2000,10,0,13,0,0,10,0,0,0,0,0,1,400,13,1000,1000,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,1000,1000,10,0,13,0,0,10,0,0,0,0,0,1,0,13,0,0,10,0,13,0,0,10,0,0,0,0,0,"
           ,["mo"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1800,200,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,4000,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,200,600,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,250,250,10,0,13,0,0,10,0,0,0,0,0,"
           ,["my"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1800,200,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,3000,10,0,13,1000,5000,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,200,600,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,250,10,0,13,0,0,10,0,0,0,0,0,"
           ,["mx"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1800,200,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,4000,10,0,13,0,0,10,0,0,0,0,0,1,425,13,250,250,10,0,13,0,0,10,0,0,0,0,0,1,425,13,200,600,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,250,250,10,0,13,0,0,10,0,0,0,0,0,1,0,13,0,0,10,0,13,0,0,10,0,0,0,0,0,"
           ,["nl"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,50,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,4000,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,250,250,10,0,13,0,0,10,0,0,0,0,0,"
           ,["nz"] = "0,400,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,400,13,1800,200,10,0,13,0,0,10,0,0,0,0,0,1,400,13,400,200,10,450,13,400,200,10,0,0,0,0,0,1,400,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,100,100,10,0,13,100,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,400,13,500,500,10,450,13,500,2500,10,0,0,0,0,0,1,400,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,250,250,10,0,13,0,0,10,0,0,0,0,0,"
           ,["no"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1800,200,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,4000,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,200,600,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,250,250,10,0,13,0,0,10,0,0,0,0,0,"
           ,["pk"] = "0,400,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,400,13,1800,200,10,0,13,0,0,10,0,0,0,0,0,1,400,13,1000,2000,10,0,13,0,0,10,0,0,0,0,0,1,400,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,250,250,10,0,13,0,0,10,0,0,0,0,0,"
           ,["pa"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1800,200,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1200,4650,10,0,13,0,0,10,0,0,0,0,0,1,425,13,320,320,10,0,13,0,0,10,0,0,0,0,0,1,425,13,180,180,10,0,13,180,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,425,13,180,180,10,0,13,500,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,320,320,10,0,13,0,0,10,0,0,0,0,0,1,0,13,0,0,10,0,13,0,0,10,0,0,0,0,0,"
           ,["py"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1800,200,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,4000,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,300,0,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,250,250,10,0,13,0,0,10,0,0,0,0,0,"
           ,["pe"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,100,10,0,13,0,0,10,0,0,0,0,0,1,425,13,150,300,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,180,180,10,0,13,180,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,200,200,10,0,13,0,0,10,0,0,0,0,0,"
           ,["ph"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,0,400,13,500,0,10,25,13,0,0,10,0,0,0,0,0,1,425,13,1000,4000,10,480,13,0,0,10,0,0,0,0,0,1,480,13,500,500,10,620,13,0,0,10,0,0,0,0,0,1,440,13,300,0,10,0,13,0,0,10,0,0,0,0,0,1,1400,13,200,300,10,1800,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,950,13,650,320,10,1400,13,330,2000,10,0,0,0,0,0,1,480,13,500,500,10,620,13,0,0,10,0,0,0,0,0,1,480,13,250,250,10,620,13,0,0,10,0,0,0,0,0,"
           ,["pl"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,0,425,13,500,0,10,350,13,0,0,10,0,0,0,0,0,1,425,13,1000,4000,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,150,150,10,0,13,150,4000,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,500,500,10,0,13,0,0,10,0,0,0,0,0,"
           ,["pt"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,200,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,5000,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,200,200,10,0,13,200,0,10,0,0,0,0,0,1,1400,13,200,270,10,1800,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1500,5120,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,360,360,10,0,13,0,0,10,0,0,0,0,0,"
           ,["ro"] = "0,450,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,450,13,1800,200,10,0,13,0,0,10,0,0,0,0,0,1,450,13,1850,4150,10,0,13,0,0,10,0,0,0,0,0,1,450,13,150,180,10,0,13,0,0,10,0,0,0,0,0,1,450,13,200,600,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,450,13,150,180,10,0,13,0,0,10,0,0,0,0,0,1,450,13,170,170,10,0,13,0,0,10,0,0,0,0,0,"
           ,["sa"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1800,200,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1200,4600,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,150,200,10,0,13,150,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,250,250,10,0,13,0,0,10,0,0,0,0,0,"
           ,["sg"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1800,200,10,0,13,0,0,10,0,0,0,0,0,1,425,13,400,200,10,0,13,400,2000,10,0,0,0,0,0,1,425,13,750,750,10,0,13,0,0,10,0,0,0,0,0,1,425,13,300,200,10,0,13,300,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,750,750,10,0,13,0,0,10,0,0,0,0,0,1,425,13,250,250,10,0,13,0,0,10,0,0,0,0,0,"
           ,["za"] = "0,400,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,400,13,25,25,10,0,13,10000,0,10,0,0,0,0,0,1,400,13,400,200,10,0,13,400,2000,10,0,0,0,0,0,1,400,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,400,0,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,250,250,10,0,13,0,0,10,0,0,0,0,0,"
           ,["es"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,100,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,4000,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,180,180,10,0,13,180,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1400,13,400,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,250,250,10,0,13,0,0,10,0,0,0,0,0,"
           ,["sz"] = "0,400,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,400,13,1800,200,10,0,13,0,0,10,0,0,0,0,0,1,400,13,400,200,10,0,13,400,4000,10,0,0,0,0,0,1,400,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,200,600,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,400,400,10,0,13,0,0,10,0,0,0,0,0,"
           ,["se"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,320,20,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,4000,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,200,600,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1400,13,100,100,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,250,250,10,0,13,0,0,10,0,0,0,0,0,"
           ,["ch"] = "0,400,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,400,13,1800,200,10,0,13,0,0,10,0,0,0,0,0,1,400,13,400,200,10,0,13,400,4000,10,0,0,0,0,0,1,400,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,200,600,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,400,400,10,0,13,0,0,10,0,0,0,0,0,"
           ,["tw"] = "0,350,13,500,0,10,440,13,0,0,10,0,0,0,0,0,1,350,13,1000,100,10,440,13,0,0,10,0,0,0,0,0,1,440,13,1000,2000,10,480,13,0,0,10,0,0,0,0,0,1,480,13,500,500,10,620,13,500,500,10,0,0,0,0,0,1,350,13,250,250,10,440,13,250,0,10,0,0,0,0,0,0,1400,13,10000,0,10,1750,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,480,13,500,500,10,620,13,500,500,10,0,0,0,0,0,1,480,13,250,250,10,620,13,250,250,10,0,0,0,0,0,"
           ,["th"] = "0,400,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,400,13,400,100,10,0,13,400,10000,10,0,0,0,0,0,1,400,13,1000,4000,10,0,13,0,0,10,0,0,0,0,0,1,400,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,100,0,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,300,300,10,0,13,0,0,10,0,0,0,0,0,"
           ,["tr"] = "0,450,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,450,13,1000,250,10,0,13,0,0,10,0,0,0,0,0,1,450,13,2000,4000,10,0,13,0,0,10,0,0,0,0,0,1,450,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,450,13,200,600,10,0,13,200,0,10,0,0,0,0,0,1,1400,13,100,100,10,1980,13,100,100,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,450,13,200,200,10,0,13,600,200,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,450,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,450,13,200,200,10,0,13,0,0,10,0,0,0,0,0,"
           ,["gb"] = "0,350,13,500,0,10,440,13,0,0,10,0,0,0,0,0,1,350,13,750,750,10,440,13,0,0,10,0,0,0,0,0,1,400,13,1000,200,10,450,13,400,2000,10,0,0,0,0,0,1,400,13,380,380,10,0,13,0,0,10,0,0,0,0,0,1,400,13,100,0,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,380,380,10,0,13,0,0,10,0,0,0,0,0,1,400,13,400,350,10,0,13,220,520,10,0,0,0,0,0,"
           ,["uk"] = "0,350,13,500,0,10,440,13,0,0,10,0,0,0,0,0,1,350,13,750,750,10,440,13,0,0,10,0,0,0,0,0,1,400,13,1000,200,10,450,13,400,2000,10,0,0,0,0,0,1,400,13,380,380,10,0,13,0,0,10,0,0,0,0,0,1,400,13,100,0,10,0,13,0,0,10,0,0,0,0,0,0,1500,13,10000,0,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,380,380,10,0,13,0,0,10,0,0,0,0,0,1,400,13,400,350,10,0,13,220,520,10,0,0,0,0,0,"
           ,["us"] = "0,350,13,500,0,10,440,13,0,0,10,0,0,0,0,0,1,620,13,1800,200,10,425,13,0,0,10,0,0,0,0,0,1,440,13,2000,4000,10,480,13,0,0,10,0,0,0,0,0,1,480,13,500,500,10,620,13,0,0,10,0,0,0,0,0,1,440,13,300,0,10,0,13,0,0,10,0,0,0,0,0,1,1440,13,100,100,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1400,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,480,13,500,500,10,620,13,0,0,10,0,0,0,0,0,1,480,13,250,250,10,620,13,0,0,10,0,0,0,0,0,"
           ,["ve"] = "0,425,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1800,200,10,0,13,0,0,10,0,0,0,0,0,1,425,13,1000,4000,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,400,13,300,0,10,450,13,0,0,10,0,0,0,0,0,1,1440,13,500,500,10,0,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,425,13,250,250,10,0,13,0,0,10,0,0,0,0,0,"
           ,["vn"] = "0,425,13,300,0,10,425,13,0,0,10,0,0,0,0,0,1,350,13,1000,100,10,440,13,0,0,10,0,0,0,0,0,1,440,13,1000,2000,10,480,13,0,0,10,0,0,0,0,0,1,450,13,200,200,10,450,13,200,200,10,0,0,0,0,0,1,350,13,250,250,10,440,13,250,0,10,0,0,0,0,0,0,1400,13,10000,0,10,1750,13,0,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,900,13,500,0,10,0,13,0,0,10,0,0,0,0,0,1,900,13,500,500,10,0,13,0,0,10,0,0,0,0,0,1,450,13,200,200,10,450,13,200,200,10,0,0,0,0,0,1,425,13,500,500,10,425,13,500,500,10,0,0,0,0,0,"
        },
        media5 = {
            en = "UK1",
            it = "Italy1",
            fr = "France1",
            de = "Germany1"
        },
        dinstar = {
            en = "UK",
            it = "Italy",
            fr = "France",
            de = "Germany"
        },
        dinstar2 = {
            ["custom"] = "255",-- custom
            ["au"]     = "13", -- AUSTRALIA
            ["at"]     = "1",  -- AUSTRIA
            ["be"]     = "2",  -- BELGIUM
            ["cn"]     = "14", -- CHINA
            ["dk"]     = "16", -- DENMARK
            ["nl"]     = "21", -- NETHERLANDS (DUTCH)
            ["fi"]     = "3",  -- FINLAND
            ["fr"]     = "4",  -- FRANCE
            ["de"]     = "5",  -- GERMANY
            ["gr"]     = "6",  -- GREECE
            ["hk"]     = "15", -- HONGKONG
            ["it"]     = "5",  -- ITALY original is 7 changed to GERMANY dialtone
            ["jp"]     = "8",  -- JAPAN
            ["no"]     = "9",  -- NORWAY
            ["pl"]     = "18", -- POLAND
            ["pt"]     = "19", -- PORTUGAL
            ["ru"]     = "17", -- RUSSIA
            ["es"]     = "10", -- SPAIN
            ["se"]     = "11", -- SWEDEN
            ["tr"]     = "20", -- TURKEY
            ["en"]     = "12", -- UK
            ["uk"]     = "12", -- UK
            ["ba"]     = "22", -- BOSNIA
            ["us"]     = "0",  -- USA
            ["en-us"]  = "0",  -- USA
            ["kp"]     = "23"  -- DPRK (North Korea)
        },
        dinstar3 = {
            ["custom"] = "255"-- custom
        },
        akuvox = {
            custom = "0", -- custom
                  -- "1", -- default
            cn     = "2", -- CHINA
            es     = "3", -- SPAIN
            lu     = "4", -- LUXEMBOURG
            se     = "5", -- SWEDEN
            tw     = "6", -- TAIWAN
            be     = "7", -- BELGIUM
            dk     = "8", -- DENMARK
            fi     = "9", -- FINLAND
            de     = "10",-- GERMANY
            nl     = "11",-- NETHERLANDS
            no     = "12",-- NORWAY
            pt     = "13",-- PROTUGAL
            it     = "1"  -- ITALY  "14" changed to "1" because of interruption dialtone
        },
        rtx = {
            de = "00", -- GERMANY
            lu = "01", -- LUXEMBOURG
            fr = "02", -- FRANCE
            it = "03", -- ITALY
            nl = "04", -- HOLLAND
            gr = "05", -- GREECE
            ie = "06", -- IRELAND
            se = "07", -- SWEDEN
            es = "08", -- SPAIN
            dk = "09", -- DENMARK
            fi = "0A", -- FINLAND
            tr = "0B", -- TURK
            ch = "0C", -- SWISS
            pt = "0D", -- PORTUGAL
            be = "0E", -- BELGIUM
            at = "0F", -- AUSTRIA
            uk = "10", -- ENGLAND
            no = "11", -- NORWAY
            us = "12", -- US
            au = "13", -- AUSTRALIA
            nz = "14", -- NEW ZEALAND
            ru = "2B", -- RUSSIA
            br = "2F"  -- BRAZIL
        }
    }

    if not dialtone_t[vendor] then
        return false
    end

    local tone = false

    local function __getIndicationsConfSettings(country)
        rv = nil
        local indications = require"inilazy".get("/rw2/etc/callweaver/indications.conf")
        if indications and indications[country] then
            rv = {}
            for toneSet, str in pairs(indications[country]) do
                str = string.gsub(str, "*%d+", "")
                rv[toneSet] = {}
                for freqAndDur in string.gmatch(str, "([^,]+)") do
                    local freq, dur = string.match(freqAndDur, "(.*)%/(.*)")
                    local f1, f2 = string.match(freqAndDur, "(%d+)%+?(%d*)")
                    f2    = "" ~= f2 and f2 or nil
                    rv[toneSet][#rv[toneSet]+1] = {freq1 = f1, freq2 = f2, duration = dur}
                end
            end
        end
        return rv
    end

    if country and "" ~= country then
        if dialtone_t[vendor][country] then -- set country from predefined tone template
            tone = dialtone_t[vendor][country]
        elseif "akuvox" == vendor or "welltech" == vendor or "dinstar2" == vendor then -- set "custom" country
            tone = __getIndicationsConfSettings(country)
            if tone and "table" == type(tone) then
                tone.country = dialtone_t[vendor]["custom"]
            end
        end
    end

    if "dinstar3" == vendor then
        tone = __getIndicationsConfSettings(country)
        if "table" == type(tone) then
            tone.country = dialtone_t[vendor]["custom"]
        else
            tone = {
                ["country"] = "255", -- set Italy in case if file /rw2/etc/callweaver/indications.conf corrupted
                ["busy"] = {{["freq1"] = "425",["duration"] = "500"},{["freq1"] = "0",["duration"] = "500"}},
                ["dial"] = {{["freq1"] = "425",}},
                ["ring"] = {{["freq1"] = "425",["duration"] = "1000"},{["freq1"] = "0",["duration"] = "4000"}},
                ["congestion"] = {{["freq1"] = "425",["duration"] = "200"},{["freq1"] = "0",["duration"] = "200"}}
                }
        end
    end

    local admin_lang = get_admin_lang()
    if not tone then
        tone = dialtone_t[vendor][admin_lang] or dialtone_t[vendor]["it"]
    end

    if "welltech04" == vendor and tone then
        local tones = {
                   [1]  = "dial"
                  ,[2]  = "stutterdial"
                  ,[3]  = "ringback"
                  ,[4]  = "busy"
                  ,[5]  = "callwaiting"
                  ,[6]  = "roh"
                  ,[8]  = "warning"
                  ,[9]  = "holding"
                  ,[10] = "disconnect1"
                  ,[11] = "disconnect2"
              }
        local t = split(tone, ",")
        tone = {}
        for k, toneType in pairs(tones) do
            tone[toneType] = {["signaltype"] = t[((k-1)*16 + 1)]}
            tone[toneType]["cell1"] = {
                ["hz"]        = t[((k-1)*16 + 2)],
                ["freqlevel"] = t[((k-1)*16 + 3)],
                ["freqon"]    = t[((k-1)*16 + 4)],
                ["freqoff"]   = t[((k-1)*16 + 5)],
                ["deviation"] = t[((k-1)*16 + 11)]
            }
            tone[toneType]["cell2"] = {
                ["hz"]        = t[((k-1)*16 + 7)],
                ["freqlevel"] = t[((k-1)*16 + 8)],
                ["freqon"]    = t[((k-1)*16 + 9)],
                ["freqoff"]   = t[((k-1)*16 + 10)]
            }
        end
    end

    return tone
end

-- Try find ucarp interface
function get_ifucarp (address)
	local network = get_network()
	local interface = network["wbr0"] and "wbr0" or "eth0"
	if not network[interface .. ":ucarp"] then
		return address, "localiface"
	end

	-- wan eth0
	local wanaddr   = string.match(network[interface]["address"],"([0-9.]*)/(%d+)")
	local ucarpaddr = string.match(network[interface .. ":ucarp"]["address"],"([0-9.]*)/(%d+)")
	if wanaddr == address then
		return ucarpaddr, "ucarpiface"
	else
		return address, "localiface"
	end
end

-- Get registrar
function get_registrar(arg, dst_ip)
	local err

	-- return address if env[HTTP_HOST] is present. Executed magnet. It mast always have correct ip, even NAT
	if arg and type(arg) == "table" and arg["HTTP_HOST"] then
		local ipaddr = arg["HTTP_HOST"]:match("([^:]*)")
		if ipaddr then
			return trim(ipaddr), "environment", err
		end
	end

	-- return address for wmsbind
	if type(arg) == "string" then
		local storage = name and nmap.registry or _G
		if not storage["pbx:externip"] then
			local registrar
			if dst_ip and not isLocalIP(dst_ip) and not is_local_route(dst_ip) then
				-- Get from kamailio "cache"
				local kam_externip = "/tmp/provisioning_externip"
				local fp = io.open(kam_externip)
				if fp then
					local content = fp:read("*a"); io.close(fp)
					if content and content ~= "" and content ~= "empty" then
						registrar = content:match("(%d+%.%d+.%d+%.%d+)")
						if registrar then
							storage["pbx:externip"] = trim(registrar)
							return storage["pbx:externip"], "kamcache", err
						else
							err = "File '/tmp/provisioning_externip' has invalid ip address"
						end
					end
				else
					err = "Can't read address, because '/tmp/provisioning_externip' not found"
				end
			end

			-- Check from ucarp address
			if not registrar then
				local case_ifucarp
				registrar, case_ifucarp = get_ifucarp(arg)
				storage["pbx:externip"] = trim(registrar)
				return storage["pbx:externip"], case_ifucarp, err
			end
		else
			return storage["pbx:externip"], "globalspace", err
		end
	end
	return false, "notfound", err
end

--- Resolve specified ip address or domain name via hosts file
-- @param addr	    An IP address or domain name for resolving
-- @param source    A table with hosts files where to search
-- @param cfg       A table with flags for search
-- @return table value with resolved domain or address
-- @note Exception is internal Wildix domain "auto.wildixin.com"
function resolve_by_hosts(addr, source, cfg)
	if not addr then
		return false
	end
    if not source or "table" ~= type(source) or #source < 1 then
        source = {"/rw2/etc/dnsmasq.d/hosts-wms.conf", "/rw2/etc/hosts"}
    end

    local function __getHosts(source)
        local rt      = nil
        local err_msg = nil
        local n       = 0
        for _, file_path in pairs(source) do
            local file, err = io.open(file_path)
            if file then
                local cont = file:read("*all")
                file:close()
                if cont then
                    for line in cont:gmatch("[^\n]*") do
                        local t = require"core.pbxutils".split(line, "%s+")
                        if t and t[1] and "" ~= t[1] then
                            n = n + 1 -- records count
                            if not rt then rt = {} end
                            rt[n]    = t     -- add enumerated element
                            if not rt[t[1]] then rt[t[1]] = {} end    -- add subtable by ip address
                            for k, v in pairs(t) do
                                if k ~= 1 then
                                    rt[t[1]][#rt[t[1]] + 1] = v -- add record by ip addres
                                    rt[v] = t[1]                -- add record by hostname
                                end
                            end
                        end
                    end
                else
                    err_msg = "__getHosts: content is empty"
                end
            elseif err_msg then
                err_msg = err_msg .. ". " .. string.format("Couldn't read file: %s", err)
            else
                err_msg = string.format("Couldn't read file: %s", err)
            end
        end
        return rt, err_msg
    end

    local function __getIpByName(name, hosts)
        -- let's resolve ip by domain name
        if not name or not hosts or "table" ~= type(hosts) then
            return nil
        end
        local ip = hosts[name]
        return ip
    end

    local function __getNameByIp(addr, hosts)
        -- let's resolve name by ip address
        if not addr or not hosts or "table" ~= type(hosts) then
            return nil
        end
        local fix_addr = addr:gsub("%.", "%%%.")
        local domain = nil
        if hosts[addr] then
           for _, v in pairs(hosts[addr]) do
               if v ~= addr then
                   domain = v
                   break
               end
           end
        end
        return domain
    end

    local rt = nil
    local hosts, err_msg = __getHosts(source)
    local isIp = require"core.pbxutils".get_ipv4(addr)
    repeat
        if cfg then
            if cfg.FailoverByDns then
                rt = {
                     domain = not isIp and addr or __getNameByIp(addr, hosts)
                    ,ip     = isIp and addr or __getIpByName(addr, hosts)
                }
                break
            end

            if cfg.ForceResolveName then
                rt = {
                     domain = not isIp and addr or __getNameByIp(addr, hosts)
                    ,ip     = isIp and addr or __getIpByName(addr, hosts)
                }
                break
            end

            if cfg.ForceResolveIp then
                rt = {ip = isIp and addr or __getIpByName(addr, hosts)}
                break
            end
        end

        rt = {ip = isIp and addr or __getIpByName(addr, hosts)}
    until true

    return rt, err_msg
end

--- Checking address for wideband codec supporting
-- @param string    The specified address for checking
-- @return boolen value as result
function is_wideband(checkAddr)
    local rv = false
    if not checkAddr or "" == checkAddr then
        return rv
    end

    -- initialize storage
    local storage = nmap ~= nil and nmap.registry or _G

    -- clear up cache storage
	local recache_indc = "/dev/shm/.magnet_up_wideband"
	if is_exists(recache_indc) then
		storage["pbx:wideband"] = nil
		os.remove(recache_indc)
	end

    -- set up cache storage by wideband data
	if not storage["pbx:wideband"] then
		local filename = "/etc/callweaver/sip-wideband.conf"
		if not is_exists(filename) then
			return rv
		end

		local fp       = io.open(filename)
        if not fp then
            return rv
        end
        local contents          = fp:read("*a"); fp:close()
        storage["pbx:wideband"] = {}

        -- processing configuration contents
        local searchIfaces = {}
        repeat
            if contents:find("forcewideband%s*=%s*yes%s*") then
                -- enable wideband codecs for all networks (0.0.0.0/0)
                searchIfaces[1] = "0.0.0.0/0"
                break
            end

            if not contents:find("disableautowideband%s*=%s*yes%s*") then
                -- retrieve wideband network from list of the local interfaces
                for _, info in pairs(get_network()) do
                    if info and info.address then
                        table.insert(searchIfaces, info.address)
                    end
                end
                break
            end

            -- read wideband network from the list
            for data in contents:gmatch("wideband%s*=%s*([0-9./]*)%s*") do
                if data and "" ~= data then
                    table.insert(searchIfaces, data)
                end
            end
        until true

        -- push found data to local storage
        if searchIfaces[1] then
            for _, row in ipairs(searchIfaces) do
                local addr, bitmask = row:match("([0-9.]*)/(%d+)")
                if addr == nil then
                    addr, bitmask = row:match("([0-9.]*)"), 32
                end
                local net = __ip_calculator(addr, bitmask)
                if net and "" ~= net then
                    table.insert(storage["pbx:wideband"], {net = net, bitmask = bitmask})
                end
            end
        end
    end

    -- check input address with wideband networks
	for _, data in ipairs(storage["pbx:wideband"]) do
		local cnet     = data.net
        local cbitmask = data.bitmask
		local vnet     = __ip_calculator(checkAddr, cbitmask)
        if cnet == vnet then
            rv = true
            break
        end
    end

	return rv
end

-- Get DNS settings from DHCP configuration
function get_dns_servers ()
	-- Get from dhcp settings
	local t = {}
	local output = trim(pbxutils.run[[ sed -ne 's/.*\:dns-server,\(.*\)/\1/p' /etc/dnsmasq.d/dhcp.conf 2>/dev/null |tr '\n' ',' ]])
	if output ~= "" then
		t = split(output, ",")
	end
	return unpack(t)
end

--- Get primary, secondary DNS settings
-- @param prim_dns a string with ip address of primary dns server
-- @param sec_dns a string with ip address of secondary dns server
-- @return dns_server and second_dns_server or "", if param prim_dns or sec_dns is null or ""
function get_prim_sec_dns_servers (prim_dns, sec_dns)
	local dns_server	    = nil
	local second_dns_server = nil
	if sec_dns and "" ~= sec_dns then
    	if prim_dns and "" ~= prim_dns then
    		dns_server 		  = prim_dns
    		second_dns_server = sec_dns
    	elseif not prim_dns or "" == prim_dns then
    		dns_server = sec_dns
    	end
    elseif not sec_dns or "" == sec_dns then
    	if prim_dns and "" ~= prim_dns then
    		dns_server 		  = prim_dns
    	end
    end
    return dns_server or "", second_dns_server or ""
end

-- Get NTP settings from DHCP configuration
function get_ntp_servers ()
	-- Get from dhcp settings
	local t = {}
	local output = trim(pbxutils.run[[ sed -ne 's/.*\:ntp-server,\(.*\)/\1/p' -e 's/dhcp-option=42,\(.*\)/\1/p' /etc/dnsmasq.d/dhcp.conf 2>/dev/null |tr '\n' ',' ]])
	if output ~= "" then
		t = split(output, ",")
	end
	return unpack(t)
end

-- Check format for detect new token
function is_new_token (token)
	if not token or token == "" then
		return "null"
	elseif token == "reset" then
		return token
	elseif token:len() == 15 and token:sub(13,13) == "W" and token:sub(10,10) == "l" and token:sub(7,7) == "d" and token:sub(3,3) == "x" then
		return "new"
	else
		return "old"
	end
end

--- Retrieve a custom password for devices
-- @note currently return a constant passowrd.
--       Main using was in WMS2.0!
function get_gpassword ()
	local storage = nmap ~= nil and nmap.registry or {}
	if not storage["gpassword"] then
		storage["gpassword"] = "wildix"
	end
	return storage["gpassword"]
end

--- Get files from directory by extension
-- @return table
-- TODO: can be replaced by readdir in fsutils.lua
function readDir(dir, ext)
	if ext == nil then
		ext = "*"
	end
	local result = {}
	for i in io.popen("ls " .. dir .. " 2>/dev/null"):lines() do
		if string.find(i, "%." .. ext .. "$") then
			table.insert(result, i)
		end
	end
	return result
end

-- get inter digit timeout from features.conf, ms
function get_interdigit_timeout()
	local filename = "/etc/callweaver/features.conf"
	local timeout = "5"

	if is_exists (filename) then
		local fp = io.open (filename, "r")
		for line in fp:lines() do
			if string.find (line,"^transferinterdigittimeout%s=>%s(%d+)") then
				timeout = string.match (line,"^transferinterdigittimeout%s=>%s(%d+)")
				break
			end
		end

		io.close (fp)
	end

	return timeout
end

-- get jitter buffers
function get_jitter_buffer (remoteDevice)
	if remoteDevice then
		return {min = "150",nom="300",max="500"} end
	local core = require"common.core"
	local connection = core.connection
	local jbsettings = {
		min = "50",
		nom = "100",
		max = "200"
	}

	connection:switch ("devdb")

	local sql = "select variable, value from system where variable like 'JITTER_%%'"
	for key, value in connection:rows (sql) do
		local pkey = string.sub (key, 8)
		jbsettings[pkey:lower()] = value
	end
	connection:close()
	return jbsettings
end

-- get exten number
function get_extenNumber_str ()
	local extenNumber = "telephoneNumber"
	local filename = "/etc/callweaver/extensions-variables.conf"

	if is_exists (filename) then
		local fp = io.open (filename, "r")
		for line in fp:lines() do
			if string.find (line,"^LDAP_EXTEN_NUMBER") then
				extenNumber = string.match (line,"^LDAP_EXTEN_NUMBER=(.*)")
				break
			end
		end

		io.close (fp)
	end

	return extenNumber
end

-- Get end of digit
function get_eod()
	local eod = "2"
	local filename = "/etc/callweaver/extensions-variables.conf"

	if is_exists(filename) then
		local file = io.open(filename, "r")
		local contents = file:read("*a")
		file:close()

		eod = string.match (contents, "GEOD=([^\n]+)") or "2"
	end

	return eod
end

-- Get button sound
function get_button_sound()
	local button_sound = "yes"
	local filename = "/etc/callweaver/extensions-variables.conf"

	if is_exists(filename) then
		local file = io.open(filename, "r")
		local contents = file:read("*a")
		file:close()

		button_sound = string.match(contents, "GBUTTON_SOUND=([^\n]+)") or "yes"
	end

	return button_sound == "yes" and "1" or "0"
end

--- Get Sssion Timer parameters from sip.conf
-- session-timers=originate
-- session-expires=90
-- session-minse=90
-- session-refresher=uac
-- @return a table with parsed parameters
function getSessionTimer ()
	local filename = "/etc/callweaver/sip.conf"

	if is_exists (filename) then
		local file = io.open(filename, "r")
		local contents = file:read("*a")
		file:close()

		return {
			["timers"]    = string.match (contents,"session%-timers=([^\n]+)"),
			["expires"]   = string.match (contents,"session%-expires=([^\n][0-9]+)"),
			["minse"]     = string.match (contents,"session%-minse=([^\n][0-9]+)"),
			["refresher"] = string.match (contents,"session%-refresher=([^\n]+)")
		}
	end

	return false
end

-- Get ToS parameters, flach for TLS/SRTP from sip-wms.conf
-- @return a table with parsed parameters
function get_sip_settings()
	local filename = "/etc/callweaver/sip-wms.conf"

	if is_exists (filename) then
		local file = io.open(filename, "r")
		local contents = file:read("*a")
		file:close()

		return {
			["tos"]     = tonumber (string.match (contents, "tos=([^\n]+)")) or 184,
			["tos_sip"] = tonumber (string.match (contents, "tos_sip=([^\n]+)")) or 184
		}, string.match (contents, "tonezone=([^\n]+)") or false
	end

	return {
		["tos"] = 184,
		["tos_sip"] = 184
	}, false
end

-- Parse ini file without sections
function parse_simple_ini(filename)
	local f = io.open(filename)
	if f then
		local config = {}
		for line in f:lines() do
			local k, v = string.match(line, '(%S+)%s*=%s*"?([^%s"]*)"?')
			if k then
				config[k] = v
			end
		end
		f:close()
		return config
	end
end

-- Create proxy url for http client if proxy enabled
function get_proxy_url()
	local filename = "/rw2/etc/pbx/httpproxy.ini"
	local config = parse_simple_ini(filename)
	if config and "1" == config.enable then
		local cred = "none" ~= config.auth_method and config.login .. ":" .. config.password .. "@" or ""
		return "http://" .. cred .. config.host .. ":" .. config.port
	end
end

-- Check of custom remote firmware repository
function isCustomFirmwareServer()
	local filename = "/var/www/upgrade/upgrade.ini"
	local conf     = require"inilazy".get(filename)
	return conf and "table" == type(conf["customFirmwareServer"]) and "yes" == conf["customFirmwareServer"]["enabled"]
end

--- Function for getting device's firmware host for upgrade
function getFirmwareUpgradeHost()
	local url = "http://firmwares.wildix.com"
	local isCustomFirmware = false
	local customFirmware = require"inilazy".get("/rw2/etc/pbx/upgrade.ini")
	if customFirmware and "table" == type(customFirmware["customFirmwareServer"]) and "string" == type((customFirmware["customFirmwareServer"]["enabled"])) then
		isCustomFirmware = "yes" == (customFirmware["customFirmwareServer"]["enabled"]):match("%w+")
	end	
	if isCustomFirmware and "string" == type(customFirmware["customFirmwareServer"]["url"]) then
		url = (customFirmware["customFirmwareServer"]["url"]):gsub('"', ""):gsub('/$', "")
	end	
	return url
end

--- Function for parsing firmware version of the devices
-- @param version	A firmware formatted version, like 7.61.10.35
-- @return table with parsed firwaver version values otherwise FALSE value
function parse_firmware(version, type)
	if not version then
		return false
	end

	local rv = false

	if "sht_v1" == type then
		local model_spec, major_version = string.match(version, "(%d+)%.(%d+)")
		if model_spec and major_version then
			rv = tonumber(model_spec .. major_version)
		end
	elseif "v_sfu" == type then  -- v_sfu = version single file upgrade (WMS-6966)
		local major_version, build_version, minor_version = string.match(version, "(%d+)%.(%d+)%.(%d+)")
		if major_version and build_version and minor_version then
			rv = {
				["sht_v2"]  = tonumber(major_version .. minor_version),
			}
		end	
	else
		local model_spec, major_version, build_version, minor_version = string.match(version, "(%d+)%.(%d+)%.(%d+)%.(%d+)")
		if model_spec and major_version and build_version and minor_version then
			rv = {
				["model"]  = tonumber(model_spec),
				["major"]  = tonumber(major_version),
				["build"]  = tonumber(build_version),
				["minor"]  = tonumber(minor_version),
				["sht_v1"] = tonumber(model_spec .. major_version),
				["sht_v2"] = tonumber(build_version .. minor_version)
			}
		end
	end

	return rv
end

--- Retrieve information about remote firmwares
-- @param family	A device family name
-- @param model		A specific device
-- @return info of the last firmwares for specified device otherwise FALSE with error message
-- @note can used for tests manually creted file (with suffix .beta)
-- @note see partially analog of this function in devices/index.php 'check_version'
function getRemoteFirmwaresInfo(family, model)
	local function __loadInfoFile(filename)
		local rv = false
		if not filename or "" == filename then
			return rv
		end

		local hc         = require"core.httpClient":new()
		local reqURL     = "https://firmwares.wildix.com/remoteupgrade/fwversions-5.0.json"
		local proxy_url = nil
		if isCustomFirmwareServer() then
			-- proxy via local nginx
			reqURL     = "http://127.0.0.1/remoteupgrade/fwversions-5.0.json"
		else
			proxy_url = get_proxy_url()
		end
		local rData, err = hc:get(reqURL, {timeout = 10, proxy = proxy_url})
		if rData and 401 ~= rData.code and 403 ~= rData.code and rData.body then
		    local fp = io.open(filename, "w")
		    if fp then
		       fp:write(rData.body)
		       fp:close()
		       rv = true
		    end
		end 

		return rv
	end
	
	local rv              = { tag = "(remote source)"}
	local err             = nil
	local supportedFamily = {
		 WP  = true		-- Wildix phones revision 1 (pre 2015)
		,MGW = true		-- Media 5 gateways
	}
	
	repeat
	    -- processing device model
	    if not model then
	       err = "not specified device model"
	       break
	    end
	    
    	-- processing family data
    	family = family and family:upper()
    	if not supportedFamily[family] then
    	    err = "the device family does not found"
    	    break
    	end
    	
    	-- updating firmware file data
    	local infoFilePath  = "/dev/shm/.fwversions.json"
    	local requireUpdate = false
    	if is_exists(infoFilePath .. ".beta") then
    	    rv.tag   = "(beta file)"
    	    infoFilePath = infoFilePath .. ".beta"
    	elseif is_exists(infoFilePath) then
    	   local actualTime = 86400                                    -- 24 hours
    	   local infoAttrs  = require"lfs".attributes(infoFilePath)
    	   local modTime    = infoAttrs and infoAttrs.change or 0
    	   if actualTime < (os.time() - modTime) then
    	       requireUpdate = true
	       end
       else
           requireUpdate = true
       end
       
       if requireUpdate and not __loadInfoFile(infoFilePath) then
           err = "could not update file: " .. infoFilePath
           break
       end
        
        -- parsing firmware file
        local fp, errMsg = io.open(infoFilePath)
        if not fp then
            err = errMsg
            break
        end
        local infoData = fp:read("*a")
        fp:close()
        infoData = require"dkjson".decode(infoData)
        if not infoData then
            err = "decoding error for file: " .. infoFilePath
            break
        end
        
        -- processing firmware data file for specified family and model        
        if not infoData[family] then
            err = "firmware data was not found for family: " .. family
            break
        end
        infoData = infoData[family]
        if not infoData[model] then
            err = "firmware data was not found for model: " .. model
            break
        end
        infoData = infoData[model]
        
        if not infoData.path or not infoData.version then
            err = string.format("invalid firmware data for %s/%s", family, model)
            break
        end

        rv.data = infoData
	until true
	
    return rv, err
end

--- SIP ping. Send SIP OPTION method to determinate sip device is reachable
-- @param bindhost A string with local ip addr optionaly with port, i.e "192.168.1.2:5067" (by default 5065-5070)
-- @param desthost A string with target ip addr optionaly with port, i.e "192.168.1.3:5060" (by default 5060)
-- @return A string with response and error messoge if present
function sip_ping (bindhost, desthost)
	local socket = socket or require"socket"
	local math = math
	local s = "OPTIONS sip:$dhost:$dport SIP/2.0\r\nVia: SIP/2.0/UDP $shost:$sport;branch=0\r\nFrom: sip:pinger@kamailio.org\r\nTo: sip:$dhost:$dport\r\nCall-ID: $callid@$shost\r\nCSeq: 1 OPTIONS\r\nContent-Length: 0\r\n\r\n"

	local shost, sport = unpack (split (bindhost, ":"))
	if not sport then
		sport = tostring (math.random (5065, 5070))
	end

	local dhost, dport = unpack (split (desthost, ":"))
	if not dport then
		dport = "5060"
	end

	return socket.protect (function()
		local co = socket.udp()
		local try = socket.newtry (function() co:close() end)

		co:setsockname (shost, sport)
		co:settimeout (1)

		local data = string.gsub (s, "%$(%w+)", {
			shost = shost,
			sport = sport,
			dhost = dhost,
			dport = dport,
			callid = string.random (8, "%l%d")
		})

		try (co:sendto (data, dhost, dport))
		local answer = try(co:receive())
		co:close()
		return answer
	end)()
end

local function charsets()
	local chars = {}
	for loop = 0, 255 do
		chars[loop+1] = string.char (loop)
	end
	local str = table.concat(chars)

	return {['.'] = chars}, str
end

local function addlookup (charset)
	local built, str = charsets()
	local substitute = string.gsub (str, '[^'..charset..']', '')

	local lookup = {}
	for loop = 1, string.len (substitute) do
		lookup[loop] = string.sub (substitute, loop, loop)
	end

	built[charset] = lookup

	return lookup
end

-- Generate random string
-- length (number)
-- charset (string, optional); e.g. %l%d for lower case letters and digits
function string.random (length, charset)
	local charset = charset or '.'

	if charset == '' then
		return ''
	else
		local chrs = charsets()
		local result = {}
		local lookup = chrs[charset] or addlookup(charset)
		local range = #lookup

		for loop = 1,length do
			result[loop] = lookup[math.random(1, range)]
		end

		return table.concat(result)
	end
end

------------------------------------------------------------------------------------------

function ip_to_bin ( ip )
	local bit = bit
	local err

	-- ip, err = expand_ip( ip )
	if err then return nil, err end

	local t, mask = {}

	if not ip:match( ":" ) then
		-- ipv4 string
		for octet in string.gmatch( ip, "%d+" ) do
			t[#t+1] = bit.tohex( tonumber(octet) )
		end
		mask = "00"
	else
		-- ipv6 string
		for hdt in string.gmatch( ip, "%x+" ) do
			t[#t+1] = hdt
		end
		mask = "0000"
	end

	-- padding
	for i, v in ipairs( t ) do
		t[i] = mask:sub( 1, string.len( mask ) - string.len( v ) ) .. v
	end

	return hex_to_bin( table.concat( t ) )
end

function bin_to_ip ( binstring )
	local bit = bit
	if type( binstring ) ~= "string" or binstring:match( "[^01]+" ) then
		return nil, "Error in ipOps.bin_to_ip: Expected string of binary digits."
	end

	local af
	if string.len( binstring ) == 32 then
		af = 4
	elseif string.len( binstring ) == 128 then
		af = 6
	else
		return nil, "Error in ipOps.bin_to_ip: Expected exactly 32 or 128 binary digits."
	end

	local t = {}
	if af == 6 then
		local pattern = string.rep( "[01]", 16 )
		for chunk in string.gmatch( binstring, pattern ) do
			t[#t+1] = bit.tohex( tonumber( chunk, 2 ) )
		end
		return table.concat( t, ":" )
	end

	if af == 4 then
		local pattern = string.rep( "[01]", 8 )
		for chunk in string.gmatch( binstring, pattern ) do
			t[#t+1] = tonumber( chunk, 2 ) .. ""
		end
		return table.concat( t, "." )
	end
end

function hex_to_bin ( hex )
	local bit = bit
	if type( hex ) ~= "string" or hex == "" or hex:match( "[^%x]+" ) then
		return nil, "Error in ipOps.hex_to_bin: Expected string representing a hexadecimal number."
	end

	local t, mask = {}, "0000"
	local binchar
	for hexchar in string.gmatch( hex, "%x" ) do
		binchar = bit.tobit( tonumber( hexchar, 16 ) )
		t[#t+1] = mask:sub( 1, string.len( mask ) - string.len( binchar ) ) .. binchar
	end
	return table.concat( t )
end

--- Retreive from table specified index range values
-- @param values	An original table with data
-- @param i1		A beginner index
-- @param i2		A ended index
-- @return table contains only specified values range from original
-- @note An ended intex can be lower than zero
function table_slice (torig, i1, i2)
	local tres = {}
	local n   = #torig

	-- default values for range
	i1 = tonumber(i1) or 1
	i2 = tonumber(i2) or n

	if i2 < 0 then
		i2 = n + i2 + 1
	elseif i2 > n then
		i2 = n
	end

	if i1 < 1 or i1 > n then
		return tres
	end

	local k = 1
	for i = i1,i2 do
		tres[k] = torig[i]
		k = k + 1
	end
	return tres
end

function create_sip_extra()
    local phpCode = [[ php -r '
        require_once("/var/www/init.php");
        ini_set("include_path", BASEPATH . ":" . ini_get("include_path"));
        $tr = new trunk();
        $tr->set_cw_file();
        '
    ]]
	os.execute(phpCode)
end

--- Retreive provision configuration form file to internal table
-- @return parsed and deserialized configuration, else error message
-- @note Set default table in parent function !!!
-- @note Need path to pbxutils library
function get_provConfig()

	--- Check validation of the parameters
	-- SIPServerPort		(value: number)				- the port number on SIP server
	-- SIPTransport			(value: UDP,TCP,TLS)		- the transport used for SIP packages
	-- SIPServerPortRemote	(value: number)				- the port number on SIP server at device behind NAT
	-- SIPTransportRemote	(value: UDP,TCP,TLS)		- the transport used for SIP packages at device behind NAT
	-- FailoverPbxIp        (value: ip address)         - an ip addres for force set Sip Server 2 address (failover)
	-- HTTPForceUse			(value: yes, no)			- disable provision request via HTTPS if device has open secure port
	-- BLFTransType			(value: attended, blind) 	- type of BLF transfer
	-- BLFNotifyType        (value: ActiveCall, ActiveAndIncomingCall) - [WPR2]: BLF blinking type
	-- MelodyBLF			(value: yes, no)			- use melody BLF audio notification
	-- PowerLEDBlinkTime    (value: 0-14400)            - BLF blinking duration for new event (in seconds)
	-- TrustLocalDevice		(value: yes, no)			- automatic device adding to provision database
	-- EnableMWISubscribe	(value: yes, no)			- enable SUBSCRIBE for MWI(Message Waiting Indicator) on W24FXSR2
	-- DisableShowUsername	(value: yes, no)			- if "yes" show only internal number (old behavior)
	-- BaseMaxSIPReg		(value: 4-30)				- W-AIR number of SIP accounts before distributed load
	-- SRTPForceUse			(value: yes, no)			- enable or disable SRTP on devices that supports SRTP mode
	-- AuthLockTimeout		(value: number > 0)			- time in seconds for XML browser auth caching
	local paramsCallback = {
		 ["SIPServerPort"]		= function(value)
		 	if value:match("%d") then
				local num_value = tonumber(value)
				if num_value > 0 and num_value < 65535 then
					return num_value
				end
			end
		end

		,["SIPTransport"]		= function(value)
			local norm_value = value:upper()
			if "UDP" == norm_value or "TCP" == norm_value or "TLS" == norm_value then
				return norm_value
			end
		end

		,["FailoverPbxIp"]         = function(value)
			if require"core.pbxutils".get_ipv4(value) then
				return value
			end
		end

		,["HTTPForceUse"]		= function(value)
		 	local norm_value = value:lower()
		 	return "yes" == norm_value and true
		end
    
    ,["DisableShowUsername"]		= function(value)
		 	local norm_value = value:lower()
		 	return "yes" == norm_value and true
		end

		,["BLFTransType"]		= function(value)
		  	local norm_value = value:lower()
		  	return "blind" == norm_value and "2" or "1"	-- default value "attended" blf transfer
		end
		
		,["BLFNotifyType"]      = function(value)
		    value = value and value:lower()
		    return "activecall" == value and "0" or "1"
		end
		
		,["MelodyBLF"]		= function(value)
		 	local norm_value = value:lower()
		 	return "yes" == norm_value and true
		end

		,["PowerLEDBlinkTime"]  = function(value)
		    local num_value = tonumber(value)
            if num_value and num_value >= 0 and num_value <= 14400 then
                return num_value
            end
            return 0
		end

		,["POWER_LED"]		    = function(value)
		 	local norm_value = value:lower()
		 	if "yes" == norm_value or "no" == norm_value then
		 		return norm_value
		 	end
		end
        ,["FailoverByDns"]     = function(value)
            local norm_value = value:lower()
            return "yes" == norm_value and true
        end
        ,["ForceResolveName"]     = function(value)
            local norm_value = value:lower()
            return "yes" == norm_value and true
        end
        ,["ForceResolveIp"]     = function(value)
            local norm_value = value:lower()
            return "yes" == norm_value and true
        end
        ,["HIDE_DTMF"]     = function(value)
            local norm_value = value:lower()
            return "yes" == norm_value and true
        end
        ,["g711only"]     = function(value)
            local norm_value = value:lower()
            return "yes" == norm_value and true
        end
        ,["DisableCodecG729"]     = function(value)
            local norm_value = value:lower()
            return "yes" == norm_value and true
        end
        ,["BaseMaxSIPReg"]  = function(value)
            local num_value = tonumber(value)
            if num_value and num_value >= 4 and num_value <= 30 then
                return num_value
            end
        end
        ,["EnableMWISubscribe"]     = function(value)
            local norm_value = value:lower()
            return "yes" == norm_value and true
        end
        ,["ResetHandset"] = function(value)
            local norm_value = value:lower()
            return "yes" == norm_value and true
        end
        ,["LossPktRateMin"] = function(value)
            local num_value = tonumber(value)
            if num_value and num_value >=0 and num_value <= 100 then
                return num_value
            end
        end
        ,["OffHookDialOut"]     = function(value)
            local norm_value = value:lower()
            if "yes" == norm_value or "no" == norm_value then
                return norm_value
            end
        end
        ,["SidetoneHandset"]  = function(value)
            local num_value = tonumber(value)
            if num_value and num_value >= 0 and num_value <= 7 then
                return num_value
            end
        end
        ,["SidetoneHeadset"]  = function(value)
            local num_value = tonumber(value)
            if num_value and num_value >= 0 and num_value <= 7 then
                return num_value
            end
        end
        ,["EnableReasonCause"] = function(value)
            local norm_value = value:lower()
            if "yes" == norm_value or "no" == norm_value then
                return norm_value
            end
        end
        ,["w-air-sb-pin"] = function(value)
            if tonumber(value) and 4 == #value then
                return value
            end
        end
        ,["DelayDevicesUpgrade"]     = function(value)
            local norm_value = value:lower()
            if "yes" == norm_value or "no" == norm_value then
                return norm_value
            end
        end
        ,["SRTPForceUse"] = function(value)
            local norm_value = value:lower()
            if "yes" == norm_value or "no" == norm_value then
                return norm_value
            end
        end
        ,["AuthLockTimeout"]  = function(value)
            local num_value = tonumber(value)
            if num_value and num_value >= 0 then
                return num_value
            end
        end
        ,["LLDP"] = function(value)
            local norm_value = value:lower()
            if "yes" == norm_value or "no" == norm_value then
                return norm_value
            end
        end
        ,["LLDPInterval"] = function(value)
            local num_value = tonumber(value)
            if num_value and num_value >=10 and num_value <= 3600 then
                return num_value
            end
        end
        ,["DelayAfterOffhook"] = function(value)
            local num_value = tonumber(value)
            if num_value and num_value >=1 and num_value <= 255 then
                return num_value
            end
        end
        ,["SpeakerDisable"]     = function(value)
            local norm_value = value:lower()
            -- values can be yes/no or comma separeted list of extensions
            repeat
                if "yes" == norm_value or "no" == norm_value then
                    break
                end
                if norm_value then
                    local tbl = {}
                    local RawList = require"core.pbxutils".split(norm_value .. ",", ",")
                    for _, val in ipairs(RawList) do
                        val = tonumber(val)
                        if val then
                            table.insert(tbl, tostring(val))
                        end
                    end
                    if #tbl >= 1 then
                        norm_value = tbl
                    else
                        norm_value = "no"
                    end
                end
            until true
            return norm_value
        end
	}

	-- extend previous table by parameters with the same callbacks
	paramsCallback["SIPServerPortRemote"] 	= paramsCallback["SIPServerPort"]
	paramsCallback["SIPTransportRemote"] 	= paramsCallback["SIPTransport"]
	paramsCallback["TrustLocalDevice"]		= paramsCallback["HTTPForceUse"]
    paramsCallback["LossPktRateMax"]        = paramsCallback["LossPktRateMin"]

	-- default values for provision parameters. All other by default is "false"
	local DefaultParamValues = {
		TrustLocalDevice = false
	}

	-- parameter parser
	local function parse_param(key, value)
		if not value then
			return false
		end

		if paramsCallback[key] then
			return paramsCallback[key](value)
		end
		return "error"
	end

	-- loading ini file
	local ini_parser 	   = require"inilazy"
	local conf_file  	   = "/rw2/etc/provision.conf"
	local conf, err_parser = ini_parser.get(conf_file)
	if not conf then
		return {}, string.format("GConfig: %s", err_parser or "file not found or file is empty")
	end
	if not conf.general then
		return {}, "GConfig: not find 'general' section"
	end

	-- parsing general configuration
	local errors	= {}
	local param_val = "error"
	local tgeneral  = {}
	for key, value in pairs(conf.general) do
		if "RawScan" == key then
			tgeneral[key] = nil
			-- skip this options it is used in the parent PHP and BASH scripts (nmap executing)
		else
			param_val = parse_param(key, value)
    		if "error" ~= param_val then
    			tgeneral[key] = param_val
    		else
    			tgeneral[key] = nil
    			table.insert(errors, string.format("GConfig: wrong parameter at 'general' section %q: %s", key or "", value or ""))
    		end
    	end
	end
	-- set default parameters for not found keys in config
	for pName, pDefValue in pairs(DefaultParamValues) do
		if not conf.general[pName] then
			tgeneral[pName] = pDefValue
		end
	end

	-- parsing configuration for all found modules and merging with general
	local tconf = {["general"] = tgeneral}
	for mname in pairs(conf) do
		if "general" ~= mname then
			local mconf = require"core.pbxutils".tcopy(tgeneral)
			for key, value in pairs(conf[mname]) do
				param_val = parse_param(key, value)
    			if "error" ~= param_val then
    				mconf[key] = param_val
    			else
    				mconf[key] = nil
    				table.insert(errors, string.format("GConfig: wrong parameter at '%s' section %q: %s", mname, key or "", value or ""))
    			end
    		end
    		-- set default parameters for not found keys in config
        	for pName, pDefValue in pairs(DefaultParamValues) do
        		if not conf.general[pName] and not conf[mname][pName] then
        			mconf[pName] = pDefValue
        		end
        	end
    		tconf[mname] = mconf
    	end
    end

	return tconf, #errors ~= 0 and errors
end

--- Get provision alternative secure port from file
-- @return port number, or nill
function get_alternate_port()
	local filename	= "/etc/wildix/tproxy_wms_include.conf"
	local alternate_port
	if is_exists(filename) then
		local fp = io.open (filename, "r")
		for line in fp:lines() do
			if line and "" ~= line then
				alternate_port = line:match("port=(%d+);")
				if alternate_port and "" ~= alternate_port then
					break
				end
			end
		end
		io.close (fp)
	end
	return alternate_port or false
end

--- Сonvert ip to the long number
-- @param ip string with the ipv4 address
-- @return long
local function __ip2long(ip)
	-- Converts ipv4 to long
	-- http://forums.verlihub-project.org/viewtopic.php?f=15&t=252
	local _,_,a,b,c,d = string.find(ip, "(%d+).(%d+).(%d+).(%d+)")
	if (tonumber(a) and tonumber(b) and tonumber(c) and tonumber(d)) then
	        return a*16777216 + b*65536 + c*256 + d
	else
	        return 0
	end
end

--- check ip address belonging to local networks
-- @param ip string with the ipv4 address
-- @return boolean
function isLocalIP(ip)
	if "string" ~= type(ip) then
		return false
	end
	local ipLong = __ip2long(ip);
	if (ipLong > __ip2long("10.0.0.0") and ipLong < __ip2long("10.255.255.255")) or
		(ipLong > __ip2long("172.16.0.0") and ipLong < __ip2long("172.31.255.255")) or
		(ipLong > __ip2long("192.168.0.0") and ipLong < __ip2long("192.168.255.255")) then
		return true;
	end
	return false;
end

--- Build XML message for dialnow request
-- @param agent		A table with general device information
-- @param request	A table with requested parameters
-- @return xml content with dialnow settings
function dialnow_settings (agent, request)
	if not agent or not request then
		return false
	end
	local core = require"common.core"
	local connection = core.connection
	connection:switch ("devdb")
	local sql = "SELECT value FROM system WHERE variable=%q"
	local cursor = connection:execute (sql, "QUICK_DIAL_PATTERNS")
	local dpatterns = cursor:fetch(); cursor:close()
	connection:close()
	local content = {}
	content[#content + 1] = "<?xml version=\"1.0\" encoding=\"utf-8\" ?>"
	content[#content + 1] = "<DialNow>"
	if dpatterns and "" ~= dpatterns then
		local i = 1
		for patt in dpatterns:gmatch("([^%|]*)|?") do
			if "" ~= patt then
				content[#content+1] =  " <Data DialNowRule=\"" .. patt .. "\" LineID=\"\" />"; i = i + 1
			end
		end
	end
	content[#content + 1] = "</DialNow>"
	return content
end

--- Get sip encryption key for welltech 01/02 fxs devices
-- @return port number, or nill
function getWelltechEncryptionKey()
	local rv = nil
	local encryptKeyConf = "/rw2/etc/kamailio/welltech_secret.cfg"
	local fp = io.open(encryptKeyConf)
	if fp then
		local content = fp:read("*a"); fp:close()

		rv = string.match(content, "^.*welltech_secret=s:(.*)\"")
	end

	return rv
end

-- Get PBX time zone
-- @return a string with PBX time zone
function getPbxTimeZone()
    local filename = "/etc/timezone"
    local rv = "Europe/Rome" --default value
    if is_exists (filename) then
        local file = io.open(filename, "r")
        local contents = file:read("*a")
        file:close()

        local timeZone = string.match(contents or "", "[%a%-_]+/[%w%+%-_]+/?[%w%+%-_]+")
        if timeZone then
            rv = timeZone
        end
    end

    return rv
end

-- Get SIP redirect port
function getSipRedirectPort()
    local core = require"common.core"
    local connection = core.connection 

    connection:switch ("devdb")

    local sql = "select value from system where variable=%q"
    local cursor = connection:execute (sql, "SIP_REDIRECT_PORT")
    local port = cursor:fetch(); cursor:close()
    connection:close() 

    return tonumber(port)
end

-- Check if run in cloud
function isInChrootEnvironment()
	local rv = false
	local filename = "/proc/mounts"

	if is_exists(filename) then
		local file = io.open(filename, "r")
		local contents = file:read("*a")
		file:close()

		if string.match (contents, "initctl") then
			rv = true
		end
	end

	return rv
end

function escapeLuaString(x)
    return (x:gsub('%%', '%%%%') 
      :gsub('^%^', '%%^') 
      :gsub('%$$', '%%$') 
      :gsub('%(', '%%(') 
      :gsub('%)', '%%)') 
      :gsub('%.', '%%.') 
      :gsub('%[', '%%[') 
      :gsub('%*', '%%*') 
      :gsub('%+', '%%+') 
      :gsub('%-', '%%-') 
      :gsub('%?', '%%?')) 
end

function isTableEmpty(t)
    local rv = true

    if not t or "table" ~= type(t) then
        return rv
    end

    for _ in pairs(t) do rv = false; break end
    return rv
end

--[[
Copyright (c) 2013 Enrique García Cota

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to
    permit persons to whom the Software is furnished to do so, subject to
    the following conditions:

    The above copyright notice and this permission notice shall be included
    in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
--]]

-- Numbers taken from table 3-7 in www.unicode.org/versions/Unicode6.2.0/UnicodeStandard-6.2.pdf
-- find-based solution inspired by http://notebook.kulchenko.com/programming/fixing-malformed-utf8-in-lua
function utf8_validate(str)
	local find = string.find
	local i, len = 1, #str
	while i <= len do
		if     i == find(str, "[%z\1-\127]", i) then i = i + 1
		elseif i == find(str, "[\194-\223][\128-\191]", i) then i = i + 2
		elseif i == find(str,        "\224[\160-\191][\128-\191]", i)
			or i == find(str, "[\225-\236][\128-\191][\128-\191]", i)
			or i == find(str,        "\237[\128-\159][\128-\191]", i)
			or i == find(str, "[\238-\239][\128-\191][\128-\191]", i) then i = i + 3
		elseif i == find(str,        "\240[\144-\191][\128-\191][\128-\191]", i)
			or i == find(str, "[\241-\243][\128-\191][\128-\191][\128-\191]", i)
			or i == find(str,        "\244[\128-\143][\128-\191][\128-\191]", i) then i = i + 4
		else
			return false, i
		end
	end

	return true
end

-- wms-6715
-- creates hash password for unknown account
-- returns md5 hash (first 20 symbols) or "1234"
function pass_for_unknown(uname)
    if not uname and "" == uname then
        return "1234", "invalid argument"
    end
    local user_passwd = string.sub(require"md5".digest(uname .. require"core.pbxutils".pbxKey()) or "", 1, 20)
    if "" == user_passwd then
        return "1234", "md5 hash is incorrect"
    end
    return user_passwd
end

-- Create random password
-- @param length generated password length, 15 by default
-- @return string
function generate_pass(length)
	local socket = require"socket"
	local password = require"md5".digest(common.core.get_random_bytes(6) .. tostring(socket.gettime() * 10000))
	return password:sub(0, length or 15)
end

function get_sip_pass(trunkname, trunkid, port, mac)
	if not trunkname or "" == trunkname or not port or not trunkid then
		logger:error("error while getting or creating sip password: incorrect input arguments")
		return false
	end
	local core = require"common.core"
	local connection = core.connection
	connection:switch("htables")
	local username
	if string.find(trunkname, "fxo") then
		username = string.format("%s_%d", trunkname, port)
	else
		username = string.format("%s_%d_0", trunkname, port)
	end
	local sql = "SELECT key_value FROM dynamic_peers WHERE key_name='%s'"
	local cursor = connection:execute(sql, username .. "::secret")
	local password = cursor:fetch(); cursor:close(); connection:close()
	-- WMS-8153 checks if existing sip password is old password type (7 symbols)
	local needUpdate = false
	if not password or "" == password then
		needUpdate = true
		logger:error("Sip password not found in database")
	elseif 7 == #password then
		needUpdate = true
		logger:info("Updating old sip password with new one")
	end

	if needUpdate then
		password = update_old_pass(username, trunkid, mac)
	end
	return password
end

function update_old_pass(username, trunkid, mac)
	if not username or "" == username then
		return false
	end
	local password = generate_pass()
	local core = require"common.core"
	local hc = require"core.httpClient":new()
	local connection = core.connection
	connection:switch("htables")
	local sql = "INSERT or REPLACE INTO dynamic_peers (id, key_name,key_value,wms_id) VALUES ((SELECT id FROM dynamic_peers WHERE key_name=%q AND wms_id=%q),%q,%q,%q)"
	connection:execute(sql, username .. "::secret", trunkid, username .. "::secret", password, trunkid)
	connection:close()
	create_sip_extra()
    local response = hc:get(string.format("http://127.0.0.1:5060/v1/unconfirmed_pwd/?mac=%s", mac:lower()))
	if 200 ~= response.code then
		logger:error("Can't connect to kamailio api to change config from v1 to v2!")
	end
	os.execute "sudo kamcmd htable.reload dynamic_peers"
	os.execute "sudo callweaver -rx \"sip reload \" > /dev/null"
	return password
end

function get_cg_managers()
	local managers = {}
	local conf = "/rw2/etc/callweaver/queues-entities.conf"
	local content = ""

	local fp = io.open(conf)
	if fp then
		content = fp:read("*a")
		fp:close()
	end

	for manager in string.gmatch(content, "queue%-manager=(%S+)[\r\n]") do
		managers[manager] = true
	end

	return managers
end

function get_cg_managers_cache()
	local conf = "/rw2/etc/callweaver/queues-entities.conf"
	local cname = "/tmp/cg.managers"

	local function __rebuild_and_get()
		local managers = get_cg_managers()
		local tmpname = cname .. "." .. string.sub(tostring(require"socket".gettime() * 10000), -8)
		local tfile = io.open(tmpname, "w")
		for m, _ in pairs(managers) do
			tfile:write(m .. "\n")
		end
		tfile:close()
		os.rename(tmpname, cname)
		return managers
	end

	local function __get_cached(content)
		local managers = {}
		for manager in string.gmatch(content, "(%S+)\n") do
			managers[manager] = true
		end
		return managers
	end

	local cfile = io.open(cname)
	if not cfile then
		return __rebuild_and_get()
	end

	local fsutils = require"core.fsutils"
	if fsutils.getmTime(conf) > fsutils.getmTime(cname) then
		cfile:close()
		return __rebuild_and_get()
	end

	local cached = cfile:read("*a")
	local managers = __get_cached(cached)
	cfile:close()
	return managers
end

--- Return environment variable by key
-- @param key			Variable key
-- @param defaultValue	Default value
-- @return environment variable value or default value
function env(key, defaultValue)
	if not key or "" == key then
		return defaultValue
	end

	local filename = "/rw2/etc/env.ini"
	local filename_custom = "/rw2/etc/env.custom.ini"

	local config = {}
	if is_exists(filename) then
		config = parse_simple_ini(filename)
	end
	if is_exists(filename_custom) then
		local custom = parse_simple_ini(filename_custom)
		require"core.pbxutils".table_concat(config, custom)
	end

	if config[key] then
		return config[key]
	end

	return defaultValue
end

-- Assign user extension to w-air handset/headset via wms-interface
-- Example: 55001 login5_1 unknown5_1 9c7514340014 02:53:d3:a3:8f w-air-sbr2
-- Example: kamctlwrapper 'unknown1_1@wildix' 'sip:unknown1_1@10.100.5.42' &
function assign_wair_user(target_extension, display_name, current_extension, base_mac, headset_ipei, base_model, sipuri)
    local logger = require"core.logger":init("[WMS 3.0]", 3)
    if not base_mac or not headset_ipei or not target_extension or not current_extension then
        logger:error("WAIR base mac address, headset IPEI or extension is empty. No way to assign the extension.")
        return false
    end
    if not base_model then
        base_model = "W-AIR"
    end
    local removeContact = false
    if sipuri and "" ~= sipuri then
        removeContact = "kamctlwrapper " .. "'" .. current_extension .. "@wildix' '" .. sipuri .. "'" .. " &"
    end
    package.path = package.path..";/var/www/provision/?.lua"
    package.path = package.path..";/var/lib/callweaver/dialplan/?.lua"
    local useragent = string.format("Wildix_%s/WAIR-HS_00.00.00.00_%s_%s", base_model:upper(), base_mac, headset_ipei)
    local dst_target = unpack(require"core.wmsldap":get("telephoneNumber", target_extension))
    local dst_current = unpack(require"core.wmsldap":get("telephoneNumber", current_extension))
    local options = {}
    if dst_target and 0 ~= #dst_target then -- assign extension
        options = {
            ["username"] = target_extension,
            ["callername"] = display_name,
            ["cid"] = current_extension,
            ["userid"] = dst_target[2]["uid"],
            ["useragent"] = useragent,
            ["removecommand"] = removeContact
        }
    elseif dst_current and 0 ~= #dst_current then -- deassign extension
        local utils = require"utils"
        local unknown_extension = utils:get_free_unknown_account()
        options = {
            ["username"] = unknown_extension,
            ["callername"] = display_name,
            ["cid"] = current_extension,
            ["userid"] = dst_current[2]["uid"],
            ["useragent"] = useragent,
            ["removecommand"] = removeContact
        }
    else
        logger:error("Assign extension to headset fails due to the inability to receive information from LDAP")
        return false
    end
    if options["userid"] then
        require"modules.wildixair".setuser(_, _,options)
        logger:info("Assign extension %s to headset ipei %s and w-air base mac %s", target_extension, headset_ipei, base_mac)
        return true
    end
end
