--- E-mail notifications without and with templates
-- 
-- @file    mailNotifications.lua 
-- @author  Yuriy Zveryanskyy
-- @created 2017/12/14

local mailNotifications = {}

--- Send an email with attach using sendemail.php script
-- @param message       A mail body
-- @param subject       A mail subject
-- @param recipient     A mail addresses
-- @param attachments   An absolute pathes to files to be attached
-- @param com_id        A mail settings per company (currently not used)
-- @param extra         A table with extra parameters for sendemail.php
function mailNotifications.send(message, subject, recipient, attachments, com_id, extra)
    local cmd         = "sendemail.php "
    local cmdbase_len = #cmd
    local rcp_params  = {to="--to %s ", cc="--cc %s ", bcc="--bcc %s "}
    local helper      = require"common.helper"

    recipient = "table" ~= type(recipient) and {to=recipient} or recipient

    for k, param in pairs(rcp_params) do
        local dst = recipient[k]
        if dst then
            cmd = cmd .. string.format(param, string.gsub(helper.trim(dst), "%s+", ","))
        end
    end

    -- empty table
    if #cmd == cmdbase_len then
        return false, "Recipient is empty"
    end

    if "table" == type(extra) then
        for param, val in pairs(extra) do
            cmd = cmd .. string.format("--%s %s ", param, val)
        end
    end

    -- subject and message can be created inside sendemail.php
    if subject then
        local subject = subject ~= "" and subject or "PBX"
              subject = require"base64".encode(subject)
        cmd = cmd .. string.format("--subject64 %s ", subject)
    end

    if message then
        local message = message ~= "" and message or subject
              message = require"base64".encode(message)
        cmd = cmd .. string.format("--body64 %s ", message)
    end

    -- processing attachments
    local fsutils = require"core.fsutils"
    local attparam = ""
    if attachments and "" ~= attachments then
        attachments = "table" ~= type(attachments) and {attachments} or attachments
        for _, attach_file in pairs(attachments) do
            if not fsutils.is_readable(attach_file) then
                return false, "Not found attachment file: " .. attach_file
            end
            attparam = attparam .. string.format("--attach %s ", attach_file)
        end
    end

    cmd = cmd .. attparam .. "&"

    local pbxutils = require"core.pbxutils"
    if "" ~= attparam then
        -- wait output, because after sending file, it can be removed
        return pbxutils.run(cmd)
    else
        return pbxutils.exec(cmd)
    end
end

--- Send missed call notification via template
-- @param recipient     Recipient email address
-- @param name          Caller name
-- @param number        Caller number
-- @param language      Recipient language
function mailNotifications.missedCallNotify(recipient, name, number, language)
    local encoded_name = require"base64".encode(name)
    local timestamp = tostring(os.time())
    local extra_params = {["type"]      = "missedCall",
                          ["name64"]    = encoded_name,
                          ["number"]    = number,
                          ["timestamp"] = timestamp}

    if language and "" ~= language then
        extra_params.language = language
    end

    return mailNotifications.send(nil, nil, recipient, nil, 0, extra_params)
end

--- Send voicemail via template
-- @param recipient     Recipient email address
-- @param message       Message body
-- @param subject       Message subject
-- @param attachment    Message attachment
-- @param language      Caller language (for transcription feature)
function mailNotifications.sendVoicemail(recipient, message, subject, attachment, language, fileForASR, transcript)
    local extra_params = nil

    if language and "" ~= language then
        extra_params = {["type"] = "voicemail",
                        ["asrLanguage"] = language,
                        ["transcript"] = transcript}
        if "" ~= fileForASR then
            extra_params["asrFile"] = fileForASR
        end
    end

    return mailNotifications.send(message, subject, recipient, attachment, 0, extra_params)
end

return mailNotifications
