--- Procedures with international number
--
-- @file    numberDetectImpl.lua
-- @author  Vyacheslav Khudan
-- @created 2015/11/25

--- @module numberDetectImpl
local numberDetectImpl = {
    cache = {}   -- pool of the processed numbers (avoids secondary data request)
}

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local numberObj = { -- meta Class
    data = nil          -- object internal data          
--         number           -- looked up number
--         type             -- number type (see database field 'type')
--         country          -- result country data
--         area             -- result area data
}

--- Checking number normalization 
-- @param number    A source number for checking
-- @return boolean value
function numberObj:isNormalized(number)
    local rv = false
    if not number or "" == number then
        return rv
    end
    
    -- validate normalization
    if "string" == type(number) and "+" == number:sub(1,1) then
        rv = true
    end
    
    return rv
end

--- Search request to retrieving data by phone number
-- @param number    A requested number for searching
-- @return found object othervice error message
function numberObj:search(number)
    number = number and tostring(number) or ""
    if not self:isNormalized(number) then
        return nil, "Not valid/normalized search number value"
    end
    
    -- checking cache for requested number
    local rv = numberDetectImpl.cache[number]
    if not rv then
        local storageSearch, errMsg = require"common.numberDetect.storageSearchRequest":new()
        if not storageSearch then
            return nil, errMsg
        end
        
        _, errMsg = storageSearch:countryData(number)
        if errMsg then 
            return nil, errMsg
        end

        _, errMsg = storageSearch:areaData(number)
        if errMsg then 
            return nil, errMsg
        end

        local result, errMsg = storageSearch:processResult()
        if not result then
            return nil, errMsg
        end
        
        -- filling result data
        local numberType   = result.areaData.type
        rv      = {number = number}
        rv.type = {ident  = numberType}
        rv.type.name = ({
             FIX   = "geographic number"
            ,VOIP  = "voip telephony"
            ,MOB   = "mobile number"
            ,SAT   = "satellite number"
            ,SPS   = "supplementary service"
            ,EMER  = "emergency number"
            ,FREE  = "free-of-charge service"
            ,PREM  = "premium rate service"
            ,PREM2 = "premium rate service"
            ,PREM3 = "premium rate service"
            ,PREM4 = "premium rate service"
        })[numberType] or "uknown service"
        
        rv.country = {
             code      = result.countryData.code
            ,name      = result.countryData.name
            ,continent = result.countryData.continent
        }
        
        rv.area    = {
             code              = result.areaData.number
            ,description       = result.areaData.description 
        }
        
        -- push found data to cache
        numberDetectImpl.cache[number] = rv
    end
        
    self.data = rv 
    
    return self
end

--- Retrieving data after searching
-- @return table with number data (see number object meta class 'data' field)
function numberObj:getData()
    return self.data
end

--- Retrieve input number value
-- @return number value
function numberObj:getNumber()
    return self.data.number
end

--- Retrieving number type ident
-- @return identificator of the number type
function numberObj:getTypeIdent()
    return self.data.type.ident
end

--- Retrieving number type name
-- @return string representation of the number type
function numberObj:getTypeName()
    return self.data.type.name or ""
end

--- Retrieving country code
-- @return country code related to searching number
function numberObj:getCountryCode()
    return self.data.country.code or ""
end

--- Retrieving country name
-- @return country name related to searching number
function numberObj:getCountryName()
    return self.data.country.name or ""
end

--- Retrieving area code
-- @return area code related to searching number
function numberObj:getAreaCode()
    return self.data.area.code or ""
end

--- Retrieving area description (location/network name, etc.)
-- @return area description related to searching number
function numberObj:getAreaDescription()
    return self.data.area.description or ""
end


--- Object constructor
function numberDetectImpl:new()
    local obj = {}
    setmetatable(obj, {__index = numberObj})
    
    -- initialize object data
    obj.data = {
         number  = nil
        ,type    = {ident = nil, name = nil}
        ,country = {code = nil, name = nil, continent = nil}
        ,area    = {code = nil, geograficLocation = nil, description = nil}
    }
    
    return obj
end

return numberDetectImpl
