--- Executing request to storage for searching data
--
-- @file    storageSearchRequest.lua
-- @author  Vyacheslav Khudan
-- @created 2015/11/25

--- @module storageSearch
local storageSearch = {}

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local searchObj = {             -- meta Class
     storageConnection = nil
     ,maxLen           = {      -- constants checked manually from database
         country = 6
        ,area    = 8            -- max length from Spain area
     }
     ,result           = nil    -- result data structure with following format
--         country                  -- contry lookup data
--             codeLen    
--             primaryISO
--             list
---        area                     -- area lookup data
--             numberLen
--             list
}

--- Procedure for searching country data for specified number
-- @param number    A requested number for searching
-- @return table with found country data otherwise error message
function searchObj:countryData(number)
    if not number or "" == number then
        return nil
    end
    
    local countryReqNumbers = {}
    for l = self.maxLen.country,1,-1 do
        local numberPart = number:sub(1+1, l+1) -- skip '+'
        if numberPart and "" ~= numberPart then
            table.insert(countryReqNumbers, "country_code='"..numberPart.."'")
        end
    end
    if 0 == #countryReqNumbers then
        return nil, "List with country code numbers part is empty"
    end
    
    local countryCodeQuery = "SELECT country_code, iso_2, country_name, continent, tag FROM Countries WHERE "..table.concat(countryReqNumbers, " OR ")
    local sqlRes           = self.storageConnection:rows(countryCodeQuery)
    if not sqlRes then
        return nil, string.format("Not found country code in the number %q", number)
    end
    local countryCodeLen = 0
    local countryList    = {}
    for countryCode, isoCode, countryName, continent, isoTag in sqlRes do
        if countryCode and isoCode then
            countryList[isoCode] = {
                 code      = countryCode
                ,codeLen   = countryCode:len()
                ,name      = countryName
                ,continent = continent 
                ,iso       = isoCode
                ,isoTag    = isoTag
                ,areaTable = string.format("Area_%s_%s", isoCode, countryCode)
            }
            -- detect maximum country code length
            if countryList[isoCode].codeLen > countryCodeLen then
                countryCodeLen = countryList[isoCode].codeLen
            end
        end
    end
    if 0 == countryCodeLen then
        return nil, string.format("Not found country with number %q", number)
    end
    
    -- processing country list
    local primaryISOCode = nil
    for isoCode, country in pairs(countryList) do
        if country.codeLen == countryCodeLen then
            self.result.country.list[isoCode] = country
            if not primaryISOCode then
                primaryISOCode = country.iso
            end
            if "p" == country.isoTag then
                primaryISOCode = country.iso
            end           
        end
    end
    
    self.result.country.codeLen    = countryCodeLen
    self.result.country.primaryISO = primaryISOCode
    
    return self.result.country
end

--- Procedure for searching area data for specified number
-- @param number    A requested number for searching
-- @return table with found area data otherwise error message
function searchObj:areaData(number)
    if not number or "" == number or 0 == self.result.country.codeLen then
        return nil
    end
    
    local countryCodeLen = self.result.country.codeLen
    
    -- remove conutry code data
    number       = number:sub(1+1+countryCodeLen)   -- skip '+' and CC
    if not number then
        return nil
    end
    local numLen         = number:len()
    local indexLen       = self.maxLen.area > numLen and numLen or self.maxLen.area
    local areaReqNumbers = {}
    for l = indexLen,1,-1 do
        local numberPart = number:sub(1, l)
        if not numberPart or "" == numberPart then
            break
        end
        table.insert(areaReqNumbers, "significant_number='"..numberPart.."'")
    end
     if 0 == #areaReqNumbers then
        return nil, "List with area code numbers part is empty"
    end
    local areaNumbersFilterStr = table.concat(areaReqNumbers, " OR ")
    
    local areaCodeQuery = {}
    for _, country in pairs(self.result.country.list) do
        local queryStr = "SELECT significant_number, '%s' AS country_iso, type, description FROM %s WHERE (%s) AND (number_length='%s' OR number_length='0')"
        local query    = string.format(queryStr, country.iso, country.areaTable, areaNumbersFilterStr, numLen)
        table.insert(areaCodeQuery, query)
    end
    if 0 == #areaCodeQuery then
        return nil, "Internal error with area code database request"
    end
    
    local sqlRes      = self.storageConnection:rows(table.concat(areaCodeQuery, " UNION "))
    if sqlRes then
        local areaCodeLen = 0
        local areaList    = {}
        for areaNumber, isoCountryCode, type, description in sqlRes do
            if areaNumber and type then
                areaList[isoCountryCode..areaNumber] = {
                     number         = areaNumber
                    ,numberLen      = areaNumber:len()
                    ,isoCountryCode = isoCountryCode
                    ,type           = type
                    ,description    = description
                }
                
                -- detect maximum country code length
                if areaList[isoCountryCode..areaNumber].numberLen > areaCodeLen then
                    areaCodeLen = areaList[isoCountryCode..areaNumber].numberLen
                end
            end
        end
        
        -- remove not valid area data (checking by MAX area number length)
        self.result.area.numberLen = areaCodeLen
        for _, area in pairs(areaList) do
            if areaCodeLen == area.numberLen then
                table.insert(self.result.area.list, area)
            end
        end
    end
    
    if 0 == self.result.area.numberLen then
        return nil, string.format("Not found area with number %q", number)
    end
    
    return self.result.area
end

--- Processing result after search procedure
-- @return found data for searched number otherwise error message
-- @note This step must be executed always for closing storage connection
function searchObj:processResult()
    self.storageConnection:close()
    if 0 == self.result.country.codeLen then
        return nil, "Not found any data"
    end
    
    local areaData       = self.result.area.list[1]
    local isoCountryCode = areaData and areaData.isoCountryCode or self.result.country.primaryISO
    if 1 < #self.result.area.list then
    	isoCountryCode = self.result.country.primaryISO
    	for k, v in pairs(self.result.area.list) do
    		if isoCountryCode == v.isoCountryCode then
                areaData       = self.result.area.list[k]
            end
        end
    end
    local rv = { 
         countryData = self.result.country.list[isoCountryCode] or {}
        ,areaData    = areaData or {}
    }
    
    -- remove trash from result data
    rv.countryData.areaTable   = nil
    rv.countryData.iso         = nil
    rv.countryData.isoTag      = nil
    rv.areaData.isoCountryCode = nil
    
    return rv
end


--- Object constructor
-- @note make connection to storage
function storageSearch:new()
    local obj = {}
    setmetatable(obj, {__index = searchObj})
    
    -- initialize object result data
    obj.result = {
        country = {
             codeLen    = 0
            ,primaryISO = nil
            ,list       = {}
        }
        ,area    = {
             numberLen = 0
            ,list      = {} 
        }
    }
    
    -- initialize storage
    local storageConn = require"core.wsqlite":new("wnpdb")
    if not storageConn then
        return nil, "Storage connection error"
    end    
    obj.storageConnection = storageConn

    return obj
end

return storageSearch