module(... or "permissions", package.seeall)

local lfs        = require"lfs"
local logger     = require"core.logger":init("permissions", 3)
local CACHE_TIME = 300 -- 5 minutes

--TODO: use function from pbxutils.lua
local function split(str, pat)
   local str = str or ""
   local t = {}  -- NOTE: use {n = 0} in Lua-5.0
   local fpat = "(.-)" .. pat
   local last_end = 1
   local s, e, cap = tostring(str):find(fpat, 1)
   while s do
      if s ~= 1 or cap ~= "" then
         table.insert(t,cap)
      end
      last_end = e+1
      s, e, cap = str:find(fpat, last_end)
   end
   if last_end <= #str then
      cap = str:sub(last_end)
      table.insert(t, cap)
   end
   return t
end

--TODO: use function from pbxutils.lua
local function tprint (tt, indent, done)
	if tt == nil then return "" end
	
	local dstring = ""
	done = done or {}
	indent = indent or 0
	if type(tt) == "table" then
		for key, value in pairs (tt) do
			dstring = dstring .. string.rep (" ", indent) -- indent it
			if type (value) == "table" and not done [value] then
				done [value] = true
				dstring = dstring .. string.format("[\"%s\"] = {\n", tostring (key));
				dstring = dstring .. tprint (value, indent + 7, done)
				dstring = dstring .. string.rep (" ", indent+4) -- indent it
				dstring = dstring .. "},\n";
			else
				dstring = dstring .. string.format("[\"%s\"] = \"%s\",\n", tostring (key), tostring(value))
			end
		end
	else
		dstring = dstring .. tt .. "\n"
	end
	return dstring
end

local function updateDefaults(rules, what, can_type)
    local rv = rules or {}

    if "VOICEMAIL_WITHOUT_PIN" == what then
        if "1" == require"core.pbxutils".getDialplanGeneralSettings("GCOUNTRY_CODE") then
            if rv["canuse"] and "table" == type(rv["canuse"]) then
                rv["canuse"]["VOICEMAIL_WITHOUT_PIN"] = "no"
            else
                rv["canuse"] = { VOICEMAIL_WITHOUT_PIN = "no" }
            end
        end
    end

    return rv
end

local function isInternational(callClass)
    local classList = require"callClass.callClassDefines".classList
    local regionCallClasses = require"callClass.callClassDefines".continentsCallClasses
    for k, _ in pairs(regionCallClasses) do
        if tostring(callClass) == tostring(k["p"]) then
            return true
        end
    end
    return false
end

local function getMaxLicense()
    local license_file = {}
    local f = io.open("/rw2/etc/sf2", "r")
    if f then
        for fline in f:lines() do
            table.insert(license_file, fline)
        end
        io.close(f)
    else
        logger:error("Can not read license file!")
    end

    if license_file[19] then
        local licenses = require"core.pbxutils".split(license_file[19], "|")
        if licenses and #licenses >= 4 then
            local license_types = {"basic", "essential", "business", "premium"}
            for i = 4, 1, -1 do
                if tonumber(licenses[i]) and tonumber(licenses[i]) ~= 0 then
                    return license_types[i] or "essential"
                end
            end
        end
    end
    return "essential"
end

--- Get group entries
-- @param gid A string with group id identificatoer ("dn")
-- @return table
local function get_group (param, value)
	local attrs, dn = {}
	
	local result, err = require"core.wmsldap":get(param, value, "wGroup")
	if #result ~= 0 then
		dn, attrs = unpack(result[1])
	else
		logger:error(err)
	end
	
	return attrs, dn
end

local function update_rules_table(rules_t, wg2gacl_rows, paternity_iteration)
	for r = 1, #wg2gacl_rows do
		local prior, gid, action, value = unpack (split (wg2gacl_rows[r], "|"))
		local key = ""
		if value == "yes" or value == "no" then
			key = "yesno"
		elseif value == "use" or value == "nouse" then
			key = "canuse"
		elseif value == "set" or value == "noset" then
			key = "canset"
		elseif value == "call" or value == "nocall" then
			key = "cancall"
		end
		
		if key == "yesno" then
			if not rules_t[key][action] then rules_t[key][action] = {} end
			if rules_t[key][action][gid] == nil then
				if rules_t[key][action]["all"] == nil then
					-- "all" is not present yet
					rules_t[key][action][gid] = value
					if gid == "all" then rules_t[key][action]["tmp_pos_of_all"] = paternity_iteration end
				else
					-- "all" is present
					-- action "all" we don't need any more
					if action ~= "all" then
						local pos_of_all = rules_t[key][action]["tmp_pos_of_all"] or 0
						if pos_of_all == paternity_iteration then
							-- so it's the same group we iterate yet - not the parent
							-- cause if we already have "all" we don't need parent rules
							rules_t[key][action][gid] = value
						end
					end
				end
			end
		elseif key ~= "" then
			if key == "cancall" and action ~= "all" then action = "call" .. action end
			if rules_t[key][action] == nil then
				if rules_t[key]["all"] == nil then
					-- "all" is not present yet
					rules_t[key][action] = value
					if action == "all" then rules_t[key]["tmp_pos_of_all"] = paternity_iteration end
				else
					-- "all" is present
					-- action "all" we don't need any more
					if action ~= "all" then
						local pos_of_all = rules_t[key]["tmp_pos_of_all"] or 0
						if pos_of_all == paternity_iteration then
							-- so it's the same group we iterate yet - not the parent
							-- cause if we already have "all" we don't need parent rules
							rules_t[key][action] = value
						end
					end
				end
			end
		end
		
	end
	return rules_t
end

local function get_group_rules(group_dn,rules_t,paternity_iteration,parents_list)
	rules_t = rules_t or {
		["yesno"] = {},
		["canuse"] = {},
		["canset"] = {},
		["cancall"] = {}
	}
	
	parents_list = parents_list or {
		[group_dn] = group_dn
	}
	
	local attrs = get_group("dn", group_dn)
	local wg2gacl_rows = attrs["wg2gAcl"] or {}
	if type(wg2gacl_rows) ~= "table" then
		wg2gacl_rows = { wg2gacl_rows }
	end
	
	local paternity_iteration = paternity_iteration or 1
	
	rules_t = update_rules_table(rules_t, wg2gacl_rows, paternity_iteration)
	
	-- use only one parent if present
	if paternity_iteration == 2 then return rules_t end
	
	paternity_iteration = paternity_iteration + 1
	
	if attrs["owner"] ~= nil then
		if parents_list[attrs["owner"]] ~= nil then
			-- we have a loop of parents
			do return rules_t end
		end
		get_group_rules(attrs["owner"],rules_t,paternity_iteration,parents_list)
	end
	return rules_t
end

local function get_group_table(group_dn) 
    local rules_t = 0
	local lfs = lfs
	local is_recache = 0, true
	if not group_dn or group_dn == "" then
		return rules_t
	end
	
	-- check if a cache file present /dev/shm/wg<gid>.acl
	local gid = string.match (group_dn, "gid=(%d+),.*")
	if not gid or gid == "" then
		return rules_t
	end
	local filename = "/dev/shm/wg" .. gid .. ".acl"
	if require"core.fsutils".is_readable(filename) then
		local attrs = lfs.attributes (filename)
		local diff = os.difftime (os.time(), attrs.modification)
		local expired = CACHE_TIME - diff
		if attrs and (diff < CACHE_TIME) then -- 1 minutes
			logger:debug("Use cached file. Exp: %d", expired)
			
			do  -- loading data from cache without global variables
                local env = setmetatable({}, {__index=_G})
                pcall(setfenv(loadfile(filename), env))
                setmetatable(env, nil)
                if env and "table" == type(env.group_acl) then
                    rules_t = env.group_acl
                end
			end
			is_recache = false
		else
			logger:debug("Update cache file. Exp: %d", expired)
		end
	end

	if is_recache or not require"core.fsutils".is_readable(filename) then
		rules_t = get_group_rules(group_dn)
		-- {{ dump file
		local content = "group_acl = {" .. tprint (rules_t) .. "}"
		local chunks = loadstring (content, filename)  -- compiled chunks
		local out, err = io.open(filename, "wb")
		if out then
			out:write (string.dump (chunks))
			out:close()
		elseif err then
			logger:warning("Cannot write ACL cache to file: %s", err)
		end
		-- }}
	end
    
	return rules_t
end

function check(self, group_dn, can_type, what, dest_group_dn, license_type)
    local rulesValueByDefault = {
        yesno = {
            CHANGE_PRESENCE = "no"
           ,DELETE_CALLS    = "no"
           ,SEE_VOICEMAIL   = "no"
        }
       ,DEFAULT        = "yes"
    }
    rulesValueByDefault = updateDefaults(rulesValueByDefault, what, can_type)

    -- license check
    local licenses = require"common.licensesData" or {}
    license_type = "string" == type(license_type) and license_type:lower()
    if not license_type or "" == license_type then
        license_type = getMaxLicense()
        logger:debug("The license is not set for the user. '%s' will be applied. ACL: '%s'", license_type, what)
    end
    if not licenses[license_type] then
        local unknown_license = license_type
        license_type = getMaxLicense()
        logger:error("Unknown license type: '%s'. '%s' will be applied. ACL: '%s'", unknown_license, license_type, what)
    end
    if licenses[license_type][tostring(what)] then
        logger:info("Forbidden to use '%s' due to the type of user license: '%s'", what, license_type)
        local rv = {["yesno"]   = "no",
                    ["canset"]  = "noset",
                    ["canuse"]  = "nouse",
                    ["cancall"] = "nocall"
                   }
        return rv[can_type]
    end

    local rv          = rulesValueByDefault[can_type] and rulesValueByDefault[can_type][what] or rulesValueByDefault.DEFAULT 
    local group_rules = get_group_table(group_dn)
    if "table" ~= type(group_rules) then
       return rv
    end
    if not group_rules[can_type] then
       return rv
    end

    what_id = what
    if "yesno" == can_type then
        if group_rules[can_type][what] then
            if group_rules[can_type][what][dest_group_dn] then
                rv = group_rules[can_type][what][dest_group_dn]
            elseif group_rules[can_type][what]["all"] then
                rv = group_rules[can_type][what]["all"]
            end
        end
    elseif "canset" == can_type or "canuse" == can_type or "cancall" == can_type then
        what = "cancall" == can_type and "call" .. what or what
        if group_rules[can_type][what] then
            rv = group_rules[can_type][what]
        elseif "cancall" == can_type and group_rules[can_type]["call21"] and isInternational(what_id) then
            rv = group_rules[can_type]["call21"]
        elseif group_rules[can_type]["all"] then
            rv = group_rules[can_type]["all"]
        end
        if "set" == rv or "use" == rv or "call" == rv then
            rv = "yes"
        end
    end
    return rv
end

function getGroupsByCondition (self, cid, param, value)
   -- check arguments
   if type(cid) ~= "table" and 
   	  tonumber(cid) ~= 0 then

	  -- get entries from ldap
	  local fp = io.popen[[ 
			extenNumber=$(sed -ne '/^LDAP_EXTEN_NUMBER/s/LDAP_EXTEN_NUMBER=\(.*\)/\1/p' /etc/callweaver/extensions-variables.conf); 
			[ -z "$extenNumber" ] && echo -n "telephoneNumber" || echo -n "$extenNumber";
	  ]]

	  local extenNumber = fp:read("*a"); fp:close()
	  local result = require"core.wmsldap":get(extenNumber, cid)
	  if #result ~= 0 then
		 _, cid = unpack (result[1])
	  end
   end

   local tgroup = get_group_table (cid or {})
   local tgroups = {}
   for gid, t in pairs (tgroup) do
	  if t[param] and t[param] == value then
		 tgroups[#tgroups+1] = gid
	  end
   end
   return tgroups
end

--- Retrieve PBX licenses information
-- @param lic A string with what kind of license information to retreive, e.g. "xbees", "classound" etc.
-- @return the value for requested license information - "0" or "1" or nil in case of error
function getPbxLicense(self, param)
	if not param then
		return nil, "Invalid license name"
	end

	self.pbxLicenses = self.pbxLicenses or {}
	if self.pbxLicenses[param] then
		return self.pbxLicenses[param]
	end

	local file, error = io.open("/rw2/etc/sf2", "r")
	if not file then
		return nil, error
	end

	local sf = {}
	for line in file:lines() do
		sf[#sf+1] = line
	end
	file:close()

	local sf19 = require"core.pbxutils".split(sf[19], "|") or {}
	self.pbxLicenses.classound = sf19[10] or "0"
	self.pbxLicenses.xbees     = sf[21] or "0"

	return self.pbxLicenses[param]
end