--- Authorization basic class for an user login procedure
-- 
-- @file    userAuthBasic.lua 
-- @author  Vyacheslav Khudan
-- @created 2018/01/22

package.path = package.path .. ";/usr/local/share/lua/5.1/common/userAuth/?.lua"

--- @module userAuthBasic
local userAuthBasic = {}

----------------------------------------------------------------
-- Class implementations
----------------------------------------------------------------
local auth = {      -- meta class
     mAuthSystemDir   = "/tmp/session/"
    ,mAuthPrefix      = "authBasic_"            -- prefix in the authorization system files names
    ,mAuthIdent       = nil                     -- authorization ident/session name
    ,mAuthPassLen     = nil                     -- length of the password for authorization
    ,mAuthBackend     = nil                     -- authorization backend object
    ,mReqData         = nil                     -- the build/normalized input request data for authorization
    ,mBackendData     = nil                     -- retrieved backend data
    ,mAuthLockTimeout = 0                       -- default user lock timeout in seconds
    
    ,mAuthFreezeAttemptList = {                  -- list of the attempts with timeout for freeze sessions
         [5]   = 30                                 -- freeze for 30 seconds if 5th invalid password entering
        ,[10]  = 30                                 -- freeze for 30 seconds if 10th invalid password entering
        ,["*"] = 60                                 -- freeze for 60 seconds after every attempt starting from 10 + 1
    } 
}

--- Retrieve the requestor identificator for this login procedure
-- @return an identificator value
function auth:__getRequestorIdent()
    return self.mReqData.requestorIdent
end

--- Retrieve identificator for current authorization process
-- @return authorization module identifier
function auth:__getIdent()
    local requestor = self:__getRequestorIdent()
    if not requestor then
        return nil
    end
    
    local hash  = string.format("%s", requestor)
    hash        = require"md5".digest(hash)
    local ident = self.mAuthSystemDir .. self.mAuthPrefix .. hash
    return ident
end

--- Retrieve an identificator for a freeze autorization session
-- @return a freeze session module identifier
function auth:__getFreezeIdent()
    local t = {}
    local idents = "table" == type(self.mAuthIdent) and self.mAuthIdent or {self.mAuthIdent}
    for _, val in pairs(idents) do
        t[#t+1] = val .. ".freeze"
    end
    return t
end

--- Retrieve an identificator for a saving autorization session credentials
-- @return a credentials session module identifier
function auth:__getCredentialsIdent()
    return self.mAuthIdent .. ".cred"
end

--- Authentificate by input data
function auth:__authenticate(exten, password)
    local rv  = false
    if not exten or not password then
        return rv
    end
    
    local function _check(backendData)
        local backendPass = backendData.userPIN or require"core.pbxutils".chars2number(backendData.userPassword)
        if not backendPass then
            return nil
        end
        local backendPassLen  = backendPass:len()
        
        local requiredPassLen = self.mAuthPassLen
        if not requiredPassLen and backendData.userPIN then
            -- set up required pass length to predefined (1..5 characters) in the case userPIN present
            requiredPassLen = backendPassLen
        end
        
        local reqPass         = password
        local reqPassLen      = password:len()
        
        if requiredPassLen then
            if requiredPassLen < backendPassLen then
                backendPassLen = requiredPassLen
            else
                requiredPassLen = backendPassLen
            end
            
            reqPass     = reqPass:sub(1, requiredPassLen)
            backendPass = backendPass:sub(1, requiredPassLen)
        end
        
        return backendPass == reqPass 
    end
    
    local backendAuthData = self.mAuthBackend:getData(exten)
    for idx, data in ipairs(backendAuthData) do
        if _check(data) then
            self.mBackendData = backendAuthData[1]  -- save user attributes
            rv                = true
            break
        end
    end
    
    return rv
end

--- Set up credentials
function auth:__setCredentials()
    local rv        = nil
    local err       = nil
    local identFile = self:__getCredentialsIdent()
    if "file" == lfs.attributes(identFile, "mode") then
        rv, err = lfs.touch(identFile)
    else
        rv, err = io.open(identFile, "w")
        if rv then
            rv:close()
            rv = true
        end
    end
    return rv, err
end

--- Clearing non-valid credentials
function auth:__clearNonValidCredentials()
    local identFile  = self:__getCredentialsIdent()
    local cacheTime = lfs.attributes(identFile, "access")
    if cacheTime then
        local actTime    = os.time()
        if cacheTime and (actTime - cacheTime) >= self.mAuthLockTimeout then
            os.remove(identFile)
        end
    end
end

--- Update a required data to freeze a session
-- @return boolean value as a freeze status
function auth:__setSessionAsFreeze()
    local rv = false
    local idents = self:__getFreezeIdent()

    for _, ident in pairs(idents) do
        local freezeCnt = 0
        local fd        = io.open(ident)
        if fd then
            freezeCnt = fd:read("*n")
            fd:close()
            freezeCnt = tonumber(freezeCnt) or 0
        end
        freezeCnt = freezeCnt + 1
        self:__setSessionFreezeCounter(ident, freezeCnt)
        
        if self.mAuthFreezeAttemptList[freezeCnt] then
            -- making freeze for an specified attempt
            self:__setSessionFreezeTime(ident, self.mAuthFreezeAttemptList[freezeCnt])
            rv = true
        elseif self.mAuthFreezeAttemptList["*"].start < freezeCnt then
            -- making freeze for every attempt
            self:__setSessionFreezeTime(ident, self.mAuthFreezeAttemptList["*"].timeout)
            rv = true
        end
    end
    
    return rv
end

--- Set up freeze counter (invalid attempts) for a session
-- @param cnt   A number of the invalid password input
-- @param ident   A freeze ident
-- @return boolean value as an operation result
function auth:__setSessionFreezeCounter(ident, cnt)
    cnt = tonumber(cnt)
    if not cnt then
        return false
    end
    
    local fd = io.open(ident, "w")
    if not fd then  -- could not open file for writing
        return false
    end

    fd:write(cnt)
    fd:close()
    return true
end

--- Set up a freeze time for a session
-- @return TRUE in case of success or NIL and a system-dependent error code in case of error
function auth:__setSessionFreezeTime(ident, timeOffset)
    timeOffset = tonumber(timeOffset)
    if not timeOffset or 0 >= timeOffset then
        return false
    end
    
    local mTime = os.time() + timeOffset
    return require"core.fsutils".setmTime(ident, mTime)
end

--- Clear a freeze state for an specified session
function auth:__clearSessionFreeze()
    local idents = self:__getFreezeIdent()
    for _, ident in pairs(idents) do
        os.remove(ident)
    end
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Authentificate by input data with smart logic of the session freeze
-- @return boolean value of the operation with error message if required
function auth:authenticate(exten, password)
end

--- Retrieve data for authorized user
-- @note updated after every authentification
function auth:getUserBackendData()
    return self.mBackendData
end

--- Check timeout for saved credentials
function auth:getCredentialsTime()
    local cacheTime = lfs.attributes(self:__getCredentialsIdent(), "access")
    if not cacheTime then
        return 0
    end
    
    -- try to clear a non valid credentials
    self:__clearNonValidCredentials()
    
    local leftTime = os.time() - cacheTime
    return leftTime < self.mAuthLockTimeout and self.mAuthLockTimeout - leftTime or 0
end

--- Retrieve a freeze time for a session
-- @return freeze time in seconds 
function auth:getSessionFreezeTime()
    local freezeIdents       = self:__getFreezeIdent()
    local t = {}
    for _, ident in pairs(freezeIdents) do
        t[#t+1] = require"core.fsutils".getmTime(ident) or 0
    end
    local sessionFreezeTime = math.max(unpack(t))
    local currentTime = os.time()
    return currentTime < sessionFreezeTime and sessionFreezeTime - currentTime or 0
end

--- Set up list of the blocking logic
-- @param list  A table with the attempt - seconds to describe blocking
-- @note used default list described above if this parameter empty
-- @note a list must include element '*' as a default locking timeout for every attempt
--       after a latest described version
function auth:setFreezeAttemptList(list)
    local argList = self.mAuthFreezeAttemptList
    if "table" == type(list) and list["*"] then
        argList = list
    end
    
    local tmpList     = {}
    local lastAttempt = 0
    for k,v in pairs(argList) do
        if tonumber(k) and v then
            tmpList[k]  = v
            lastAttempt = k
        end
    end
    
    tmpList["*"] = {
         start   = lastAttempt
        ,timeout = argList["*"]
    }
    self.mAuthFreezeAttemptList = tmpList
end

--- Object constructor
-- @param reqRawData    The raw request data
-- @params options      The additional options for object initialization
-- @param childObj      The child object for current one 
-- @return object otherwise FALSE following error string
function userAuthBasic:new(reqRawData, options, childObj)
    options = options or {}
    if "table" ~= type(reqRawData) then
        return false, "invalid type of the raw request data"
    end
    
    -- build input request environment
    local reqData    = nil
    local authReqEnv = require"authReqEnvFactory":new(reqRawData)
    for ifName, stat in authReqEnv:iterator("staticCheck") do
        if ifName and stat then
            reqData = authReqEnv:instantiate(ifName)
            break
        end
    end
    if not reqData then
        return false, "unknown request data"
    end
    if not reqData.requestorIdent then
        return false, "not found an identificator of the requested user"
    end
    
    -- initialize authorization backend
    local authBackend = require"authBackend":new("LDAP")
    if not authBackend then
        return false, "error authorization backend building"
    end
    
    -- create auth object
    local obj = "table" == type(childObj) and childObj or {}
    setmetatable(obj, {__index = auth})
    
    -- initialize auth object
    obj.mReqData     = reqData
    obj.mAuthBackend = authBackend
    obj.mAuthPassLen = options.passLen
    
    obj.mAuthIdent    = obj:__getIdent()
    if not obj.mAuthIdent then
        return false, "authorization identificator error"
    end
    
    obj:setFreezeAttemptList(options.FreezeAttemptList)
    
    return obj
end

return userAuthBasic
