--- Authorization class for an user lock procedure in the case device request.
--  Commonly used to make an authorization in a provisioning.
-- 
-- @file    userAuthDevLock.lua 
-- @author  Vyacheslav Khudan
-- @created 2017/01/16

package.path = package.path .. ";/usr/local/share/lua/5.1/common/userAuth/?.lua"
local lfs    = require"lfs"

--- @module userAuthDevLock
local userAuthDevLock = {}

----------------------------------------------------------------
-- Class implementations
----------------------------------------------------------------
local auth = {      -- meta class (main part in userAuthBasic)
     mAuthPrefix      = "authDevLock_"
    ,mAuthLockTimeout = 30 
}

--- Retrieve identificator for current authorization process
-- @see userAuthBasic.__getIdent
function auth:__getIdent()
    local requestor = self:__getRequestorIdent()
    local ua        = self.mReqData.userAgent
    
    if not requestor or not ua then
        return nil
    end
    
    local hash  = string.format("%s|%s", ua, requestor)
    hash        = require"md5".digest(hash)
    local ident = self.mAuthSystemDir .. self.mAuthPrefix .. hash
    return ident
end


----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Authentificate by input data with smart logic of the session freeze
--  and using an entered credentials for defined time.
-- @see input and output values in userAuthBasic:authenticate
-- @return boolean value of the operation with error message if required
function auth:authenticate(exten, password)
    local err         = nil
    local rv          = false
    self.mBackendData = nil
    
    local sessionFreezeTime = self:getSessionFreezeTime()
    if 0 < sessionFreezeTime then
        err = string.format("a session for the requestor '%s' locked for %s seconds", self:__getRequestorIdent(), sessionFreezeTime)
        return rv, err
    end
    
    if 0 < self:getCredentialsTime() then
        self:__setCredentials()
        rv = true
        return rv
    end
    
    if not exten then
        return rv, "not found user extension in the request"
    end
    if not password then
        return rv, "not found user password in the request"
    end
    
    rv = self:__authenticate(exten, password)
    if rv then
        self:__clearSessionFreeze()
        self:__setCredentials()
    else
        if self:__setSessionAsFreeze() then
            err = string.format("a session for the requestor '%s' locked for %s seconds", self:__getRequestorIdent(), self:getSessionFreezeTime())
        else
            err = string.format("a password for the user '%s' is not correct (%d limit length)", exten, self.mAuthPassLen or 0)
        end
    end
    return rv, err 
end

--- Object constructor
-- @see userAuthBasic
-- @mandatory fields reqRawData: uri, ua, remoteIP, user
function userAuthDevLock:new(reqRawData, options)
    local obj, err = require"userAuthBasic":new(reqRawData, options, auth)
    if not obj then
        return false, err
    end
    
    -- additional object parameters
    if options and options.lockTimeout then
        obj.mAuthLockTimeout = options.lockTimeout
    end
    
    return obj
end

return userAuthDevLock
