--- Authorization class for an user login procedure.
--  Commonly used to make an authorization in a dialplan.
-- 
-- @file    userAuthLogin.lua 
-- @author  Vyacheslav Khudan
-- @created 2017/01/16

package.path = package.path .. ";/usr/local/share/lua/5.1/common/userAuth/?.lua"

--- @module userAuthLogin
local userAuthLogin = {}

----------------------------------------------------------------
-- Class implementations
----------------------------------------------------------------
local auth = {      -- meta class
     mAuthPrefix    = "authLogin_",
     extension      = nil,
     srcIP          = nil
}

--- Retrieve identificator for current authorization process
-- @see userAuthLogin.__getIdent
function auth:__getIdent()
    local requestor = self:__getRequestorIdent()
    local ua        = self.mReqData.userAgent
    if not requestor or not ua then
        return nil
    end
    
    local hash  = string.format("%s|%s", requestor, ua)
    hash        = require"md5".digest(hash)
    local ident = self.mAuthSystemDir .. self.mAuthPrefix .. hash

    local ident2 = nil
    if self.extension and self.srcIP then
        hash = string.format("%s|%s", self.extension, self.srcIP)
        hash = require"md5".digest(hash)
        ident2 = self.mAuthSystemDir .. self.mAuthPrefix .. hash
    end

    return {ident, ident2}
end

function auth:__setExtraIdent(exten)
    if channel then
        self.extension = exten
        self.srcIP     = channel["X-SRC-ADDRESS"]:get()
        -- refresh ident
        self.mAuthIdent = self:__getIdent()
    end
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Authentificate by input data with smart logic of the session freeze
-- @see input and output values in userAuthBasic:authenticate
-- @return boolean value of the operation with error message if required
function auth:authenticate(exten, password)
    local err         = nil
    local rv          = false
    self.mBackendData = nil
    local lockError = "a session for the requestor '%s' locked for %s seconds"

    if not exten then
        return rv, "not found user extension in the request"
    end
    if not password then
        return rv, "not found a user password in the request"
    end

    -- we know an extension here
    self:__setExtraIdent(exten)
    local sessionFreezeTime = self:getSessionFreezeTime()
    if 0 < sessionFreezeTime then
        err = string.format(lockError, self:__getRequestorIdent(), sessionFreezeTime)
        return rv, err
    end


    rv = self:__authenticate(exten, password)
    if rv then
        self:__clearSessionFreeze()
    else
        if self:__setSessionAsFreeze() then
            err = string.format(lockError, self:__getRequestorIdent(), self:getSessionFreezeTime())
        else
            err = string.format("a password for the user '%s' is not correct (%d limit length)", exten, self.mAuthPassLen or 0)
        end
    end
    return rv, err 
end

--- Object constructor
-- @see userAuthBasic
-- @mandatory fields reqRawData: ua, source
function userAuthLogin:new(reqRawData, options)
    auth.extension = nil
    auth.srcIP     = nil
    local obj, err = require"userAuthBasic":new(reqRawData, options, auth)
    if not obj then
        return false, err
    end
    
    return obj
end

return userAuthLogin
