--- Module for importing user features and avatar from remote PBX
--
-- @file    importUserFeatures.lua
-- @author  Vyacheslav Khudan
-- @created 2016/05/26

package.path = package.path .. ";/usr/local/share/lua/5.1/common/userFeatures/?.lua"

local logger   = require"core.logger":init("userFeatures", 2)
local pbxutils = require"core.pbxutils"

--- @module importUserFeatures
local importUserFeatures = {
     requestTimeout  = 10     -- timeout in seconds per an every request
    ,requestAttempts = 3      -- attempts number for an every procedure 
}
----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
local function __getUserAvatarFromRemotePBX(reqAddress, userIdent, userExten)
    local result = false
    local err    = nil
    
    repeat
        -- prepare the request
        local httpClient  = require"core.httpClient":new()
        local reqURL      = string.format("https://%s/user_avatar/%s.png", reqAddress, userExten)
        local tmpFilePath = string.format("/tmp/user_avatar_%s.png", userIdent)
        local userAgent   = "Wildix PHP script " .. pbxutils.pbxSerial()
        local timeout     = importUserFeatures.requestTimeout
        local options     = {timeout = timeout, userAgent = userAgent, insecure = true}
        
         -- processing repsonse with attempts
        logger:debug("URL (%s): %s", userAgent, reqURL)
        local rv = nil
        for i = 1, importUserFeatures.requestAttempts do
           rv, err = httpClient:getFile(reqURL, tmpFilePath, options)
           if rv and 200 == rv.code then
               rv = true
               break
           end
           
           -- some timeout before the next attempt
           require"socket".sleep(math.random(100,200)/1000)
        end
        if not rv then
            break
        end
        
        -- move the avatar to a user settings directory
        local userFeatDir = require"userFeatures":getSettingsDir()
        local avatarDir   = string.format("%s/avatars/", userFeatDir)
        local rv, errMsg  = require"core.fsutils".mvfile(tmpFilePath, avatarDir)
        if not rv then
            err = errMsg
            break
        end
        local avatarFile = string.format("%s/user_avatar_%s.png", avatarDir, userIdent)
        require"core.fsutils".chmod(avatarFile, "rw-rw-r--")
        
        result = true
    until true
    
    if not result then
        err = err or "undefined error"
    end
    
    return result, err
end

local function __getUserFeaturesFromRemotePBX(reqAddress, userIdent)
    local result = false
    local err    = nil
    
    repeat
        -- preapre the request
        local httpClient  = require"core.httpClient":new()
        local reqURL      = string.format("https://%s/?action=getUser&userId=%s", reqAddress, userIdent)
        -- the temporary file name shoult be the same as for a user features initialize module
        local tmpFileName = string.format("%s.lua", userIdent)
        local tmpDir      = "/tmp/"
        local tmpFilePath = tmpDir .. "/" .. tmpFileName
        local userAgent   = "WMSCommunicate " .. pbxutils.pbxSerial()
        local timeout     = importUserFeatures.requestTimeout
        local options     = {timeout = timeout, userAgent = userAgent, insecure = true}
        
        -- processing repsonse with attempts
        logger:debug("URL (%s): %s", userAgent, reqURL)
        local rv = nil
        for i = 1, importUserFeatures.requestAttempts do 
            rv, err = httpClient:getFile(reqURL, tmpFilePath, options)
            if rv and 200 == rv.code then
                rv = true
                break
            end

            -- some timeout before the next attempt
            require"socket".sleep(math.random(100,200)/1000)
        end
        if not rv then
            break
        end
        
        -- processing retrieved user features
        local userFeatures, errMsg = require"userFeatures":new(userIdent)
        if not userFeatures then
            err = "couldn't initialize user features: " .. errMsg
            break 
        end
        local rv, errMsg = userFeatures:loadData(nil, tmpDir)
        if not rv then
            err = errMsg
        end
        
        -- fix up user features
        userFeatures:set("SHAREDVM", {status = 0, value = ""})
        
        -- save user features
        if not userFeatures:save() then
            err = "the user features saving error has occurred"
            break
        end
        
        -- remove temporary file
        require"core.fsutils".rmfile(tmpFilePath)
        
        result = true
    until true
    
    if not result then
        err = err or "undefined error"
    end
    
    return result, err
end

local function __removeUserDataFromRemotePBX(reqAddress, userIdent)
    local rv  = false
    local err = nil
    
    repeat
        -- preapre the request
        local httpClient = require"core.httpClient":new()
        -- XXX: use HTTP to avoid an issue with callweaver CURL request (required CURLOPTION 'CURL_SSL_VERIFYHOST = 0')
        local reqURL     = string.format("http://%s/?action=removeUser", reqAddress)
        local reqBody    = {userId = userIdent}
        local userAgent  = "WMSCommunicate " .. pbxutils.pbxSerial()
        local timeout    = importUserFeatures.requestTimeout
        local options    = {timeout = timeout, userAgent = userAgent}
        
        -- processing repsonse with attempts
        for i = 1, importUserFeatures.requestAttempts do
            rv, err = httpClient:post(reqURL, reqBody, options)
            if rv and 200 == rv.code then
                rv = true
                break
            end

            -- some timeout before the next attempt
            require"socket".sleep(math.random(100,200)/1000)
        end
    until true
    
    return rv, err
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Executing an import/moving for user data
-- @param userIdent     The moving user identificator
-- @param userExten     The moving user extension
-- @param address       The address of remote pbx
-- @return boolean value as a status
function importUserFeatures:exec(userIdent, userExten, address)
    local rv = false
    
    repeat
        if not userIdent or "" == userIdent then
            break
        end
        if not userExten or "" == userExten then
            break
        end
        if not pbxutils.get_ipv4(address) then
            break
        end
        
        local actionsCnt = 0
        
        
        -- moving the user avatar file
        logger:debug("Try to import an avatar file from the remote PBX(%s) for the user '%s'", address, userExten)
        local result, err = __getUserAvatarFromRemotePBX(address, userIdent, userExten)
        if result then
            logger:info("An avatar file for the user '%s' was loaded and saved successfully!", userExten, userIdent)
            actionsCnt = actionsCnt + 1
        else
            logger:warning("Error in retrieving the user '%s' avatar from the remote PBX: %s", userExten, err)
        end
        
        -- retrieving the user preferences data
        logger:debug("Try to import a features from the remote PBX(%s) for the user '%s'", address, userExten)
        local result, err = __getUserFeaturesFromRemotePBX(address, userIdent)
        if result then
            logger:info("User '%s' features were loaded and saved successfully!", userExten)
            actionsCnt = actionsCnt + 1
        else
            logger:warning("Error in retrieving the user '%s' features from the remote PBX: %s", userExten, err)
        end
        
        -- try to remove user related data from remote PBX
        if 1 <= actionsCnt then
            logger:debug("Removing user '%s' features and avatar file from the remote PBX(%s) ...", userExten, address)
            local result, err =  __removeUserDataFromRemotePBX(address, userIdent)
            if result then
                logger:info("The request for removing user related data was processed successfully!")
            else
                logger:info("An error occurred during processing the request to remove user related data ...")
                
            end
        else
            logger:info("The user related data will be saved on the remote PBX!")
        end
    until true
    
    return rv
end

return importUserFeatures
