--- Sending SIP request for synchronize user features with device status
-- like forward settings or user status
--
-- @file    syncUserFeaturesBySip.lua
-- @author  Vyacheslav Khudan
-- @created 2016/05/20

package.path = package.path .. ";/usr/local/share/lua/5.1/common/userFeatures/?.lua"

--- @module syncUserFeaturesViaSip
local syncUserFeaturesViaSip = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
--- Retrieve user data
local function __getUserData(flag, value)
    local rv = nil
    if not flag or not value then 
        return rv
    end
    
    if "exten" == flag then
        flag = "telephoneNumber"
    end
    
    local userData  = require"core.wmsldap":get(flag, value)
    if userData and userData[1] and userData[1][2] then
        rv       = userData[1][2]
        rv.exten = rv.telephoneNumber
    end
    
    return rv
end
    
--- Build command for sync procedure
local function __buildSyncCommand(data)
    if data and data.name and data.exten then
        local syncFeatStatus = data.status and "true" or "false"
        local syncFeatValue  = data.value or ""
        local syncCmd        = string.format("sip sync feature %s %s %s %s", data.exten, data.name, syncFeatStatus, syncFeatValue)
        
        return syncCmd
    end
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Executing sync for all user features located on the disk
function syncUserFeaturesViaSip:full()
    local featIterator = require"userFeatures":iterator()
    for featObj in featIterator do
        local userData  = nil
        local userIdent = featObj:get("uid")
        if userIdent then
            userData = __getUserData("uid", userIdent)
        end
        if self:exec(featObj, userData) then
            require"socket".sleep(1)	-- sending PUBLISH per second
        end
    end
    
    return true
end

--- Retrieve list of the default features that will be synced
-- @return list of the default sync features 
function syncUserFeaturesViaSip:getDefaultSyncFeatures()
    local rv = {"STATUS", "FCF"}
    for idx, name in ipairs(rv) do
        rv[name] = idx
    end
    return rv
end

--- Executing user features sync procedure
-- @param userData  The user information
-- @param featObj   The loaded user features object 
-- @param syncList  The list of the user features for sync (optional)
-- @return boolean value as operation result
function syncUserFeaturesViaSip:exec(featObj, userData, syncList)
    local rv = false
    repeat
        if "table" ~= type(featObj) or "table" ~= type(userData) then
            break
        end
        if not userData.exten then
            break
        end
        
        -- prepare user status and active features
        local userExten = userData.exten
        local usedActiveClass    = featObj:getActiveClass()
        local usedActiveSubClass = featObj:getActiveSubClass()
        if "internal" ~= usedActiveClass and "external" ~= usedActiveClass then
            break
        end
        featObj:setUpActiveClassValuesByActivatedFlag()
        
        
        -- processing user features
        local cmdList = {}
        syncList      = syncList or self:getDefaultSyncFeatures()
        
        -- fix up double notification for user status
        if syncList[syncList.STATUS] then
            syncList[syncList.STATUS] = nil
            syncList[#syncList + 1]   = "DND"
            syncList[#syncList + 1]   = "AWAY"
            syncList[#syncList + 1]   = "ONLINE"
            syncList[#syncList + 1]   = "OFFLINE"
        end
        local custom = require"base64".encode(require"core.pbxutils".tojson({
            location = tostring(featObj:getUserStatusData("location")),
            message = tostring(featObj:getUserStatusData("message")),
            untilTime = tostring(featObj:getUserStatusData("untilTime"))}))
        local presence = "ONLINE"

        for idx, feature in pairs(syncList) do
            local skip_command = false
            if "number" == type(idx) and feature then
                local featName   = nil
                local featStatus = false
                local featValue  = nil
                local userStatus = featObj:get("STATUS")
                if "DND" == feature then
                    -- notification about user status like DND
                    featName         = "DND"
                    if "dnd" == userStatus then
                        featStatus = true
                        presence = "DND"
                    else
                        featStatus = false
                    end
                elseif "AWAY" == feature then
                    -- notification about user status like AWAY
                    featName         = "AWAY"
                    if "away" == userStatus then
                        featStatus = true
                        presence = "AWAY"
                    else
                        featStatus = false
                    end
                elseif "FCF" == feature then
                    -- notification about user forward feature
                    -- checking all forwards, but send only one
                    featName  = "UCF"
                    featValue = featObj:get("FCF")
                    if featValue and "VOICEMAIL" == featValue:upper() then
                        featValue = "VM"
                    end
                    if featValue then
                        featStatus = true
                    else
                        featValue = "0"
                    end
                elseif "ONLINE" == feature then
                    skip_command = true
                    featName = "ONLINE"
                    if "available" == userStatus then
                        featStatus = true
                        presence = "ONLINE"
                    else
                        featStatus = false
                    end
                elseif "OFFLINE" == feature then
                    skip_command = true
                    featName = "OFFLINE"
                    if "unavailable" == userStatus then
                        featStatus = true
                        presence = "OFFLINE"
                    else
                        featStatus = false
                    end
                end
                if not skip_command then
                    cmdList[#cmdList + 1] = __buildSyncCommand({exten = userExten, name = featName, status = featStatus, value = featValue})
                end
            end
        end
        cmdList[#cmdList + 1] = string.format("sip peer set presence %s %s %s", userExten, presence, custom)

        -- roll back used feeatures class and sub class
        featObj:setUpActiveClass(usedActiveClass)
        featObj:setUpActiveSubClass(usedActiveSubClass)
        
        -- execute background script for sending command
        local syncData = require"dkjson".encode(cmdList)
        local syncCmd  = string.format("send_sip_sync_feature 'execute' '%s' &", syncData)
        local cmdPipe  = io.popen(syncCmd)
        rv             = "userdata" == type(cmdPipe)
        if rv then
            cmdPipe:close()
        end
    until true
    
    return rv
end

--- Executing feature sync by some user ident
-- @param identData     The user identification data 
-- @param data          The list of the features for sync (optional)  
-- @param userFeatures  The features object (optional)
-- @return boolean value as operation result 
function syncUserFeaturesViaSip:byIdent(identData, data, userFeatures)
    local rv           = false
    local userData     = nil
    repeat
        if "table" ~= type(identData) then
            break
        end
        
        -- retrieve user data
        userData = __getUserData(identData.flag, identData.value)
        if not userData or not userData.uid then
            break
        end

        -- loading user features
        if "table" ~= type(userFeatures) then
            userFeatures = require"userFeatures":new(userData.uid)
            if not userFeatures then
                break
            end
            userFeatures:load()
        end
        
        -- preapring sync data
        data = data and "table" ~= type(data) and {data} or data
        
        rv = true
    until true
    
    if rv then
        return self:exec(userFeatures, userData, data)
    end
    
    return rv
end

--- Executing feature sync by user extension
-- @see 'syncUserFeaturesViaSip:byIdent' 
function syncUserFeaturesViaSip:byExten(exten, data, userFeatures)
    return self:byIdent({flag = "exten", value = exten}, data, userFeatures)
end

--- Executing feature sync by user ident
-- @see 'syncUserFeaturesViaSip:byIdent' 
function syncUserFeaturesViaSip:byUID(uid, data, userFeatures)
    return self:byIdent({flag = "uid", value = uid}, data, userFeatures)
end

--- Executing user features sync procesure via provision command
-- @param identData     The user identification data
-- @return boolean value as operation result 
function syncUserFeaturesViaSip:general(identData)
    local rv = false
    repeat
        if "table" ~= type(identData) then
            break
        end
        
        -- retrieve user data
        local userData = __getUserData(identData.flag, identData.value)
        if not userData or not userData.exten then
            break
        end
        local userExten = userData.exten
        
        -- preparing sync command
        local registrations = require'core.sipproxy.sipproxy':new():getRegistrationByExten(userExten)
        local syncMode      = 'check-sync'
        local userAgent     = registrations and registrations:first():getUserAgent()
        if userAgent and string.find(userAgent, 'Wildix W%-PA.*') then
            syncMode = 'reboot-phone'
        end
        local cmdList = { string.format("sip notify %s %s", syncMode, userExten) }

        -- execute background script for sending command
        local syncData = require"dkjson".encode(cmdList)
        local syncCmd  = string.format("/usr/sbin/send_sip_sync_feature 'execute' '%s' &", syncData)
        local cmdPipe  = io.popen(syncCmd)
        rv             = "userdata" == type(cmdPipe)
        if rv then
            cmdPipe:close()
        end
    until true
    
    return rv
end

--- Executing general sync by user ident
-- @see 'syncUserFeaturesViaSip:generalSync' 
function syncUserFeaturesViaSip:generalByUID(uid)
    return self:general({flag = "uid", value = uid})
end

return syncUserFeaturesViaSip





