--- Proxy interface to user features with processing logs
--
-- @file    userFeatures.lua
-- @author  Vyacheslav Khudan
-- @created 2016/05/11

package.path = package.path .. ";/usr/local/share/lua/5.1/common/userFeatures/?.lua"

local logger = require"core.logger":init("userFeatures", 2)

--- @module userFeatures
local userFeatures = {}

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Constructor
function userFeatures:new(uid)
    local extUserFeatures, err = require"userFeaturesExtend":new(uid)
    if not extUserFeatures then
        logger:error("Could not load user (%s) features: %s", uid, err)
        return nil
    end
    
    logger:debug("Initialized features for user UID '%s' ...", uid)
    
    --- Proxy interface ---
    -- extend object by proxy methods
    local proxy = setmetatable({}, {__index = extUserFeatures})
    
    -- Processing logs for loading user features
    function proxy:load(data)
        local rv, err = extUserFeatures:load(data)
        if rv then
            logger:debug("The features for user UID '%s' were loaded successfully!", uid)
        else
            logger:debug("Could not load features for user UID '%s' - %s", uid, err)
            logger:info("Using default user features!")
        end
        
        return rv
    end
    
    -- Bind user feautes data (ex. get via front end) to internal structure
    function proxy:bind(data)
        logger:debug("Binding user features ...")
        if "table" ~= type(data) then
            return false
        end
        
        local rv, err = extUserFeatures:load(data)
        if not rv then
            logger:error("Binding user features for user UID '%s' - %s", uid, err)
        end
        
        return rv
    end
    
    -- Processing logs for saving user featuress
    function proxy:save()
        local rv, err = extUserFeatures:save()
        if rv then
            logger:debug("The features for user UID '%s' were saved successfully!", uid)
        else
            logger:error("Could not save features for user UID '%s' - %s", uid, err)
        end
        
        return rv
    end
    
    return proxy
end

--- Retrieve path to user settings folder
function userFeatures:getSettingsDir()
    return require"userFeaturesExtend":getSettingsDir()
end

--- Iterator over user features directory
-- @return user features object otherwise nil 
function userFeatures:iterator()
    logger:debug("Loading user features iterator ...")
    return require"userFeaturesExtend":iterator()
end

--- Direct function for loading user features object with processing disk data
-- @param uid   The user identificator 
-- @return list of the user features
-- @note must be used only on PHP/WMS side (read only mode)
function userFeatures:load(uid)
    logger:debug("Direct loading features for user UID '%s'!", uid or "")
    
    local rv  = {}
    local obj = self:new(uid)
    if obj then
        obj:load()
        rv = obj:getData4FrontEnd() or {}
    end
    
    return rv
end

---  Direct method for saving user features via hide object initialization
-- @param uid   The user identificator 
-- @param data  The user features for saving
-- @return boolean value
-- @note must be used only on PHP/WMS side
function userFeatures:save(uid, data)
    logger:debug("Direct saving features for user UID '%s'!", uid or "")
    local rv           = false
    local obj          = self:new(uid)
    if not obj then
        return rv
    end
    
    local frontEndData = obj:getFrontEndData(data)
    local notifyTypes   = obj:compareFrontEndData(frontEndData)
    if obj:bind(frontEndData) then
        rv = obj:save()
    end
    
    -- send notification to device (class can be disabled)
    local syncFeaturesViaSip = require"syncUserFeaturesViaSip"
    if notifyTypes["feature-sync"] then
        logger:debug("Sending notify with type 'feature-sync' to user UID '%s'...", uid)
        local syncFeatList       = syncFeaturesViaSip:getDefaultSyncFeatures()
        syncFeatList[syncFeatList.STATUS] = nil
        syncFeaturesViaSip:byUID(uid, syncFeatList, obj)
    end
    if notifyTypes["check-sync"] then
        logger:debug("Sending notify with type 'check-sync' to user UID '%s'...", uid)
        syncFeaturesViaSip:generalByUID(uid)
    end

    -- send roster change notification to ejabberd
    if notifyTypes["roster"] then
        logger:debug("Sending roster change to user UID '%s'...", uid)
        require"core.pbxutils".roster_notify(uid, notifyTypes["roster"])
    end

    return rv
end

--- Direct method for loading default user features without disk processing
-- @return list of the user features
-- @note must be used only on PHP/WMS side (multiply users editing)
function userFeatures:default()
    logger:debug("Direct methods for loading default user features ...")
    local rv  = {}
    local obj = self:new("multiply")
    if obj then
        rv = obj:getData4FrontEnd()
    end
    
    return rv
end

return userFeatures
