--- Interface for processing user features
--
-- @file    userFeaturesBasic.lua
-- @author  Vyacheslav Khudan
-- @created 2016/05/10

--- @module userFeaturesBasic
local userFeaturesBasic = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
--- Check if argument is string and has correct value
local function __isStringArg(val)
    local rv = false
    if val and "" ~= val then
        rv = true
    end
    
    return rv
end

--- Loading data from lua file
local function __readLuaFileData(filePath, fileName)
    local rv  = nil
    local err = nil
    
    repeat
        if not filePath or not fileName then
            err = "invalid args"
            break
        end
        
        if not filePath then
            err = "not specified path for user settings directory"
            break
        end

        rv, err = require"core.fsutils".is_dir(filePath)
        if not rv then
            err = string.format("specified data path '%s' is invalid(%s)", filePath, err)
            break
        end

        local userFeaturesFile = filePath.."/"..fileName
        rv, err                = require"core.fsutils".is_file(userFeaturesFile)
        if not rv then
            err = string.format("user features file '%s' is invalid!", userFeaturesFile)
            break
        end
        
        -- loading data from file without global variables
        local userFeaturesChunk = nil
        userFeaturesChunk, err = loadfile(userFeaturesFile)
        if not userFeaturesChunk then
            err = string.format("error parsing user features file: %s", err or "unknown error")
            break
        end
        local env = setmetatable({}, {__index=_G}) 
        pcall(setfenv(userFeaturesChunk, env))
        setmetatable(env, nil)
        rv = env and env.features
        if "table" ~= type(rv) then
            err = string.format("error loading user features")
            break
        end
    until true
    
    return rv, err
end

--- Save data to lua file
local function __saveLuaFileData(filePath, fileName, binData)
    local rv  = false
    local err = nil
    
    repeat
        if not filePath or not fileName or "string" ~= type(binData) then
            err = "invalid args"
            break
        end
        
        if not filePath then
            err = "not specified path for user settings directory"
            break
        end

        rv, err = require"core.fsutils".is_dir(filePath)
        if not rv then
            err = string.format("specified data path '%s' is invalid(%s)", filePath, err)
            break
        end

        local featuresFile = filePath.."/"..fileName
        
        -- Write to temporary file
        local rnd_part = string.sub(require"md5".sumhexa(require"common.core".get_random_bytes(5)), 1, 10)
        local tempFile = featuresFile.."~"..rnd_part
        rv, err = io.open(tempFile, "wb")
        if not rv then
            err = string.format("error oppening user temp file(%s)", err)
            break
        end
        rv:write(binData)
        rv:flush()
        rv:close()
        
        -- Move to original file name
        rv, err = os.rename(tempFile, featuresFile)
        if not rv then
            os.remove(tempFile)
            err = string.format("error during features file saving '%s'(%s)", featuresFile, err)
            break
        end
        
        -- Update file modes
        require"core.fsutils".chmod(featuresFile, "rw-rw-rw-")
        
    until true
    
    return rv, err
end

----------------------------------------------------------------
-- Public interface for general methods
----------------------------------------------------------------
local preferences = {  -- meta class
     _uid      = nil           -- user identificator
    ,_class    = "internal"    -- user actual call type/class
    ,_subClass = "default"     -- user actual sub class value (user status)
    ,_data     = nil           -- user features
    ,_fsPath   = nil           -- path to user feeatures on file system
}

--- Convert features to string representation
-- @param data  A table value with some features data
-- @return string representation of the user features data 
function preferences:tostring(data)
    if data and "table" ~= type(data) then
        return ""
    end
        
    local data = data or self._data
    return require"core.pbxutils".tprint(data)
end

--- Read user features from file system
-- @param data       A data table with user features
-- @param featFolder A custom directory with user features
-- @return boolean value with error message 
function preferences:loadData(data, featFolder)
    local err      = nil
    local fromFile = false
    
    if not data then    -- loading data from file system
        fromFile  = true
        data, err = __readLuaFileData(featFolder or self._fsPath, self._uid..".lua")
    end
    
    if "table" ~= type(data) then
        err  = "user features not found/has invalid format"
        data = nil
    end
    
    -- normalize data to actual version and merge with default values
    local rv = require"userFeaturesNormalize".run(self._data, data, fromFile)
    if rv then
        self:setUpActiveSubClass(self:getUserStatus())   -- update sub class by user status
    end
    
    return rv, err
end

--- Write user features to file system
function preferences:saveData()
    local userFile = self._uid .. ".lua"
    
    -- compile data to binary representation
    local featuresContent = string.format("features = { %s }", require"core.pbxutils".tprint(self._data))
    local binData         = string.dump(loadstring(featuresContent, userFile))
    
    return __saveLuaFileData(self._fsPath, userFile, binData)
end

--- Push value for the user feature
-- @param name   A feature name
-- @param value  A feature value (can be table with status & value)
-- @return boolean value as a status
function preferences:setValueData(name, value)
    local rv = false
    if not __isStringArg(name) or not value then
        return rv
    end
    
    --- feature value look up
    local feature = nil
    if self:isSubClassFeature(name) then  -- processing classes group feature
        feature = self._data._classes[self._class][self._subClass][name]
    else                                                -- processing global value
        feature = self._data[name]
    end
    if not feature then
        return rv
    end
    
    --- update feature value
    local featureType = type(feature)
    if featureType == type(value) then
        if "table" == featureType then
            for name, val in pairs(value) do
                if feature[name] then
                    if "table" == type(val) then
                        feature[name] = val
                    else
                        val = "boolean" == type(val) and "" or val
                        feature[name] = tostring(val)
                    end
                    rv = true
                end
            end
        else
            self._data[name] = value
            rv               = true
        end
    end
    
    return rv
end

--- Check accessory of the classes group feature
-- @param name   A feature name 
-- @return boolean value of the feature accessory
function preferences:isSubClassFeature(name)
    return self._data._classes.internal.default[name]
end

--- Retrieve data of the user feature
-- @param name  A name of the user feature
-- @return table with status and value
function preferences:getValueData(name)
    local rv = {
         status = nil
        ,val    = nil
    }
    if not __isStringArg(name) then
        return rv
    end
    
    if self:isSubClassFeature(name) then  -- loading feature from classes group
        local fData = self._data._classes[self._class][self._subClass][name] or {}
        rv.status = fData.status
        rv.val    = fData.value
    else                                                -- loading feature from global group
        if "STATUS" == name then
            rv.val = self:getUserStatus()
        elseif "PHONEBOOK-ID" == name then              -- get formated phonebook data 
            local pbStr = self._data["PHONEBOOK"] or ""
            rv.val = pbStr:gsub("new%.", ""):gsub("old%.", "")
        else                                            -- TODO: processing other specific fields (HOTLINE)
            local fData = self._data[name]
            if "table" == type(fData) then
                if fData.status then
                    rv.status = fData.status
                    rv.val    = fData.value
                else
                    rv.val    = fData
                end
            elseif "string" == type(fData) then
                rv.val = fData
            end
        end
    end
    
    return rv
end

--- Retrieve feature status
-- @see 'getValueData' 
function preferences:getStatus(name)
    local rv = self:getValueData(name).status
    return "1" == rv and true or false
end

--- Retrieve feature value
-- @see 'getValueData' 
function preferences:getValue(name)
    return self:getValueData(name).val
end

--- Retrieve single user features data (like SHAREDVM)
-- @return list with requested features data 
function preferences:getSingleData()
    local data = {}
    for name, val in pairs(self._data) do
        if "_classes" ~= name then
            data[name] = val
        end
    end
    
    return data
end
 
--- Retrieve data for specified (or internal settings) class name
-- @param className     A searched features class name 
-- @return user features for some class
function preferences:getClassData(className)
    local activeClass = self._class
    if className then
        self:setUpActiveClass(className)
    end
    
    local data = self._data._classes[self._class] or {}
    self:setUpActiveClass(activeClass)
    
    return data
end

--- Retrieve data for some class and sub class name
-- @param className     An searched features class name
-- @param subClassName  An searched features sub class name
-- @return user features for some class/sub class values
function preferences:getSubClassData(className, subClassName)
    local data = self:getClassData(className)
    
    local activeSubClass = self._subClass
    if subClassName then
        self:setUpActiveSubClass(subClassName)
    end
    
    data = data[self._subClass] or {}
    self:setUpActiveSubClass(activeSubClass)
    
    return data
end

--- Retrieve user features data
-- @return list with user features 
function preferences:getData()
    return self._data or {}
end

--- Update user active class
-- @param name  A user class name
-- @return operation status as boolean value
function preferences:setUpActiveClass(name)
    local rv = false
    if self._data._classes[name] then
        self._class = name
        rv          = true
    end
    
    return rv
end

--- Getter for user active class
-- @return user active class value
function preferences:getActiveClass()
    return self._class
end

--- Update user actual sub class value
-- @param val   An sub class value
-- @return boolean value 
function preferences:setUpActiveSubClass(val)
    local rv = false
    
    if not __isStringArg(val) then
        return rv
    end
    val = string.lower(val)
    if "dnd" ~= val and "away" ~= val then
        val = "default"
    end
        
    self._subClass = val
    rv             = true
    
    return rv
end

--- Getter for user actual sub class
-- @return user actual sub class value
function preferences:getActiveSubClass()
    return self._subClass
end

function preferences:isAlternatePickupMode()
    local rv = false
    local groups = self:getValue("ALTERNATE_PICKUP_MODE") or {}
    if groups["callgroups"] and groups["pickupgroups"] then
        local callgroups = groups["callgroups"]:gsub("%s", "")
        local pickupgroups = groups["pickupgroups"]:gsub("%s", "")
        if #callgroups > 0 or #pickupgroups > 0 then
            rv = true
        end
    end
    return rv
end

--- Get pickup/call groups for alternate pickup mode
function preferences:getPickupGroups()
    local rv = { callgroups = {}, pickupgroups = {} }
    local groups = self:getValue("ALTERNATE_PICKUP_MODE") or {}
    if groups["callgroups"] and groups["pickupgroups"] then
        local callgroups = groups["callgroups"]:gsub("%s", "")
        local pickupgroups = groups["pickupgroups"]:gsub("%s", "")
        local utils = require"core.pbxutils"
        rv["callgroups"] = utils.split(callgroups, ",") or {}
        rv["pickupgroups"] = utils.split(pickupgroups, ",") or {}
    end
    return rv
end

--- Get custom identities of the user
function preferences:getCustomIdentities()
    local rv = {}
    local identities = self:getValue("CUSTOM_IDENTITIES") or ""
    if identities and "" ~= identities then
        local utils = require"core.pbxutils"
        rv = utils.split(identities:gsub("%s", ""), ",") or {}
    end
    return rv
end

--- Get allowed callgroups list
function preferences:getAllowedCallgroups()
    local status = false
    local groups = {}
    local cg_whitelist = self:getValue("CALLCENTER_WHITELIST") or {}
    if cg_whitelist["enable"] and cg_whitelist["callGroups"] then
        status = "1" == cg_whitelist["enable"]
        for _, callgroup in pairs(cg_whitelist["callGroups"]) do
            table.insert(groups, callgroup["id"])
        end
    end
    return status, groups
end
----------------------------------------------------------------
-- Public interface for processing activated class data
----------------------------------------------------------------
--- Check activated flag for user class and sub class values
-- @param class     The checked class value
-- @param subClass  The checked sub class value
-- @return boolean value of the checking
function preferences:isActivatedClassValues(class, subClass)
    local rv   = false
    local data = self:getSubClassData(class, subClass)
    
    if data["_ACTIVATED"] and "1" == data["_ACTIVATED"].status then
        rv = true
    end
    return rv
end

--- Set up activated flag for user class and sub class 
-- @param status    The flag status value 
-- @param class     The specified class value
-- @param subclass  The specified sub class value
-- @return boolean value of the set up operation
function preferences:setUpActivatedClassValue(status, class, subclass)
    local rv = false
    
    -- normalize status value
    local statusVal = "0"
    if status then
        statusVal = "1"
    else
        statusVal = self:isActivatedClassValues(class, subclass) and "0" or "1"
    end
    
    local data = self:getSubClassData(class, subClass)
    if "table" == type(data["_ACTIVATED"]) then
        data["_ACTIVATED"].status = statusVal
        rv                        = true
    end
    
    return rv
end

--- Set up active class and sub class by activated flag for this values
function preferences:setUpActiveClassValuesByActivatedFlag()
    local actClass    = self._class
    local actSubClass = self._subClass
    if self:isActivatedClassValues(actClass, actSubClass) then
        -- use predefined class/subclass values
    elseif self:isActivatedClassValues("internal", actSubClass) then
            actClass    = "internal"
    elseif self:isActivatedClassValues(actClass, "default") then
            actSubClass = "default"
    else
        actClass        = "internal"
        actSubClass     = "default"
    end
    self._class    = actClass
    self._subClass = actSubClass
end

----------------------------------------------------------------
-- Public interface to massive feature processing
----------------------------------------------------------------
--- Iterate over user active classes to retrieve feature value
-- @param name  A future name
-- @return iterator over activated classes for specified feature
function preferences:getStatusOverActiveClasses(name)
    local checkedClasses = {}
    if __isStringArg(name) then
        -- save current actual class
        local actClass       = self:getActiveClass()
        
        for className in pairs(self._data._classes) do
            if self:isActivatedClassValues(className) and self:setUpActiveClass(className) then
                table.insert(checkedClasses, {
                         class = className
                        ,val   =  self:getStatus(name)
                    })
            end
        end
        
        -- restore old actual class
        self:setUpActiveClass(actClass)
    end
    
    local i = 0
    return function()
        i = i + 1
        if checkedClasses[i] then
            return checkedClasses[i].class, checkedClasses[i].val
        end
        return nil
    end
end
 
--- Set user data for available classes
-- @param name      A feature name
-- @param value     A feature data 
-- @return list with updated classes 
function preferences:setValueDataOverActiveClasses(name, value)
    -- save current actual class
    local actClass = self:getActiveClass()
    
    local rv     = {}
    for className in pairs(self._data._classes) do
        if self:isActivatedClassValues(className) and self:setUpActiveClass(className) then
            if self:setValueData(name, value) then
                rv[#rv + 1] = className
            end
        end
    end
    
    -- restore old actual class
    self:setUpActiveClass(actClass)
    
    return rv
end

----------------------------------------------------------------
-- Public interface for user status data
----------------------------------------------------------------
--- Retrieve list of the available user statuses
-- @return list of the statuses list 
function preferences:getUserStatusList()
    return require"defUserFeatures".statusValues()
end

--- Push user status data to user features
-- @param data  A list of the status data
-- @return boolean value as operation result
function preferences:setUserStatusData(data)
    local rv = false
    if "table" ~= type(data) then
        return rv
    end
    
    for field, val in pairs(data) do
        if self._data.STATUS[field] and val then
            val = '""' == val and "" or val  -- value normalization
            if "value" == field then
                val = self:getUserStatusList()[val] and val or "available"
            else
                val = require"base64".encode(val)
            end
            if val ~= self._data.STATUS[field] then
                self._data.STATUS[field] = val
                rv = true
            end
        end
    end
    
    return rv
end

--- Set only user status value
-- @param val   An status value 
-- @return boolean value
function preferences:setUserStatus(val)
    return self:setUserStatusData({value = val})
end

--- Retrieve user status data
-- @param fieldName     Field name from status data (optional)
-- @return requested status field value 
function preferences:getUserStatusData(fieldName)
    local rv         = nil
    local statusData = self._data.STATUS
    if fieldName then
        rv = statusData[fieldName]
        if "value" == fieldName then
            if "" == rv then
                rv = "available"
            end
        elseif "untilOffset" == fieldName then
            rv = {}
            local tdObj = require"core.timedate":new(require"base64".decode(statusData.untilTime))
            if tdObj then
                tdObj:toTimeZone()
                local currentTimestamp = os.time()
                      rv.difftime      = (tdObj:getTimestamp() or currentTimestamp) - currentTimestamp
                local hour, min        = math.modf(rv.difftime/3600)
                      min              = math.ceil(min*60)
                      rv.hour          = hour
                      rv.min           = min
            end
        else
            rv = require"base64".decode(rv)
        end
    else
        rv = {}
        for field, val in pairs(statusData) do
            if field and val and "" ~= val then
                if "value" ~= field then
                    val = require"base64".decode(val)
                end
                rv[field] = val
            end
        end
    end
    return rv
end

--- Retrieve user status value
function preferences:getUserStatus()
    local rv = self:getUserStatusData("value")
    return rv
end


---  Object constructor
-- @param uid       User identificatory (LDAP UID)
-- @param fsPath    The file system path where located user features 
-- @param class     Actual user call class/type
-- @return object value otherwise nil with error message
function userFeaturesBasic:new(uid, fsPath, class)
    local obj = nil
    if not __isStringArg(uid) then
        return obj, "Invalid user ident"
    end
    
    -- initialize features object
    obj = setmetatable({}, {
         __index    = preferences
        ,__tostring = preferences.tostring
    })
    obj._uid      = uid
    obj._fsPath   = fsPath
    obj._data     = require"defUserFeatures".new(uid)
    obj.type      = "userFeatures"
    
    obj:setUpActiveClass(class)
    
    return obj
end

return userFeaturesBasic

