--- Adapter for processing user features with old interface
--
-- @file    userFeaturesExtend.lua
-- @author  Vyacheslav Khudan
-- @created 2016/05/11

package.path = package.path .. ";/usr/local/share/lua/5.1/common/userFeatures/?.lua"

--- @module userFeaturesExtend
local userFeaturesExtend = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
--- Make fix up for user data in case using in WMS
local function __featuresFixer(data)
    local rv = nil
    if "table" ~= type(data) then
        return rv
    end
    
    rv = {}
    if data._classes then   -- convert to WMS format
        for name, val in pairs(data) do
            if "_classes" ~= name then
                rv[name] = val
            end
        end
        
        rv.internal  = data._classes.internal
        rv.external  = data._classes.external
        rv.blacklist = data._classes.blacklist
        rv.whitelist = data._classes.whitelist
    else                        -- convert from WMS format
        rv            = data
        rv._classes = {
             internal  = data.internal
            ,external  = data.external
            ,blacklist = data.blacklist
            ,whitelist = data.whitelist
        }
        rv.internal  = nil
        rv.external  = nil
        rv.blacklist = nil
        rv.whitelist = nil
    end
    
    return rv
end

local function __compare_roster_data(old, new)
    if not old or not new then
        return false
    end

    local function __compare_datasets(a, b) -- a minus b
        local rv = {}
        local notempty = false
        for k, v in pairs(a) do
            if not b[k] then
                rv[k] = v
                notempty = true
            end
        end
        return notempty and rv
    end

    local function __get_nick(str)
        -- "Max Mad [213]" -> "Max Mad"
        return string.match(str, "(.-)%s*%[.+%]") or ""
    end

    local function __fix_exten(str)
        -- "213@wildix" -> "213", @wildix due to strange old days reason
        return string.match(str, "([^@]+)@?.*")
    end

    local function __flat_roster(roster)
        local flat = {}
        for group, members in pairs(roster) do
            for name, exten in pairs(members) do
                if "" ~= exten then
                    local fixed_exten = __fix_exten(exten)
                    local record = {
                        ["exten"] = fixed_exten,
                        ["name"]  = name,
                        ["group"] = group
                    }
                    -- handle all cahnges
                    local key = fixed_exten .. name .. group
                    flat[key] = record
                end
            end
        end
        return flat
    end

    local function __diff_data(add, remove)
        if not (add or remove) then
            return false
        end

        local function __to_list(t)
            local rv = {}
            for _, record in pairs(t) do
                rv[#rv + 1] = {record["exten"], __get_nick(record["name"]), record["group"]}
            end
            return rv
        end

        return {["add"] = __to_list(add or {}), ["remove"] = __to_list(remove or {})}
    end

    local old_flat = __flat_roster(old)
    local new_flat = __flat_roster(new)
    local removed  = __compare_datasets(old_flat, new_flat)
    local added    = __compare_datasets(new_flat, old_flat)
    return __diff_data(added, removed)
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Constructor
function userFeaturesExtend:new(uid, class)
    local fsPath               = self:getSettingsDir()
    local userFeaturesObj, err = require"userFeaturesBasic":new(uid, fsPath, class)
    if not userFeaturesObj then
        return nil, err
    end
    
    
    ---- Wrapper interface ----
    local extFeatures = {}  -- Wrapper object
    -- extend object by wrapper methods
    local obj         = setmetatable(extFeatures, {__index = userFeaturesObj})
    
    --- Saving user features data to file
    function extFeatures:save()
        return self:saveData()
    end
    
    --- Loading user features from file system
    -- @see 'loadData' method for basic class  
    function extFeatures:load(data)
        return self:loadData(data)
    end

    --- Set features value
    -- @see 'setValueData' method for basic class  
    function extFeatures:set(name, value)
        return self:setValueData(name, value)
    end

    --- Retrieve feature value by status and value
    -- @return following values
    --      - nil|false   feature is not activated
    --      - val|true    feature active (may have value)
    -- @see 'getValueData' method in the basic class
    function extFeatures:get(name)
        local rv       = nil
        local featData = self:getValueData(name)
        if featData.status then
            rv = "1" == featData.status and (featData.val or true)
        else
            rv = featData.val
        end
        
        -- set up default value for some features
        if "RING" == name and not rv then
            rv = "Ring1"
        end
        
        return rv
    end

    --- Cheks user features status
    -- @see 'getStatus' method for basic class 
    function extFeatures:is_active(name)
        return self:getStatus(name)
    end
    
    --- Retrieve user features data for PHP/JS front-end
    -- @note data has simle format of the classes for WMS integration
    -- @note do not change user features directly
    function extFeatures:getData4FrontEnd()
        return __featuresFixer(self._data) or {}
    end
    
    --- Processing user features data from PHP/JS front-end
    -- @note executed in case saving feature from WMS
    function extFeatures:getFrontEndData(data)
        return __featuresFixer(data)
    end
    
    --- Compare front end data with actual user features
    -- @param data  The front end data with user features
    -- @return type of the user changes
    function extFeatures:compareFrontEndData(data)
        local rv = {["check-sync"]=false, ["feature-sync"]=false, ["roster"]=false}

        if "table" ~= type(data) then
            return rv
        end
        
        extFeatures:load()  -- loading data from file system
        -- check global features needed for sync
        local actGlobalData = extFeatures:getSingleData()

        repeat
            if actGlobalData["FKEYS"] ~= data["FKEYS"] then
                rv["check-sync"] = true
                break
            end
            if actGlobalData["PHONEBOOK"] ~= data["PHONEBOOK"] then
                rv["check-sync"] = true
                break
            end
            if actGlobalData["SHAREDVM"].status ~= data["SHAREDVM"].status or 
                actGlobalData["SHAREDVM"].value ~= data["SHAREDVM"].value then
                rv["check-sync"] = true
                break
            end
            if actGlobalData["HOTLINE"].status ~= data["HOTLINE"].status or 
               actGlobalData["HOTLINE"].value ~= data["HOTLINE"].value or 
               actGlobalData["HOTLINE"].delay ~= data["HOTLINE"].delay then
                rv["check-sync"] = true
                break
            end
            if actGlobalData["CWTONE"].status ~= data["CWTONE"].status then 
                rv["check-sync"] = true
                break
            end
            if actGlobalData["DATETIME"].format ~= data["DATETIME"].format or 
                actGlobalData["DATETIME"].is24h ~= data["DATETIME"].is24h then
                rv["check-sync"] = true
                break
            end
        until true

        for _, classname in pairs({"internal", "external"}) do
            extFeatures:setUpActiveClass(classname)
            -- check class/sub class activated flag
            local actSubClass           = extFeatures:getActiveSubClass()
            local frontEndSubClassData  = data._classes[classname][actSubClass]
            local frontEndActivatedFlag = "1" == frontEndSubClassData["_ACTIVATED"].status and true or false
            if frontEndActivatedFlag ~= extFeatures:isActivatedClassValues() then
                rv["feature-sync"] = true
            end

            -- check class/sub class features for sync
            local frontEndFeatureStatus = nil
            frontEndSubClassData = data._classes[classname][actSubClass]
            for _, name in pairs({"FCF"}) do
                frontEndFeatureStatus = "1" == frontEndSubClassData[name].status and true or false
                local frontEndFeatureValue  = frontEndSubClassData[name].value
                local actFeatureStatus      = extFeatures:getStatus(name)
                local actFeatureValue       = extFeatures:getValue(name)
                if (frontEndFeatureStatus ~= actFeatureStatus) or 
                   (frontEndFeatureStatus and frontEndFeatureValue ~= actFeatureValue) then
                    rv["feature-sync"] = true
                end
            end
            if frontEndFeatureStatus and frontEndActivatedFlag then
                break
            end
        end

        rv["roster"] = __compare_roster_data(actGlobalData["ROSTER"], data["ROSTER"])

        return rv
    end
    
    return obj
end

--- Return user settings directory path
function userFeaturesExtend:getSettingsDir()
    return require"core.pbxutils".get_users_settings_dir()
end

--- Iterate over user features directory
-- @param fsPath     The file system path where located user features 
-- @return user feature object by every itarate otherwise nil value
function userFeaturesExtend:iterator()
    local userFeaturesList = {}
    local userFeaturesDir  = self:getSettingsDir()
    if userFeaturesDir then
        local lfs = require"lfs"
        for fileName in lfs.dir(userFeaturesDir) do
            if "file" == lfs.attributes(userFeaturesDir.."/"..fileName, "mode") then
                local userIdent = fileName:match("^(%d+)%.lua$")
                table.insert(userFeaturesList, userIdent)
            end
        end
    end
    local idx = 1
    return function()
        local userFeatures = nil
        repeat
            local userIdent = userFeaturesList[idx]
            idx             = idx + 1               -- set next index value
            userFeatures    = self:new(userIdent, fsPath)
            if userFeatures and userFeatures:loadData() then
                break
            end
            
            if not userFeaturesList[idx] then
                break
            end
        until true
        
        return userFeatures
    end
end

return userFeaturesExtend
