--- Module for normalization user features on some version
--
-- @file    userFeaturesNormalize.lua
-- @author  Vyacheslav Khudan
-- @created 2016/05/11

--- Versions desctiption
-- 1 -  Simple user features: one list for one class
-- 2 -  User features with extended lists: all statuses inside one class with one list per status
-- 3 -  New format for FKEYS: BLF with audio/video notifications 

--- @module userFeaturesNormilize
local userFeaturesNormalize = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
--- Make merge user data to tmplt
local function __mergeFeatures(tmpl, data, withDefault)
    local level = 0
    local function _merge(t, d)
        level = level + 1
        for n, v in pairs(t) do
            local vType = type(t[n])
            if type(d[n]) == vType then
                if 1 == level and ("ROSTER" == n or "DEFSTATUS" == n or "CALLCENTER" == n or "CALLCENTER_WHITELIST" == n) then
                    t[n] = d[n]             -- link directly actual user roster to template
                elseif "table" == vType then
                    _merge(t[n], d[n])
                elseif withDefault and "value" == n then    -- using template value in case empty data value
                    if d[n] and "" ~= d[n] then
                        t[n] = d[n]
                    end
                else
                    t[n] = d[n]
                end
            end
        end
        level = level - 1
    end

    return _merge(tmpl, data)
end

--- Normilize user features from version 1 to version 2 (WMS-1379)
local function __fixUp12(data)
    --- Updating sub class values
    local function __updateSubClass(dst, src)
        if "table" ~= type(dst) or "table" ~= type(src) then
            return false
        end

        -- fill default sub class
        dst.default = {}
        for name, val in pairs(src) do
            dst.default[name] = val
        end

        -- update DND sub class from old DND feature value
        if src.DND then
            dst.dnd = {}
            if "1" == src.DND.status then
                dst.dnd["_ACTIVATED"] = { status = "1" }
            end
            if "reject" == src.DND.value then
                dst.dnd["REJECT"] = { status = "1" }
            elseif "silent" == src.DND.value then
                dst.dnd["RING"] = {
                     status = "1"
                    ,value  = "silent"
                }
            end
        end

        return true
    end

    -- set up status values
    data["STATUS"] = {
         value     = data.STATUS
        ,message   = data.STATUS_MESSAGE
        ,untilTime = data.STATUS_UNTIL
        ,location  = data.STATUS_LOCATION
    }

    -- fix up pop-up URL
    if "string" == type(data.POPUPURL) then
        data["POPUPURL"] = {url = data.POPUPURL}
    end

    -- fill user sublcasses by new format
    data._classes = {
         internal  = {}
        ,external  = {}
        ,blacklist = {}
        ,whitelist = {}
    }
    __updateSubClass(data._classes.internal, data.internal)
    __updateSubClass(data._classes.external, data.external)
    __updateSubClass(data._classes.blacklist, data.blacklist)
    __updateSubClass(data._classes.whitelist, data.whitelist)

    return true
end

--- Normilize user features from version 2 to 3 (WMS-1678)
local function __fixUp23(data)
    if not data.FKEYS or "LINE ,LINE " == data.FKEYS then
        return true
    end
    
    local fkeys = {}
    for val in data.FKEYS:gmatch("[^,]+") do
        if val:find("^EXTAV%s") then
            local label = val:find("^EXTAV %d+$") and " " or ""
            val         = val:gsub("EXTAV", "EXT")
            val         = val .. label .. " 0 1 1"
        elseif val:find("^EXTV%s") then
            local label = val:find("^EXTV %d+$") and " " or ""
            val         = val:gsub("EXTV", "EXT")
            val         = val .. label .. " 0 0 1"
        elseif val:find("^EXT%s") then
            if not val:find("^EXT .* %d %d %d$") then       -- skip new format for EXT
                local label = val:find("^EXT %d+$") and " " or ""
                val         = val .. label .. " 0 0 0"
            end
        elseif val:find("^SD%s") then
            if not val:find("^SD .* %d$") then              -- skip new format for SD
                local label = val:find("^SD %d+$") and " " or ""
                val         = val .. label .. " 0"
            end
        elseif val:find("^QCALL%s") then
            if not val:find("^QCALL %d+ %d %d %d$") then    -- skip new format for QCALL
                val = val .. " 0 0 0"
            end
        end
        
        fkeys[#fkeys + 1] = val
    end
    
    data.FKEYS = table.concat(fkeys, ",")
    return true
end

--- Normalize data by removing deprecated data from external systems
local function __fixUp(tmpl, data, withDefault)
    data.uid = nil  -- clear some field in user data

    __mergeFeatures(tmpl, data, withDefault)
    return true
end

local __normalizationList = { -- uprade table
     [1] = __fixUp12
    ,[2] = __fixUp23
}

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Execute user features normalization
-- @param tmplt         The default user features as template
-- @param data          The loaded user features data
-- @param isFileSource  The features source specificator 
-- @return updated user features data structure
function userFeaturesNormalize.run(tmplt, data, isFileSource)
    local rv = false
    if "table" ~= type(data) or "table" ~= type(tmplt) then
        return rv
    end
    if "number" ~= type(tmplt._version) then
        return rv
    end
    
    local fromVer = data._version or 1
    local toVer   = tmplt._version
    
    for idx=fromVer, toVer-1 do     -- processing fix up for old features
        if __normalizationList[idx] then
            __normalizationList[idx](data)
        end
    end
    
    rv = __fixUp(tmplt, data, isFileSource)       -- processing default features

    return rv
end


return userFeaturesNormalize



