--- User status module
--
-- @file    userStatus.lua
-- @author  Vyacheslav Khudan
-- @created 2016/05/27

package.path = package.path .. ";/usr/local/share/lua/5.1/common/userFeatures/?.lua"

--- @module userStatus
local userStatus = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
local function __getUserData(flag, value)
    local rv = nil
    if not flag or not value then
        return rv
    end
    
    if "exten" == flag then
        flag = "telephoneNumber"
    end
    
    local userData  = require"core.wmsldap":get(flag, value)
    if userData and userData[1] and userData[1][2] then
        rv       = userData[1][2]
        rv.exten = rv.telephoneNumber
    end
    
    return rv
end

local function __loadUserFeaturesByExten(exten)
    local rv = nil
    
    -- get user ident
    local userData  = __getUserData("exten", exten)
    local userIdent = userData and userData.uid
    if not userIdent then
        return rv
    end
    
    -- loading user features for updating
    local userFeatures = require"userFeatures":new(userIdent)
    if not userFeatures then
        return rv
    end
    userFeatures:load()
    rv = userFeatures
    return rv
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Saving status data to cache file (user features)
-- @param exten     A related user for status data
-- @param data      The data with status information
-- @return boolean value
function userStatus:saveToCache(exten, data)
    local rv = false
    if not exten or "table" ~= type(data) then
        return rv
    end
    if not data.status then
        return rv
    end
    
    local userFeatures = __loadUserFeaturesByExten(exten)
    if not userFeatures then
        return rv
    end
    
    -- set up new parameter and save features to file system
    if not userFeatures:setUserStatusData({
         value     = data.status
        ,message   = data.message
        ,untilTime = data.untilTime
        ,location  = data.location
    }) then
        return rv
    end
    rv = userFeatures:save()

    -- set up user actual data with new status
    userFeatures:setUpActiveSubClass(data.status)

    -- select external class by FCF
    featureStatus = {}
    for class, status in userFeatures:getStatusOverActiveClasses("FCF") do
        featureStatus[class] = status
    end
    if not featureStatus.internal and featureStatus.external then
        userFeatures:setUpActiveClass("external")
    end

    userFeatures:setUpActiveClassValuesByActivatedFlag()
    
    -- send notification to device(s) for new status
    require"syncUserFeaturesViaSip":byExten(exten, nil, userFeatures)

    return rv
end


--- Reading user status data from cache file (user features)
-- @param exten     The checking user extension (optional)
-- @return status data for one or all user(s)
function userStatus:getFromCache(exten)
    local featIterator = nil
    if exten then   -- iterator for one user
        featIterator = function()
            local userFeatures = __loadUserFeaturesByExten(exten)
            return function()
                local rv     = userFeatures
                userFeatures = nil          -- clear for disabling second iteration
                return rv
            end
        end
        featIterator = featIterator()
    else            -- iterator for all user features
        featIterator = require"userFeatures":iterator()
    end
    
    local rv = {}
    for featObj in featIterator do
        local userIdent = featObj:get("uid")
        local userData  = __getUserData("uid", userIdent)
        local userExten = userData and userData.exten
        if userExten then
            local userStatusData = featObj:getUserStatusData()
            if "table" == type(userStatusData) then
                if userStatusData.value or userStatusData.message or 
                   userStatusData.untilTime or userStatusData.location then
                    rv[#rv + 1] = {
                         exten     = userExten
                        ,value     = userStatusData.value
                        ,message   = userStatusData.message
                        ,untilTime = userStatusData.untilTime
                        ,location  = userStatusData.location
                    }
                end
            end
        end
    end
    
    return exten and rv[1] or rv
end

--- Retrieve real-time users precence data (only status)
-- @param data  The table with request data (like users extensions)
-- @return list with users statuses 
function userStatus:getDataFromJabberbyPresence(data)
    local rv = {}
    if "table" ~= type(data) then
        return rv
    end

    local params = "?users=" .. table.concat(data, ",")
    local hc = require"core.httpClient":new()
    local resp = hc:get("http://127.0.0.1:5280/api/get_presence_status" .. params)
    local userStatuses = require"dkjson".decode(resp and resp.body or "")

    if "table" == type(userStatuses) then
        for _, val in ipairs(userStatuses) do
            if val.user and val.status then
                rv[val.user] = val.status
            end
        end
    end
    
    return rv
end

--- Retrieve user real-time status data
-- @param exten     The requested user extension
-- @return status data
function userStatus:getDataFromJabber(exten)
    local rv = nil
    if not exten then
        return rv
    end

    local hc = require"core.httpClient":new()
    local resp = hc:get("http://127.0.0.1:5280/api/get_user_status?user=" .. exten)
    local statusData = require"dkjson".decode(resp.body) or {}
    rv = { status = statusData.show }
    return rv
end

--- Set up real-time status
-- @param exten     The user extension with status data
-- @param data      The status data for specified user
-- @return boolean value
function userStatus:setDataToJabber(exten, data)
    local rv = false
    if not exten or "table" ~= type(data) then
        return rv
    end
    
    local status = data.status
    if "dnd" ~= status and "away" ~= status and "mur" ~= status then
        status = "available"
    end
    
    -- processing until time
    local untilTime   = nil
    if "available" ~= status and data.time and data.date then
        local hour, min        = data.time:match("(%d%d):(%d%d)")
        local day, month, year = data.date:match("(%d%d)/(%d%d)/(%d%d%d%d)")
        if hour and min and day and month and year then
            local timeRFCStr       = string.format("%s-%s-%sT%s:%s:00Z", year, month, day, hour, min)
            local tdObj            = require"core.timedate":new(timeRFCStr, data.tz)
            if tdObj then
                tdObj:toUTC()
                untilTime = tdObj:format("%FT%TZ")
            end
        end
    end

    -- prepare request data
    local reqData = {
        ["user"] = exten,
        ["show"] = status,
        ["status"] = "",
        ["until"] = untilTime or ""
    }

    local jsondata = require"dkjson".encode(reqData)
    local hclient = require"core.httpClient":new()
    local res = hclient:post("http://127.0.0.1:5280/api/set_user_status", jsondata, {["Content-Type"] = "application/json"})
    return 200 == res.code
end

return userStatus
