--- Module for processing Voice Mail application procedures
--
-- @file	voiceMailApplication.lua
-- @author	Gennadiy Ganchev
-- @created 2015/10/15

--- @module voiceMailApplication
local voiceMailApplication = {}
-- init logger function
local logger = logger or require"core.logger":init("[WMS 3.0]", 3)
local fsutils = require"core.fsutils"
local VOICEMAIL_ROOT = "/var/spool/callweaver/voicemail/"
--[[
[message]
duration=10
flag=
macrocontext=
callerlang=en
origdate=Wed Sep 23 01:01:14 PM UTC 2020
origtime=1600866074
exten=502
rdnis=unknown
category=
msg_id=msg1600866085-1132742934
origmailbox=502
callerchan=SIP/503-0000000b
priority=1
context=users
callerid="Jimmy" <503>
--]]
local missed_template = [[[message]
missed=1
origtime={{time}}
exten={{exten}}
msg_id={{msgid}}
origmailbox={{exten}}
callerid="{{callername}}" <{{callernum}}>
]]
----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
local function __createMissedID(time)
	return "miss" .. tostring(time) .. "-" .. string.sub(tostring(require"socket".gettime() * 10000), -10)
end

local function __getMissedName(path)
	local name = "msg0000"

	if not fsutils.is_dir(path) then
		fsutils.mkdirr(path)
		return name
	end

	local msgFiles = fsutils.readdir(path, "^(.*)%.txt")

	if msgFiles and 0 < #msgFiles then
		return "msg" .. string.sub(tostring(#msgFiles + 10000),-4)
	end
	return name
end

function voiceMailApplication.writeMissedMsg(exten, callernum, callername)
	local time = os.time()
	local data = {
		time = tostring(time),
		exten = exten,
		msgid = __createMissedID(time),
		callername = callername,
		callernum = callernum
	}
	local message = string.gsub(missed_template, "{{(%w+)}}", data)
	local path = VOICEMAIL_ROOT .. exten .. "/INBOX/"
	local filename = __getMissedName(path)
	local fp, err = io.open(path .. filename .. ".txt", "w")
	if not fp then
		return fp, err
	end
	fp:write(message)
	fp:close()
	-- cw stub
	local fp1 = io.open(path .. filename .. ".alaw", "w")
	fp1:close()
	return voiceMailApplication.updateVM(exten, true)
end

--- Update voicemail cached data in callweaver
function voiceMailApplication.updateVM(vmBoxNum, createPeer)
	local send_command = require"core.astman".send_command
	if not vmBoxNum or "" == vmBoxNum then
		logger:error("Voicemailbox number is empty")
		return false
	end
	local res, err
	if createPeer then
		res, err = send_command("sip mwi " .. vmBoxNum)
	else
		res, err = send_command("voicemail update " .. vmBoxNum)
	end
	if not res then -- show logs from astman module
		logger:debug("Voicemailbox %s was not updated: %s", vmBoxNum, err)
	else
		logger:debug("Voicemailbox %s was updated.", vmBoxNum)
	end
	return true
end

--- Send mwi notify to user
local function __mwiNotify(vmBoxNum)
	if not vmBoxNum or "" == vmBoxNum then
		logger:error("Voicemailbox number is empty")
		return false
	end
	local res, err = require"core.astman".send_command("sip mwi " .. vmBoxNum)
	if not res then -- show logs from astman module
		logger:debug("Notify with MWI was not sent: %s", err)
	else
		logger:debug("Notify with MWI was sent")
	end
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Get voicemail messages info
-- @param params: A table with requested parameters
-- @return table with voicemailbox data
function voiceMailApplication.getMsgInfo(params)
	if "table" ~= type(params) then
		return false
	end
	if not params["mailbox"] or "" == params["mailbox"] then
		return false
	end

	local ini_parser = require"inilazy"
	local pathNew = VOICEMAIL_ROOT .. params["mailbox"] .. "/INBOX/"
	local pathOld = VOICEMAIL_ROOT .. params["mailbox"] .. "/Old/"

	local function _create_vm_msg_table(mailbox, msg_type, msgId, msg_files, vm_path, result)
		result 						= result or {}
		for id, msg_file in pairs(msg_files) do
			local msg_detail = ini_parser.get(vm_path .. msg_file.. ".txt")
			if msg_detail and "table" == type(msg_detail) and (not msgId or (msgId and msg_detail.message.msg_id == msgId)) then
				result[msg_detail.message.msg_id] = {
					mailbox          = mailbox,
					type             = msg_type,
					msg_file         = msg_file,
					msgId            = msg_detail.message.msg_id,
					file             = vm_path .. msg_file,
					name             = msg_detail.message.callerid:match('%"(.*)%"'),
					number           = msg_detail.message.callerid:match('%<(.*)%>') or msg_detail.message.callerid,
					origtime         = tonumber(msg_detail.message.origtime),
					duration         = msg_detail.message.duration,
					missed           = msg_detail.message.missed -- missed call not a message
				}
				result[#result + 1] = result[msg_detail.message.msg_id]
			end
		end
		return result
	end
	local result = {}
	local new_msg_files	= nil
	local old_msg_files	= nil
	local newerr		= nil
	local olderr		= nil
	if not fsutils.is_dir(pathNew) then
		fsutils.mkdirr(pathNew)
	end
	if not fsutils.is_dir(pathOld) then
		fsutils.mkdirr(pathOld)
	end
	new_msg_files, newerr = require"core.fsutils".readdir(pathNew, "^(.*)%.txt")
	old_msg_files, olderr = require"core.fsutils".readdir(pathOld, "^(.*)%.txt")
	if new_msg_files and not newerr then
		result = _create_vm_msg_table(params["mailbox"], "new", params["msgId"], new_msg_files, pathNew, result)
	end
	if old_msg_files and not olderr then
		result = _create_vm_msg_table(params["mailbox"], "old", params["msgId"], old_msg_files, pathOld, result)
	end

	if 0 == #result then
		return false, "vmemptylist"
	end

	table.sort(result, function(a, b) return a.origtime > b.origtime end)

	-- limit for phone xml browser
	if params.limit and #result > params.limit then
		local limited_res = {}
		for i = 1, params.limit do
			local item = result[i]
			limited_res[item.msgId] = item
			limited_res[i] = item
		end
		return limited_res
	end

	return result
end

--- Move voicemail message from INBOX to Old directory (mark it as read)
-- @param vmBoxNum	A voicemail box number
-- @param msgId		A voicemail message id to be moved from INBOX
-- @return toMsgId	A voicemail message id moved to Old (will be renamed if a message with the same id exists)
function voiceMailApplication.moveMsg(vmBoxNum, msgId)
	if not vmBoxNum or "" == vmBoxNum or not msgId or "" == msgId then
		return false
	end

	local fromPath = VOICEMAIL_ROOT .. vmBoxNum .. "/INBOX/"
	local toPath = VOICEMAIL_ROOT .. vmBoxNum .. "/Old/"

	local msgFiles, err = fsutils.readdir(toPath, "^(.*)%.txt")
	local toMsgId = "msg0000"

	if msgFiles and 0 < #msgFiles then
		toMsgId = "msg" .. string.sub(tostring(#msgFiles + 10000),-4)
	end

	fsutils.mvfile((fromPath .. msgId .. ".alaw"), (toPath .. toMsgId .. ".alaw"))
	fsutils.mvfile((fromPath .. msgId .. ".txt"), (toPath .. toMsgId .. ".txt"))
	return toMsgId
end

--- Mark the voicemail message as read
-- @param params	A table with voicemail message parameters (vmbox number, message id)
-- @return true or false
function voiceMailApplication.markAsReadMsg(params)
	if "table" ~= type(params) or not params["mailbox"] or "" == params["mailbox"] or not params["msgId"] or "" == params["msgId"] then
		return false, "Error, missing or incorrect input data"
	end
	local msg, err = voiceMailApplication.getMsgInfo(params)
	if not msg then
		return false, err
	end
	if "new" == msg[params["msgId"]]["type"] then
		voiceMailApplication.moveMsg(params["mailbox"], msg[params["msgId"]]["msg_file"])
	else
		return false, "The voicemail message already marked as read"
	end
	return true
end

--- Delete voicemail message
-- @param vmBoxNum: A voicemail box number
-- @param msgId:	A voicemail message id
function voiceMailApplication.deleteMsg(params)
	if "table" ~= type(params) or not params["mailbox"] or "" == params["mailbox"] or not params["msgId"] or "" == params["msgId"] then
		return false
	end
	local msg	= nil
	local err	= nil
	local file	= params["file"]
	if not file then
		msg, err = voiceMailApplication.getMsgInfo(params)
		if not msg then
			return false, err
		end
		file = msg[params["msgId"]]["file"]
	end
	fsutils.rmfile(file .. "*")
	fsutils.rmfile(VOICEMAIL_ROOT .. params["mailbox"] .. "/INBOX/" .. params["msgId"] .. "*")
	return true
end


--- Forward voicemail message
-- @param params	A table with voicemail message parameters (vmbox number, message id)
-- @param vmBoxTo	Destination mailbox
function voiceMailApplication.fwdMsg(params, vmBoxTo)
	if "table" ~= type(params) or not params["mailbox"] or "" == params["mailbox"] or not params["msg_file"] 
		or "" == params["msg_file"] or not params["msgId"] or "" == params["msgId"] or not vmBoxTo or "" == vmBoxTo then
			return false, "wrong args"
	end

	local fromPath = VOICEMAIL_ROOT .. params["mailbox"] .. ("new" == params["type"] and "/INBOX/" or "/Old/")
	local toPath = VOICEMAIL_ROOT .. vmBoxTo .. "/INBOX/"
	local msgFiles, err = fsutils.readdir(toPath, "^(.*)%.txt")
	local toMsgId = "msg0000"

	if msgFiles and 0 < #msgFiles then
		toMsgId = "msg" .. string.sub(tostring(#msgFiles + 10000), -4)
	end

	logger:debug([[VmFwd. Copy from "%s" to "%s"]], (fromPath .. params["msg_file"] .. ".alaw"), (toPath .. toMsgId .. ".alaw"))
	local rv, err = fsutils.cpfile((fromPath .. params["msg_file"] .. ".alaw"), (toPath .. toMsgId .. ".alaw"))
	if rv then
		logger:debug([[VmFwd. Copy from "%s" to "%s"]], (fromPath .. params["msg_file"] .. ".txt"), (toPath .. toMsgId .. ".txt"))
		rv, err = fsutils.cpfile((fromPath .. params["msg_file"] .. ".txt"), (toPath .. toMsgId .. ".txt"))
	end

	if rv then
		if (VOICEMAIL_ROOT .. params["mailbox"] .. "/INBOX/") ~= toPath then
			logger:debug([[VmFwd. Copy from "%s" to "%s"]], (VOICEMAIL_ROOT .. params["mailbox"] .. "/INBOX/" .. params["msgId"] .. "*"), toPath)
			rv, err = fsutils.cpfile((VOICEMAIL_ROOT .. params["mailbox"] .. "/INBOX/" .. params["msgId"] .. "*"), toPath)
		end
	end

	if rv then
		ini_parser = require"inilazy"
		logger:debug("VmFwd. Update ini-file:", toPath .. toMsgId .. ".txt")
		fixMsg, err = ini_parser.get(toPath .. toMsgId .. ".txt")
		if fixMsg and fixMsg["message"] then
			if fixMsg["message"]["callerid"] and [["]] == fixMsg["message"]["callerid"]:sub(1, 1) then
				fixMsg["message"]["callerid"] = [["Fwd: ]] .. fixMsg["message"]["callerid"]:sub(2)
			else
				fixMsg["message"]["callerid"] = [[Fwd: ]] .. fixMsg["message"]["callerid"]
			end
			rv, err = ini_parser.set(fixMsg, toPath .. toMsgId .. ".txt")
		else
			rv = false
		end
	end

	if rv then
		local dialplanGeneralSetting = require"core.pbxutils":getDialplanGeneralSettings() or {}
		local attach_fmt = dialplanGeneralSetting["GATTACH_FORMAT_VM"] or "mp3"
		local command = "nohup wms_notify -vm %q %q %q %q %q > /tmp/notify.out 2>&1 &"
		logger:debug(string.format("VmFwd. Execute: " .. command, vmBoxTo, toPath .. toMsgId, 'alaw', attach_fmt, params["msgId"]))
		require"core.pbxutils".exec(command, vmBoxTo, toPath .. toMsgId, 'alaw', attach_fmt, params["msgId"])
	end

	return rv, err
end

return voiceMailApplication
