--- Client implementation for callweaver manager interface
-- 
-- @file    AMIClient.lua 
-- @author  Vyacheslav Khudan
-- @created 2017/02/23

--@module AMIClient
local AMIClient = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
-- Set metatable method with working __gc functionality
local function exsetmetatable(t, mt)
    local proxy              = newproxy(true)
    getmetatable(proxy).__gc = function() mt.__gc(t) end
    t[proxy]                 = true
    return setmetatable(t, mt)
end

----------------------------------------------------------------
-- AMI driver implementation
----------------------------------------------------------------
local driver = {
     sockConn = nil      -- connection socket
    ,timeout  = 10       -- timeout for socket operations
}

function driver:getInfo()
    local rv = {}
    if self.sockConn then
        rv.sockIdent                 = tostring(self.sockConn)
        rv.localAddr, rv.localPort   = self.sockConn:getsockname()
        rv.remoteAddr, rv.remotePort = self.sockConn:getpeername()
    end
    
    return rv
end

function driver:isConnected()
    local rv = self.sockConn and self.sockConn:getsockname() or nil
    return nil ~= rv 
end

function driver:connect(host, port)
    if not host or not port then
        return false, "Invalid connection data"
    end
    
    if self.sockConn then
        return true
    end
    
    local socket = require"socket".tcp()
    socket:settimeout(self.timeout)
    local status, err = socket:connect(host, port)
    if not status then
        return false, err
    end
    
    self.sockConn = socket
    return true
end

function driver:disconnect()
    if self.sockConn then
        self.sockConn:close()
        self.sockConn = nil
    end
end

function driver:send(data)
    if not data or not self.sockConn then
        return nil, "No connection for sending"
    end
    return self.sockConn:send(data)
end

function driver:recieve()
    local rv  = nil
    local err = nil
    if not self.sockConn then
        return rv, "No connection for recieving"
    end
    
    local data, err = self.sockConn:receive()
    -- decrease timeout to retrieve data
    self.sockConn:settimeout(0.003)
    rv = ""
    while data do
        rv = rv .. "\n" .. data
        data, err = self.sockConn:receive()
    end
    self.sockConn:settimeout(self.timeout)
    
    return rv, err
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local client = {   -- meta class
     driver    = nil     -- driver for AMI connections
    ,cfg       = nil     -- configuration to make connection
    ,keepAlive = false   -- flag about persistent or non-persistent connection
}

--- Builder for AMI messages
function client.__cmdBuilder(cmdList)
    if "table" ~= type(cmdList) then
        return nil
    end
    
    local cmd = {}
    for k, v in pairs(cmdList) do
        if k and v then
           cmd[#cmd + 1] = k .. ": " .. v 
        end
    end
    
    cmd = table.concat(cmd, "\r\n")
    cmd = cmd .. "\r\n\r\n"
    return cmd
end

--- Default responce processor
-- @return boolean value if request successful 
function client.__defRespProcessor(data)
    local rv = false
    if data and data:find("Response: Success") then
        rv = true
    end
    return rv
end
 
--- Simple sender data without any smart logic
-- @param cmd               The command data 
-- @param respProcessor     The responce data processor
-- @return recieved data from responce processor 
function client:__send(cmd, respProcessor)
    cmd = self.__cmdBuilder(cmd)
    if not cmd then
        return nil, "Invalid command value for sending"
    end
    
    local sentData, err = self.driver:send(cmd)
    if not sentData and err then
        return nil, err
    end
    local data, err = self.driver:recieve()
    if not data and err then
        return nil, err
    end
    local rv = nil
    if "function" == type(respProcessor) then
        rv = respProcessor(data)
    else
        rv = self.__defRespProcessor(data)
    end
    return rv
end

--- Initialize client with login procedure
-- @return boolean status of the procedure execution 
function client:init()
    if not self.driver then
        return false, "Driver not found"
    end
    
    if self.driver:isConnected() then
        return true
    end
    
    -- make connection to socket via driver
    local rv, err = self.driver:connect(self.cfg.host, self.cfg.port)
    if not rv then
        return false, err
    end
    
    -- sending login command
    local loginCMD = {
         Action   = "Login"
        ,Events   = "Off"
        ,Username = self.cfg.user
        ,Secret   = self.cfg.pass
    }
    return self:__send(loginCMD)
end

--- Retrieve info about socket connection
-- @note actual only for persistent connection 
function client:getInfo()
    local rv   = nil
    local info = self.driver:getInfo()
    if info then
        rv   = { ident  = info.sockIdent }
        if info.localAddr then
            rv.client = info.localAddr .. ":" .. info.localPort
        end
        if info.remoteAddr then
            rv.server = info.remoteAddr .. ":" .. info.remotePort
        end
    end
    return rv
end

--- Sending data to AMI with following responce processing
-- @see client:__send
function client:send(cmd, respProcessor)
    -- make connection initialization for non-persistent mdoe
    if not self:init() then
        return nil, "Client login command failed"
    end
    
    local rv, err = self:__send(cmd, respProcessor)
    
    -- try to destroy connection
    if not self.keepAlive then
        self:destroy()
    end
    
    return rv, err
end

--- Destroy AMI client object with closing related connections
function client:destroy()
    return self.driver:disconnect()
end
----------------------------------------------------------------
-- Available commands
----------------------------------------------------------------
--- Execute Asterisk CLI Command
function client:sendCommand(cmd)
    if not cmd or "" == cmd then
        return nil
    end
    local cmdCMD = {
         Action  = "Command"
        ,Command = cmd
    }
    return self:send(cmdCMD)
end

--- Get the value of a global or local channel variable
-- @note to retrieve gloval variable you have to clear channel 
function client:getVar(var, channel)
    if not var then
        return nil
    end
    
    local cmdGetVar = {
         Action   = "Getvar"
        ,Variable = var
    }
    if channel then
        cmdGetVar.Channel = channel
    end
    
    local processor = function(data)
        if data:find("Response: Success") then
            return data:match("Value: ([^\r\n]+)")
        end
    end
    return self:send(cmdGetVar, processor)
end

--- Set a global or local channel variable
-- @note to retrieve gloval variable you have to clear channel 
function client:setVar(var, value, channel)
    if not var or not value then
        return nil
    end
    
    local cmdSetVar = {
         Action   = "Setvar"
        ,Variable = var
        ,Value    = value
    }
    if channel then
        cmdSetVar.channel = channel
    end
    
    return self:send(cmdSetVar)
end

--- Update channel information with following SIP UPDATE message
-- @param channel   The channel name for updating
-- @param data      The data for updating 
-- @return boolean value as a operation status
function client:updateChannelInfo(channel, data)
    if not channel then
        return false
    end
    
    local cmdUpdChanInfo = {
         Action  = "SIPUpdatePAI"
        ,Channel = channel
    }
    if data then
        if data.remoteNum then
            cmdUpdChanInfo.ConnectedLineNum = data.remoteNum
        end
        if data.remoteName then
            cmdUpdChanInfo.ConnectedLineName = data.remoteName
        end
    end
    return self:send(cmdUpdChanInfo)
end

--- Processing DB Put with callweaver database
function client:putDBQuery(key, val, family)
    if not key or not val or not family then
        return false
    end
    
    local cmdPutDB = {
         Action = "DBPut"
        ,Family = family
        ,Key    = key
        ,Val    = val
    }
    return self:send(cmdPutDB)
end
    
--- Processing DB Get with callweaver database
function client:getDBQuery(key, family)
    if not key or not family then
        return false
    end
    
    local cmdGetDB = {
         Action = "DBGet"
        ,Family = family
        ,Key    = key
    }
    
    local processor = function(data)
        if data:match("DBGetResponse") then
            return data:match("Val:%s([^\n]+)")
        end
    end
    return self:send(cmdGetDB, processor)
end

--- Processing DB Del with callweaver database
function client:delDBQuery(key, family)
    if not key or not family then
        return false
    end
    
    local cmdDelDB = {
         Action = "DBDel"
        ,Family = family
        ,Key    = key
    }
    return self:send(cmdDelDB)
end

--- Processing DB DelTree to callweaver database
function client:delTreeDBQuery(family)
    if not family then
        return false
    end
    
    local cmdDelTreeDB = {
         Action = "DBDelTree"
        ,Family = family
    }
    return self:send(cmdDelTreeDB)
end

--- Processing adding an agent to a callgroup
function client:addQueueMember(queue, member)
    local cmdGetVar = {Queue      = queue,
                       Penalty    = 1,
                       Paused     = no,
                       Interface  = "Local/" .. member .. "@internalcalls",
                       MemberName = "Local/" .. member .. "@internalcalls",
                       Action    = "QueueAdd"
                      }
    return self:send(cmdGetVar)
end

--- Processing removing an agent from a callgroup
function client:removeQueueMember(queue, member)
    local cmdGetVar = {Queue     = queue,
                       Interface = "Local/" .. member .. "@internalcalls",
                       Action    = "QueueRemove"
                      }
    return self:send(cmdGetVar)
end

--- Processing originate request
function client:originate(chan, context, exten, priority, application, data, clid, timeout, variables, async)
    local cmdGetVar = {Action = "Originate"}
    local errors = {}

    if (context and exten) and (not application and not data) then 
        if "" ~= context and "" ~= exten then
            cmdGetVar["Context"] = context
            cmdGetVar["Exten"] = exten
            if priority and tonumber(priority) then
                cmdGetVar["Priority"] = tostring(priority)
            else
                cmdGetVar["Priority"] = "1"
            end
        else
            table.insert(errors, "empty context or exten")
        end
    elseif (not context and not exten and not priority) and (application and data) then
        if "" ~= applicationthen then
            cmdGetVar["Application"] = application
        else
            table.insert(errors, "empty application")
        end
        if data and "" ~= data then
            cmdGetVar["Data"] = data
        else
            table.insert(errors, "empty data")
        end
    else
        table.insert(errors, "wrong destination")
    end
    
    if not chan or "" == chan then
        table.insert(errors, "no channel specified")
    else
        cmdGetVar["Channel"] = chan
    end

    if timeout and tonumber(timeout) then
        cmdGetVar["Timeout"] = tostring(timeout)
    else
        cmdGetVar["Timeout"] = "30000"
    end

    if not clid or "string" ~= type(clid) or "" == clid then
        table.insert(errors, "no CallerId specified")
    else
        cmdGetVar["CallerID"] = clid
    end

    if variables and "" ~= variables then
        cmdGetVar["Variable"] = variables
    end

    if async then
         cmdGetVar["Async"] = "true"
    end

    if #errors ~= 0 then
        return false, table.concat(errors, ",")
    end

    processor = function (data)
        if data and data:find("Response: Success") then
            return true
        end
        return tostring(data)
    end

    response = self:send(cmdGetVar, processor)
    if response and "boolean" ~= type(response) then
        return false, response
    end

    return true
end

--- Processing pause monitor request
function client:pauseMonitor(channel)
    if not channel or "string" ~= type(channel) or "" == channel then
        return false, "Invalid channel"
    end
    local cmdGetVar = {Action  = "PauseMonitor",
                       Channel = channel
                      }
    return self:send(cmdGetVar)
end

--- Processing unpause monitor request
function client:unpauseMonitor(channel)
    if not channel or "string" ~= type(channel) or "" == channel then
        return false, "Invalid channel"
    end
    local cmdGetVar = {Action  = "UnpauseMonitor",
                       Channel = channel
                      }
    return self:send(cmdGetVar)
end

--- Object constructor
function AMIClient:new(keepAlive)
    -- loading client configuration
    local cfgPath = "/etc/callweaver/manager.conf"
    local cfg, err = require"inilazy".get(cfgPath)
    if "table" ~= type(cfg) then
        return nil, "Can't parse configuration for AMI client: " .. err
    end
    local objCfg = {
         user = "root"
        ,pass = cfg["root"] and cfg["root"].secret or "wildix"
        ,host = cfg["general"] and cfg["general"].bindaddr or "127.0.0.1"
        ,port = cfg["general"] and cfg["general"].port or "5039"
    }

    -- initialize object
    local mt = {
         __index = client
        ,__gc    = client.destroy
    }
    local obj     = exsetmetatable({}, mt)
    obj.cfg       = objCfg
    obj.driver    = driver
    obj.keepAlive = keepAlive
    if keepAlive and not obj:init() then
        return nil, "Client login command failed"
    end
    
    return obj
end

return AMIClient
