--- Simple file system library. Show information about the file system
-- 
-- @file	diskutils.lua 
-- @author	Vyacheslav Khudan
-- @created 2014/07/16	

local io = io
local type = type
--local string = string
local posix = require"posix"

module(... or "diskutils")

--- Check mounted partition
-- @param partition	A partition mount point
-- @return status True or False and error message
function is_mount (partition)
	if not partition or partition == "" then
		return false, "Invalid argument"
	end
	
	local mtab, err_msg = io.open("/etc/mtab", "r")
	if not mtab then
		return false, err_msg
	end
	
	local content = mtab:read("*all")
	mtab:close()
	
	for line in content:gmatch("[^\r\n]+") do
		if line:match(partition) then
			return true
		end
	end
	
	return false, "Disk not found"
end

--- Show information about mount disk space
-- @param mem_type	Type of the memory information (total, available, used)
-- @param partition A partition mount point
-- @return size of the disk free space in Kbytes
function get_space (mem_type, partition)
	local space_value = 0
	if "total" ~= mem_type and "avail" ~= mem_type and "used" ~= mem_type then
		return space_value, "Invalid type value"
	end
	if not partition or partition == "" then
		return space_value, "Invalid argument"
	end
	
	local disk_found, err_msg = is_mount(partition)
	if not disk_found then
		return space_value, err_msg
	end
	
	local statvfs, err_msg = posix.statvfs(partition)
	if not statvfs or "table" ~= type(statvfs) then
		return space_value, err_msg
	end
	
	if "total" == mem_type then
		space_value = statvfs.frsize * statvfs.blocks
	elseif "avail" == mem_type then
		space_value = statvfs.frsize * statvfs.bavail
	elseif "used" == mem_type then
		space_value = statvfs.frsize * (statvfs.blocks - statvfs.bfree)
	end
	
	return space_value/1024
end

--- Show available space of the RAM disk
-- @return size of the available space in Kbytes
function get_avail_ram_space ()
	local ram_partition = "/dev/shm"
	return get_space("avail", ram_partition)
end

--- Check storage is configured
-- @param storage	A storage mount point
-- @return status True or False and error message
function is_configured(storage)
	if not storage or "" == storage then
		return false, "Invalid argument"
	end
	
	local share, err_msg = io.open("/rw2/etc/wms.conf", "r")
	if not share then
		return false, err_msg
	end
	
	local content = share:read("*all")
	share:close()
	
	for line in content:gmatch("[^\r\n]+") do
		if line:match(storage) then
			return true
		end
	end
	
	return false, "Storage is not configured"
end