--- Simple file system library (operations with file and directories).
-- 
-- @file	fsutils.lua 
-- @author	Vyacheslav Khudan
-- @created 2014/07/17

local io = io
local os = os
local type = type
local table = table
local string = string
local tostring = tostring
local tonumber = tonumber
local pairs = pairs
local select = select
local require = require

module(... or "fsutils")

--- Format input arguments
local function __fmtarg (arg1, ...)
	if (...) then
		arg1 = string.format(arg1, ...)
	end
	
	return arg1
end

--- Get type of the input path
local function __get_type (path, without_link_info)
	if not path or "" == path then
		return false, "Invalid path"
	end
	
	if without_link_info then
		return require"lfs".attributes(path, "mode")
	end
	return require"lfs".symlinkattributes(path, "mode")
end

--- Copy source file to destination file
function __copy(src_file, dst_file)
	if not src_file or "" == src_file or not dst_file or "" == dst_file then
		return false, "Invalid argument"
	end
	
	local command = string.format("cp %s %s 2>&1", src_file, dst_file)
	local handle, exec_error = io.popen(command)
	if not handle then
		local exec_msg = string.format("Can not execute '%s'", command)
		local err_msg  = "string" == type(exec_error) and exec_msg .. ": " .. exec_error or exec_msg
		return handle, err_msg
	end
	local rv     = handle:read("*a")
	handle:close()
	
	local err    = 0 ~= rv:len() and rv:gsub("\n", "") or nil
	rv           = not err and true or false
	return rv, err
end

------------ File system checkers --------------
--- Check if specified path is directory
-- @return boolean True or False
function is_dir (path)
	if "directory" == __get_type(path, true) then
		return true
	end
	return false
end

--- Check if specified path is symbolic link
-- @return boolean True or False
function is_link (path)
	if "link" == __get_type(path) then
		return true
	end
	return false
end

--- Check if specified path is file
-- @return boolean True or False
function is_file (path)
	if "file" == __get_type(path) then
		return true
	end
	return false
end

--- Check if a file can be read with any extension which passed
-- @param filename	A string with absolute path to file
-- @param mixed		A table with extenions of the file
-- @return boolean True or False
-- @note quickly than 'get_type'
function is_readable (filename, ...)
	local texts = select(1, ...)
	
	local filename = filename or ""
	if filename == "" then
		return false
	end
	
	local file
	if type(texts) == "table" and #texts > 0 then
		for _, ext in pairs(texts) do
			file = io.open(filename .. "." .. ext)
			if file and io.close(file) then
				return true
			end
		end
		return false
	else
		 file = io.open(filename)
		 return file ~= nil and io.close(file)
	end
end

-------------- File operations -----------------
--- Copy file (can be simple regexp, like file*) to destination file or folder
-- @param src_name A source file name for the copy operation
-- @param dst_path A destination file or folder
-- @return boolean True or False with error message
-- @note Has long time execution vs shell cp
function cpfile (src_name, dst_path)
	if not src_name or "" == src_name or not dst_path or "" == dst_path then
		return false, "Invalid argument"
	end
	
	if src_name:find("%*") then -- regexp present
		if not is_dir(dst_path) then
			return false, "Can't copy multi file to file"
		end
		
		local err = nil
		local res_files = require"posix".glob(src_name)
		if res_files then
			for _, file in pairs(res_files) do
				if is_file(file) then
					local dst_name = ("/" .. file):match(".*%/(.*)")
					local copy_res, err_msg = __copy(file, dst_path .. "/" .. dst_name)
					if not copy_res then
						err = (err and err .. "\n" or "") .. err_msg
					end
				else
					err = string.format("%s: Not a file", src_name)
				end
			end
		else
			err = string.format("%s: No file found", src_name)
		end
		if err then
			return false, err 
		end
    	return true
	else -- simple file
		if is_readable(src_name) then
    		if is_dir(dst_path) then -- copy to directory
        		local dst_name = ("/" .. src_name):match(".*%/(.*)")
        		return __copy(src_name, dst_path .. "/" .. dst_name)
        	else
        		return __copy(src_name, dst_path)
        	end
    	else
    		return false, string.format("%s: Not a file", src_name)
    	end
	end
	
	return false, "Unknown error"
end

--- Remove files. Support simple regular expression (ex. name.*)
-- @param file	A file to be deleted
-- @param mixed	A arguments for string formatted file name
-- @return boolean True or False with error message 
-- @limitation Can't remove multiply files by glob searching if name contents asterisk symbol
function rmfile (file, ...)
	file = __fmtarg(file, ...)
	if not file then
		return false, "Invalid file(s)"
	end
	
	local isOneFile = true
	
	if file:find("%*") then -- regexp present
		local err 		 = nil
		local foundFiles = require"posix".glob(file)
		if "table" == type(foundFiles) then
			isOneFile = false
			for _, f in pairs(foundFiles) do
        		if not is_dir(f) then
            		local rmRes, errMsg = os.remove(f)
            		if not rmRes then
            			err = (err and err.."\n" or "")..errMsg  
            		end
            	end
        	end
        	if err then
        		return false, err 
        	end
        	return true
    	end
    end
    
    if isOneFile then	-- simple remove
		return os.remove(file)
	end
end

--- Move file
-- @param old_file	An old file which shold be moved
-- @param new_file	A destination file
-- @return boolean True or False with error message
function mvfile (old_file, new_file)
	if not old_file or "" == old_file or not new_file or "" == new_file then
		return false, "Invalid argument"
	end
	
	if not os.rename(old_file, new_file) then
    	local res, err_msg = cpfile(old_file, new_file)
    	if not res then
    		return res, err_msg
    	end
    	return rmfile(old_file)
    end
    return true
end

--- Change the mode of the path. 
-- Modes are specified in one of the following formats: "rwxrwxrwx" (e.g. "rw-rw-w--") "ugoa+-=rwx" (e.g. "u+w") "+-=rwx" (e.g. "+w")
-- @param path A existing file path
-- @param mode A file mode
-- @return boolean True or False with error message
function chmod (path, mode)
	if not path or "" == path or not mode or "" == mode then
		return false, "Invalid argument"
	end
	
	return require"posix".chmod(path, mode)
end

--- Change ownership of a file
-- @param path      The full path of the file
-- @param userName  The user name
-- @return boolean value as a status
function chown(filePath, userName)
    if not filePath or not userName then
        return nil, "Invalid arguments"
    end

    local userData = require"posix".getpasswd(userName)
    if "table" ~= type(userData) or not userData.uid or not userData.gid then
        return nil, string.format("Invalid/not found user - %s", userName)
    end

    return require"posix".chown(filePath, userData.uid, userData.gid)
end

--- Get size of the file
-- @param path A target file path
-- @return Size in bytes and error message if present
function filesize (path)
	if not path or "" == path then
		return 0, "Invalid argumets"
	end
	
	local res_size, error = require"lfs".symlinkattributes(path, "size")
	return tonumber(res_size) or 0, error
end

--- Get modification time
-- @param path A target path
-- @return Modification time and error message if present
function getmTime(path)
	if not path or "" == path then
		return 0, "Invalid argumets"
	end
	
	local modificationTime, error = require"lfs".symlinkattributes(path, "modification")
	return tonumber(modificationTime), error
end

--- Set up modification time
-- @param path  A target file path
-- @param mTime A motification time value
-- @return TRUE value otherwise FALSE with error message if present
-- @note Update access time to the same value as modification time
function setmTime(path, mTime)
    if not path or "" == path then
        return false, "invalid argumets"
    end
    
    local mTime = tonumber(mTime) or os.time()
    return require"lfs".touch(path, mTime, mTime)
end

--- Generate temporary file on target directory
-- @param target A string with target path to generate tmpfile
-- @return A string with temporary filename
function tmpname (target)
	if not target then
		return nil, "Invalid args"
	end
	
	local tmpname = os.tmpname()
	local tmpname_new = tostring(target) .. tmpname:sub(6) -- skip /tmp/
	return os.rename(tmpname, tmpname_new) and tmpname_new or tmpname
end

----------- Directory operations ---------------
--- Simple make directory
-- @return boolean True or False with error
function mkdir (path)
	if not path or "" == path then
		return false, "Invalid path"
	end
	
	return require"lfs".mkdir(path)
end

--- Recursive make directory
-- @return boolean True or False with error
function mkdirr (path)
	if not path or "" == path then
		return false, "Invalid path"
	end
	
	local is_current_dir = true
	local found_delim = path:find("/")
	if 1 == found_delim then
		is_current_dir = false
	end
	 
	if found_delim then
		local sub_path = is_current_dir and "" or "/"
		local err = nil
		for dir in path:gmatch("[^/]+") do
			sub_path = sub_path .. dir
			_, err = mkdir(sub_path)
			if err and not err:match("File exists") then
				return false, err
			end
			sub_path = sub_path .. "/" 
		end
		return true
	end
	return mkdir(path)
end

--- Simple remove directory
-- @return boolean True or False with error
function rmdir (path)
	if not path or "" == path then
		return false, "Invalid path"
	end
	return require"lfs".rmdir(path)
end

--- Recursive remove directory
-- @return boolean True or False with error
function rmdirr (path)
	if not path or "" == path then
		return false, "Invalid path"
	end
	
	if "directory" ~= __get_type(path) then
		return false, string.format("Directory %q not found", path)		
	end
	
	local err = nil
	for dirfile in require"lfs".dir(path) do
		if dirfile and "." ~= dirfile and ".." ~= dirfile then
			local err_msg = nil
			local res_dirfile = path .. "/" .. dirfile
			if is_dir(res_dirfile) then
				_, err_msg = rmdir(res_dirfile)
				if err_msg and err_msg:find("Directory not empty") then
					err_msg = nil
					_, err_msg = rmdirr(res_dirfile)
				end
			else
				_, err_msg = os.remove(res_dirfile)
			end
			if err_msg then
				err = (err and err .. "\n" or "") .. err_msg
			end
		end
	end
	
	local _, err_msg = rmdir(path)
	if err_msg then
		err = (err and err .. "\n" or "") .. err_msg
	end
	
	if err then
		return false, err
	end
	return true
end

--- Getting list of the file system objects in specified directory by some mode
-- @param dir_path	  A target directory for listing
-- @param regexp	  A regular expression for matching objects
-- @param mode		  A mode of the found object (file, link, dir). Default: file + link
-- @param first_match Return result if first file is mathced
-- @return Found filenames or nil with errors
function readdir (dir_path, regexp, mode, first_match)
	if not dir_path or "" == dir_path then
		return nil, "Invalid argument"
	end
	
	if not is_readable(dir_path) then
		return nil, "Destination directory is not readable"
	end
	
	local res_list = {}
	for file in require"lfs".dir(dir_path) do
		if file and "." ~= file and ".." ~= file then
			local check_mode = false
			local match_name = nil
			if regexp and regexp ~= "" then
				if "table" == type(regexp) then
    				for _, vregexp in pairs(regexp) do
						match_name = file:match(vregexp)
						if match_name then break end
    				end
    			else
					match_name = file:match(regexp)
				end
				if match_name then
					check_mode = true
				end
			else
				check_mode = true
			end
    		
    		if check_mode then
    			local obj_mode = nil
    			if "file" == mode or "link" == mode then  
    				obj_mode = require"lfs".symlinkattributes(dir_path.."/"..file, "mode")
    			elseif not mode or "dir" == mode then
    				obj_mode = require"lfs".attributes(dir_path.."/"..file, "mode")
    			end
    				
    			if mode == obj_mode or (not mode and ("file" == obj_mode or "link" == obj_mode)) then
    				if match_name then
    					table.insert(res_list, match_name)
    					if first_match then
    						break
    					end
    				else
    					table.insert(res_list, file)
    				end
        		end
        	end
		end
	end
	
	if #res_list > 0 then
		local sort_func = function(a,b) return a < b end
		table.sort(res_list, sort_func)
		return res_list
	end
	return nil, "Directory doesn't have files"
end

--- Remove files by prefix
-- dir    A directory where the files are located (for example "/tmp/")
-- prefix A prefix for cleaning files names started from (for example tts_ed2b5c0139cec8ad2873829dc1117d50.wav)
-- ttl    A time to live cache in seconds - files created earlier the ttl will be deleted
function rmFilesByPrefix(dir, prefix, ttl)
	if not prefix then
		return false
	end
	dir = dir or "/tmp/"
	ttl = ttl or 0
	local fileList = readdir(dir, "^("..prefix..".*)")
	if not fileList or "table" ~= type(fileList) then
		return false
	end
	local currentTime = os.time()
	for _,file in pairs(fileList) do
		local fileModTime = getmTime(dir..file)
		if fileModTime and currentTime - fileModTime >= ttl then
			rmfile(dir..file)
		end
	end
	return fileList
end
