--- HTTP request engine based executing curl via callweaver SHELL function.
--  Basically used for downloading/uploading files.
-- 
-- @file    callweaverShellCURLBackend.lua 
-- @author  Vyacheslav Khudan
-- @created 2017/11/29
-- @note errors otput for curl execution not supported for this backend

local supportedMethods = {
         GET    = true
        ,POST   = true
}

local cwShellCURLBackend = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
--- Building HTTP request
local function __buildRequest(opts, method)
    opts             = opts or {}
    local timeout    = 60
    local userAgent  = "Mozilla/5.0"
    local authData   = nil
    local authMethod = "basic"  -- default
    local file       = nil
    local content    = nil
    -- processing user agent
    if opts.userAgent then
        userAgent = opts.userAgent
    end
    
    -- processing timeout value
    local optsTimeout = tonumber(opts.timeout) 
    if optsTimeout then
        timeout = optsTimeout
    end
    
    -- processing authorization data
    if opts.authUser and opts.authPass then
        authData = string.format("%s:%s", opts.authUser, opts.authPass)
        if "digest" == opts.authMethod then
            authMethod = "digest"
        end
    end
    
    -- processing request type to a file downloading
    if "GET" == method then
        if opts.inFile then
            file = {
                 direction = "incoming"
                ,path      = opts.inFile
            }
        end
    -- processing requests type for a file uploading
    elseif "POST" == method then
        if opts.outFileByForm then
            file = {
                 direction = "outgoing"
                ,mode      = "form"
                ,field     = "file"
                ,path      = opts.outFileByForm
            }
        elseif opts.outFileByStream then
            file = {
                 direction = "outgoing"
                ,mode      = "stream"
                ,path      = opts.outFileByStream
            }
        end
        if opts.content then
            content   = opts.content
            local url = require"socket.url"
            if "table" == type(content) then
                local content_t = {}
                for key, value in pairs(content) do
                    if key and value then
                        if "table" == type(value) then
                            for _, subvalue in pairs(value) do
                                table.insert(content_t, key .. "=" .. require"core.httpClient".urlEncode(subvalue))
                            end
                        else
                            table.insert(content_t, key .. "=" .. require"core.httpClient".urlEncode(value))
                        end
                    end
                end
                content = content_t
            end
        end
    end

    local randValue
    if opts.cw_hack then
        math.randomseed(os.time())
        randValue = tostring(math.random(1000000000))
    end

    return {
         timeout    = timeout
        ,userAgent  = userAgent
        ,authData   = authData
        ,authMethod = authMethod
        ,file       = file
        ,method     = method
        ,content    = content
        ,ssl        = {insecure = opts.insecure}
        ,cw_hack    = randValue
    }
end

--- Building CURL command for execution via shell
local function __buildCURLCommand(reqData, URLData)
    if not reqData or not URLData then
        return false
    end
    local cmdTable          = {"/usr/bin/curl -D/dev/stdout -sS"}
    cmdTable[#cmdTable + 1] = string.format("'%s'", URLData.target)
    if reqData.ssl.insecure then
        cmdTable[#cmdTable + 1] = string.format("-k")
    end    
    if reqData.method then
        cmdTable[#cmdTable + 1] = string.format("-X %s", reqData.method)
    end
    if reqData.userAgent then
        cmdTable[#cmdTable + 1] = string.format("-A '%s'", reqData.userAgent)
    end
    if reqData.authData then
        cmdTable[#cmdTable + 1] = string.format("--%s -u '%s'", reqData.authMethod, reqData.authData)
    end
    if reqData.timeout then
        cmdTable[#cmdTable + 1] = string.format("--connect-timeout %d", reqData.timeout)
    end

    if reqData.content then
        if reqData.file and "form" == reqData.file.mode then
            for _, v in ipairs(reqData.content) do
                cmdTable[#cmdTable + 1] = string.format("-F '%s'", v)
            end
        end
    end
    
    if reqData.file then
        if "incoming" == reqData.file.direction then
            local fd, err = io.open(reqData.file.path, "w")
            if not fd then
                return nil, err
            end
            fd:close()
            cmdTable[#cmdTable + 1] = "-o " .. reqData.file.path
        elseif "outgoing" == reqData.file.direction then
            local fd, err = io.open(reqData.file.path, "r+")
            if not fd then
                return nil, err
            end
            fd:close()
            if "form" == reqData.file.mode then
                cmdTable[#cmdTable + 1] = string.format("-H 'Expect:'")
                cmdTable[#cmdTable + 1] = string.format("-F '%s=@%s'", reqData.file.field, reqData.file.path)
            else
                cmdTable[#cmdTable + 1] = "-T " .. reqData.file.path
            end
        end
    end

    if reqData.cw_hack then
        cmdTable[#cmdTable + 1] = string.format(">> /tmp/cwresp-%s", reqData.cw_hack)
    end

    return table.concat(cmdTable, " ")
end

--- Processing response from callweaver CURL request
local function __parseResponse(data)
     if not data or "" == data then
        return false, 400
    end
    
    local scode   = nil
    local headers = {}
    local body    = nil
    for line in data:gmatch("([^\r\n]+)%s+\r?\n") do
        if not scode then
            -- parse a return code string
            scode = line
        elseif scode and line:match("^HTTP/%d%.?%d? .*") then
            -- clear parsed data as found headers from another message
            scode   = line
            headers = {}
        else    -- make headers parsing and saving
            local f, v = line:match("(.*): (.*)")
            if f and v then
                headers[f:lower()] = v
            end
        end
    end
    
    local code = scode and scode:match("^HTTP/%d%.?%d? (%d+)")
          code = tonumber(code)
       
    if code then
        body = data:match(".*\r?\n%s*\r?\n(.+)\r?\n?$")
    end
    
    local rs = code and headers
    return rs, code, headers, scode, body
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Checking for supported request methods
function cwShellCURLBackend.isMethodSupported(method)
    return supportedMethods[method]
end

--- Sending HTTP request
function cwShellCURLBackend.request(method, URLData, opts)
    if not cwShellCURLBackend.isMethodSupported(method) then
        return false
    end
    if not URLData or not opts then
        return false
    end
    if not URLData.target then
        return false, "tagrget URL not specified"
    end
    
    -- processing request data
    local reqData, err = __buildRequest(opts, method)
    if not reqData then
        return false, err
    end
    
    local command, err = __buildCURLCommand(reqData, URLData)
    if not command then
        return false, err or "error building CURL command for shell"
    end
    
    local output = channel.SHELL(command):get() or ""

    if reqData.cw_hack then
        local resp     = "/tmp/cwresp-" .. reqData.cw_hack
        local fresp    = io.open(resp)
        if fresp then
            output = fresp:read("*a")
            fresp:close()
            os.remove(resp)
        else 
            return false, "Could not open temporary file: " .. resp
        end
    end

    local rs, rcode, rheaders, rscode, rbody = __parseResponse(output)
    if not rs then
        return false, rscode
    end
    
    return {
         code    = rcode
        ,scode   = rscode
        ,headers = rheaders
        ,body    = rbody
    }
end

return cwShellCURLBackend
