--- HTTP request engine based on callweaver CURL function.
--  Basically used for simple HTTP/HTTPS requests.
-- 
-- @file    callweaverCURLBackend.lua 
-- @author  Vyacheslav Khudan
-- @created 2017/11/23
-- 
-- @note engine does not support following things:
-- * sending HTTP request with custom headers
-- * sending/retrieving file to/from server
-- * some HTTP requests methods, like PUT, DELETE, etc.

local supportedMethods = {
         GET    = true
        ,POST   = true
        ,PUT    = true
        ,DELETE = true
}

local cwCURLBackend = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
--- Building HTTP request
local function __buildRequest(opts, method)
    opts            = opts or {}
    local file       = nil
    local content    = nil
    local userAgent  = "Mozilla/5.0"
    local authData   = nil
    local authMethod = "basic" -- default value
    local timeout    = 60
    
    -- processing user agent
    if opts.userAgent then
        userAgent = opts.userAgent
    end
    
    -- processing timeout value
    local optsTimeout = tonumber(opts.timeout) 
    if optsTimeout then
        timeout = optsTimeout
    end
    
    -- processing authorization data
    if opts.authUser and opts.authPass then
        authData = string.format("%s:%s", opts.authUser, opts.authPass)
        if "digest" == opts.authMethod then
            authMethod = "digest"
        end
    end
    
     -- processing content data
    if "POST" == method or "PUT" == method then
        content = opts.content
        if not content then
            return false, string.format("Not found data in %s request", method)
        end
        local url = require"socket.url"
        if "table" == type(content) then
            local content_t = {}
            for key, value in pairs(content) do
                if key and value then
                    key        = url.escape(key)
                    value      = url.escape(value)
                    local data = key .. "=" .. value
                    table.insert(content_t, data)
                end
            end
            content = table.concat(content_t, "&")
        end
    end
    
    return {
         userAgent  = userAgent
        ,timeout    = timeout
        ,authData   = authData
        ,authMethod = authMethod
        ,content    = content
        ,file       = file
        ,ssl        = {insecure = opts.insecure}
    }
end


--- Processing response from callweaver CURL request
local function __parseResponse(data)
    if not data or "" == data then
        return false, 400
    end
    
    local scode   = nil
    local headers = {}
    local body    = nil
    
    for line in data:gmatch("([^\r\n]+)%s+\r?\n") do
        if not scode then
            -- parse a return code string
            scode = line
        elseif scode and line:match("^HTTP/%d%.?%d? .*") then
            -- clear parsed data as found headers from another message
            scode   = line
            headers = {}
        else    -- make headers parsing and saving
            local f, v = line:match("(.*): (.*)")
            if f and v then
                headers[f:lower()] = v
            end
        end
    end
    
    local code = scode and scode:match("^HTTP/%d%.?%d? (%d+)")
          code = tonumber(code)
       
    if code then
        body = data:match(".*\r?\n%s*\r?\n(.+)\r?\n?$")
    end
    
    local rs = code and headers
    return rs, code, headers, scode, body
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Checking for supported request methods
function cwCURLBackend.isMethodSupported(method)
    return supportedMethods[method]
end

--- Sending HTTP request
function cwCURLBackend.request(method, URLData, opts)
    if not cwCURLBackend.isMethodSupported(method) then
        return false
    end
    if not URLData or not opts then
        return false
    end
    if not URLData.target then
        return false, "tagrget URL not specified"
    end
    
    -- processing request data
    local reqData, err = __buildRequest(opts, method)
    if not reqData then
        return false, err
    end
    
    -- set up requests parameters
    channel.CURLOPT("method"):set(method)
    channel.CURLOPT("httptimeout"):set(reqData.timeout)
    channel.CURLOPT("useragent"):set(reqData.userAgent)
    channel.CURLOPT("header"):set("yes")
    if reqData.authData then
        channel.CURLOPT("userpwd"):set(reqData.authData)
        if "digest" == reqData.authMethod then
            channel.CURLOPT("httpauth"):set(2) -- CURLAUTH_DIGEST
        end
    end
    if reqData.ssl.insecure then
        channel.CURLOPT("ssl_verifypeer"):set("no")
        -- XXX: required changes on callweaver side
        channel.CURLOPT("ssl_verifyhost"):set("no")
    end
    
    local rawData   = nil
    local URLtarget = string.format("%q", URLData.target)
    if reqData.content then
        rawData = channel.CURL(URLtarget, reqData.content):get() or ""
    else
        rawData = channel.CURL(URLtarget):get() or ""
    end
    local rs, rcode, rheaders, rscode, rbody = __parseResponse(rawData)
    if not rs then
        return false, rscode
    end
    
    return {
         code    = rcode
        ,scode   = rscode
        ,headers = rheaders
        ,body    = rbody
    }
end

return cwCURLBackend