local md5 = require"md5"
local s_http = require "socket.http"
local s_url = require "socket.url"
local ltn12 = require "ltn12"

local hash = function(...)
    return md5.sumhexa(table.concat({...}, ":"))
end

--- Parse the value part of the WWW-Authenticate header into a table.
local parse_header = function(h)
    local r = {}
    for k, v in (h .. ','):gmatch("(%w+)=(.-),") do
        if v:sub(1, 1) == '"' then -- strip quotes
            r[k:lower()] = v:sub(2, -2)
        else r[k:lower()] = v end
    end
    return r
end

--- Helper to build the Authorization header. `t` is a table of tables.
local make_digest_header = function(t)
    local r = {}
    for i = 1, #t do
        local x = t[i]
        if x.unquote then
            r[i] =  x[1] .. '=' .. x[2]
        else
            r[i] = x[1] .. '="' .. x[2] .. '"'
        end
    end
    return "Digest " .. table.concat(r, ', ')
end

--- Main logic. `params` is always a table and can be modified.
local __getDigestAuthHeader = function(params)
    if not params.url then return nil, "missing URL" end
    local url = s_url.parse(params.url)
    local user, password = url.user, url.password
    if not (user and password) then
        return nil, "missing credentials in URL"
    end
    url.user, url.password, url.authority, url.userinfo = nil, nil, nil, nil
    params.url = s_url.build(url)
    local ghost_source
    if params.source then
        local ghost_chunks = {}
        local ghost_capture = function(x)
            if x then ghost_chunks[#ghost_chunks+1] = x end
            return x
        end
        local ghost_i = 0
        ghost_source = function()
            ghost_i = ghost_i+1
            return ghost_chunks[ghost_i]
        end
        params.source = ltn12.source.chain(params.source, ghost_capture)
    end
    local b, c, h = s_http.request(params)
    if (c == 401) and h["www-authenticate"] then
        local ht = parse_header(h["www-authenticate"])
        assert(ht.realm and ht.nonce)
        if ht.qop ~= "auth" then
            return nil, string.format("unsupported qop (%s)", tostring(ht.qop))
        end
        if ht.algorithm and (ht.algorithm:lower() ~= "md5") then
            return nil, string.format("unsupported algo (%s)", tostring(ht.algorithm))
        end
        local nc, cnonce = "00000001", string.format("%08x", os.time())
        local uri = s_url.build({path = url.path, query = url.query})
        local method = params.method or "GET"
        local response = hash(
            hash(user, ht.realm, password),
            ht.nonce,
            nc,
            cnonce,
            "auth",
            hash(method, uri)
        )
        params.headers = params.headers or {}
        local auth_header = {
            {"username", user},
            {"realm", ht.realm},
            {"nonce", ht.nonce},
            {"uri", uri},
            {"cnonce", cnonce},
            {"nc", nc, unquote=true},
            {"qop", "auth"},
            {"algorithm", "MD5"},
            {"response", response},
        }
        if ht.opaque then
            table.insert(auth_header, {"opaque", ht.opaque})
        end
        return make_digest_header(auth_header)
    else
        return nil, "incorrect response"
    end
end

local getDigestAuthHeader = function(url, method)
    return __getDigestAuthHeader({url = url, sink = ltn12.sink.table({}), method = method})
end

return {
    getDigestAuthHeader = getDigestAuthHeader,
}
