--- Http client implementation 
-- 
-- @file    httpClient.lua 
-- @author  Vyacheslav Khudan
-- @created 2015/07/10
-- 
-- @note Can't be used by nmap engine - required modules use global variables

-- @module httpClient
local httpClient = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
--- Parse http URL data:
--      schema  A shema of the URI for current request
--      host    A target host for http client request
--      port    A port in the request
--      path    A requested http path
local function __parse_http_url_data(urlStr)
    if "string" ~= type(urlStr) then
        return false, "Wrong settings format"
    end
    
    -- parsing URI string   
    local data         = {path = "/"}
    local idxParser    = {}
          idxParser[1] = function(val)  -- scheme
              if val then
                  data.scheme = val:match("(.*):/")
                  return data.scheme and 2 or idxParser[2](val) 
              end
          end
          idxParser[2] = function(val)  -- host & port
              if val then
                  if val:find(":") then
                      data.host, data.port = val:match("([^:/]+):?(%d+)")
                  else
                      data.host = val:match("([^:/]+)")
                  end
                  return data.host and 3 or 2 
              end
          end
          idxParser[3] = function(val)  -- path
              if val then
                  data.path = data.path..val
                  return 3
              end
          end
    
    local idx = 1
    for val in urlStr:gmatch("([^/]+/?)") do
        idx = idxParser[idx](val)
    end
    
    -- port values checking
    local port = tonumber(data.port)
    if not port then
        port = "https" == data.scheme and 443 or 80     -- set default port
    elseif 0 >= port or 65535 < port then
        return false, "Port value is not correct"
    end
    data.port = port
    
    -- schema checking
    if data.port and not data.scheme then
       data.scheme = 443 == data.port and "https" or "http" 
    end
    if "http" ~= data.scheme and "https" ~= data.scheme then
        return false, "Invalid scheme in the request"
    end
    if "https" == data.scheme then  
        data.ssl = true -- set flag for default HTTPS schema
    end
    
    -- host value checking
    if not data.host or "" == data.host then
        return false, "Host value is not correct"
    end
    
    -- building result URL string
    local url       = require"socket.url"
    local url_data  = url.parse(data.path)
    url_data.scheme = data.scheme
    url_data.host   = data.host
    url_data.port   = data.port
    data.target     = url.build(url_data)
    
    return data
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local client = {}   -- meta Class

--- Applies the URL escaping content coding to a string
-- @param str   the string to be encoded
-- @return the encoded string
function client.urlEncode(str)
    if str and "" ~= str then
        str = string.gsub (str, "\n", "\r\n")
        str = string.gsub(str, "([^ A-Za-z0-9._-])", function(c)
                return string.format("%%%02x", string.byte(c)):upper()
            end)
        str = string.gsub(str, " ", "+")
    end
    return str
end

--- Removes the URL escaping content coding from a string
-- @param str   the string to be decoded.
-- @return the decoded string
function client.urlDecode(str)
    if str and "" ~= str then
        str = string.gsub (str, "+", " ")
        str = string.gsub (str, "%%(%x%x)", function(hex)
                 return string.char(tonumber(hex,16)) 
              end)
        str = string.gsub (str, "\r\n", "\n")
    end
    return str
end

--- Fetches a resource with a specified request type
-- @param method    A request method type (like GET, POST ...)
-- @param resourceIdent A URI data (like host, port, etc.)
-- @param options       A table with http request parameters
--          userAgent       A custom client User Agent (def. "Mozilla/5.0")
--          authMethod      An authentification user (def. "Basic")
--          authUser        A name of the user for authorization
--          authPass        A password for authorization field
--          timeout         A connection timeout in seconds (with millisecond precision)
--          headers         A list of the custom headers
--          insecure        Disable SSL certificate verification
--          inFile          A path to a downloading file
--          outFileByForm   A path to an uploading file via form
--          outFileByStream A path to an uploading file as stream
function client:request(method, resourceIdent, options)
    if not method or "" == method then
        return false, "HTTP method was not found"
    end
    local options = options or {}
    
    -- detect a request with the file processing
    local requiredFileLoading = options.inFile or options.outFileByForm or options.outFileByStream
    
    -- initialize a backend for processing a request
    local factory          = require"core.httpClient.backendFactory":new()
    local httpBackend, err = factory:instantiate(requiredFileLoading, self.backend)
    if not httpBackend then
       return false, err
    end
    
    if not httpBackend.isMethodSupported(method) then
        return false, "not suppoted HTTP method by the backend"
    end

    local URLData, err = __parse_http_url_data(resourceIdent)
    if not URLData then
        return false, err
    end
    
    local rv, err = httpBackend.request(method, URLData, options)
    if not rv or not rv.body then
        return rv, err
    end
    
    return rv
end

--- Fetches a resource with a GET request
-- @param resourceIdent A URI data (like host, port, etc.)
-- @param options       A table with http request parameters (like timeout or headers)
-- @return A response table otherwise FALSE value with error message
function client:get(resourceIdent, options)
    return self:request("GET", resourceIdent, options)
end

--- Fetches a resource with a GET request by raw data
-- @param host      A requested host address
-- @param port      A requested port
-- @param options   A table with http request parameters (like timeout or headers)
-- @return A response table otherwise FALSE value with error message
-- @see client:get
function client:rawget(host, port, path, options)
    local resourceIdent = string.format("%s:%s/%s", host, port, path)
    return self:get(resourceIdent, options)
end

--- Fetches a resource with a DELETE request
-- @param resourceIdent A URI data (like host, port, etc.)
-- @param options       A table with http request parameters (like timeout or headers)
-- @return A response table otherwise FALSE value with error message
function client:delete(resourceIdent, options)
    return self:request("DELETE", resourceIdent, options)
end

--- Fetches a resource with a DELETE request by raw data
-- @param host    A requested host address
-- @param port    A requested port
-- @param options A table with http request parameters (like timeout or headers)
-- @return A response table otherwise FALSE value with error message
-- @see client:delete
function client:rawdelete(host, port, path, options)
    local resourceIdent = string.format("%s:%s/%s", host, port, path)
    return self:delete(resourceIdent, options)
end

--- Fetches a resource with a POST request
-- @param resourceIdent A URI data (like host, port, etc.)
-- @param data          A data for POST request. Can be string or table with string values for key,value
-- @param options       A table with http request parameters
-- @return A response table otherwise FALSE value with error message
function client:post(resourceIdent, data, options)
    -- pushing post data to options
    options         = options or {}
    options.content = data
    
    return self:request("POST", resourceIdent, options)
end

--- Fetches a resource with a POST request by raw data
-- @param host      A requested host address
-- @param port      A requested port
-- @param data      A data for POST request. Can be string or table with string values for key,value
-- @param options   A table with http request parameters
-- @return A response table otherwise FALSE value with error message
-- @see client:post
function client:rawpost(host, port, path, data, options)
    local resourceIdent = string.format("%s:%s/%s", host, port, path)
    return self:post(resourceIdent, data, options)
end

--- Fetches a resource with a PUT request
-- @param resourceIdent A URI data (like host, port, etc.)
-- @param data          A data for PUT request. Can be string or table with string values for key,value
-- @param options       A table with http request parameters
-- @return A response table otherwise FALSE value with error message
function client:put(resourceIdent, data, options)
    -- pushing post data to options
    options         = options or {}
    options.content = data
    
    return self:request("PUT", resourceIdent, options)
end

--- Fetches a resource with a PUT request by raw data
-- @param host      A requested host address
-- @param port      A requested port
-- @param data      A data for PUT request. Can be string or table with string values for key,value
-- @param options   A table with http request parameters
-- @return A response table otherwise FALSE value with error message
-- @see client:put
function client:rawput(host, port, path, data, options)
    local resourceIdent = string.format("%s:%s/%s", host, port, path)
    return self:put(resourceIdent, data, options)
end

--- Download a file via HTTP protocol
-- @param resourceIdent A URI data (like host, port, etc.)
-- @param filePath      A path to a downloading file
-- @param options       A table with http request parameters (like timeout or headers)
-- @return A response table otherwise FALSE value with error message
function client:getFile(resourceIdent, filePath, options)
    local options  = options or {}
    options.inFile = filePath 
    return self:get(resourceIdent, options)
end

--- Upload a file via HTTP protocol via from data
-- @param resourceIdent A URI data (like host, port, etc.)
-- @param filePath      A path to a uploading file
-- @param data          A data for PUT request. Can be string or table with string values for key,value
-- @param options       A table with http request parameters (like timeout or headers)
-- @return A response table otherwise FALSE value with error message
function client:sendFileAsFormData(resourceIdent, filePath, data, options)
    local options         = options or {}
    options.outFileByForm = filePath
    return self:post(resourceIdent, data, options)
end


--- Transparent methods
function httpClient.urlEncode(str)
    return client.urlEncode(str)
end
function httpClient.urlDecode(str)
    return client.urlDecode(str)
end

--- Object constructor
function httpClient:new(backend)
    local obj = {}
    obj.backend = backend
    setmetatable(obj, {__index = client})
    
    return obj
end

return httpClient
