--- HTTP request engine based on standart libraries.
--  Uses library lua-socket and lua-sec to processing requests.
-- 
-- @file    socketBackend.lua 
-- @author  Vyacheslav Khudan
-- @created 2017/11/23

local supportedMethods = {
         GET    = true
        ,POST   = true
        ,PUT    = true
        ,DELETE = true
}

local socketBackend = {}

----------------------------------------------------------------
-- Custom socket implementation
----------------------------------------------------------------
--- Create socket object with SSL supporting
local function __createSocketObj(opts)
    local opts = opts or {}
    
    local function _settimeout(tcpObj, timeout)
        local timeout = tonumber(timeout)
        if timeout then
            tcpObj:settimeout(timeout)
        end
    end
    
    return {
        tcp  = function()
            local iSocket = {tcpObj = require"socket".tcp()}
            -- new connect method
            function iSocket:connect(host, port)
                -- set up timeout befor initial connection
                _settimeout(self.tcpObj, opts.timeout)
                local rs, err = self.tcpObj:connect(host, port)
                if not rs then
                    return rs, err
                end

                -- creating SSL socket
                if opts.isSSL then
                    self.tcpObj, err = require"ssl".wrap(self.tcpObj, {
                                        mode     = "client"
                                       ,protocol = opts.secureProtocol or "tlsv1"
                                       ,verify   = false == opts.insecureSSL and "peer" or "none"
                                       ,capath   = opts.capath
                                   })
                    if not self.tcpObj or "userdata" ~= type(self.tcpObj) then
                        return 1 -- SSL connection was not created
                    end
                end
                
                -- re-set timeout to apply for SSL connections
                _settimeout(self.tcpObj, opts.timeout)
                
                -- executing SSL handshake
                if opts.isSSL then
                    self.tcpObj:dohandshake()
                end
                
                return 1
            end
            
            -- socket metatable callback
            local function _mSocket(t, method)
                return function(t, ...)
                    return t.tcpObj[method](t.tcpObj, ...)
                end
            end

            return setmetatable(iSocket, { __index = _mSocket })
        end
    }
end

--- Building HTTP request
local function __buildRequest(opts, method, URLData)
    opts          = opts or {}
    local headers = {}
    local file    = nil
    local content = nil
    
    -- processing custom headers
    if "table" == type(opts.headers) then
        for name, val in pairs(opts.headers) do
            headers[name] = val
        end
    end
    
    -- processing user agent value
    headers["User-Agent"] = opts.userAgent or "Mozilla/5.0"

    -- processing authorization header
    if opts.authUser and opts.authPass then
        -- TODO: support for digest authorization method
        if "digest" == opts.authMethod then
            local http_digest = require"core.httpClient.http-digest"
            local url = string.format("%s://%s:%s@%s:%s%s", URLData.scheme, opts.authUser, opts.authPass, URLData.host, URLData.port, URLData.path)
            local digestHeader, err = http_digest.getDigestAuthHeader(url, method)
            if err then
                return false, "error creating header for digest auth"
            end
            headers["Authorization"] = digestHeader
        else
            local authPair           = require"mime".b64(opts.authUser..":"..opts.authPass)
            headers["Authorization"] = "Basic".." "..authPair
        end
    elseif "bearer" == opts.authMethod and opts.authToken then
            headers["Authorization"] = "Bearer "..opts.authToken
    end
    
    -- processing content data
    if "POST" == method or "PUT" == method then
        -- processing file uploading
        -- TODO: sending file via POST form
        if opts.outFileByForm or opts.outFileByStream then
            return false, "not supported file sending"
        end
        
        -- processing request content
        content = opts.content
        if not content then
            return false, "not found data in POST/PUT request"
        end
        local url = require"socket.url"
        if "table" == type(content) then
            local content_t = {}
            for key, value in pairs(content) do
                if key and value then
                    key        = url.escape(key)
                    value      = url.escape(value)
                    local data = key .. "=" .. value
                    table.insert(content_t, data)
                end
            end
            content = table.concat(content_t, "&")
        end
        
        -- set request headers
        headers["Content-Type"]   = headers["Content-Type"] or "application/x-www-form-urlencoded"
        headers["Content-Length"] = #content
        
    -- processing request type to a file downloading
    elseif "GET" == method then
        if opts.inFile then
            file = {
                 direction = "incoming"
                ,path      = opts.inFile
            }
        end
    end
    
    local rv = {
         headers = headers
        ,content = content
        ,timeout = opts.timeout
        ,file    = file
        ,ssl     = {insecure = opts.insecure, protocol = opts.secureProtocol}
        ,capath  = opts.capath
        ,proxy   = opts.proxy
    }
    
    return rv 
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Checking for supported request methods
function socketBackend.isMethodSupported(method)
    return supportedMethods[method]
end

--- Sending HTTP request
function socketBackend.request(method, URLData, opts)
    if not socketBackend.isMethodSupported(method) then
        return false
    end
    if not URLData or not opts then
        return false
    end
    if not URLData.target then
        return false, "tagrget URL not specified"
    end

    -- processing options values
    local reqData, err = __buildRequest(opts, method, URLData)
    if not reqData then
        return false, err
    end
    
    -- initialize socket for request
    local reqSocket = __createSocketObj({
         timeout     = reqData.timeout
        ,isSSL       = URLData.ssl
        ,insecureSSL = reqData.ssl.insecure
        ,capath = reqData.capath or nil
        ,secureProtocol = reqData.ssl.protocol
    })
    
    -- sending http request
    local rbody = {}
    local http  = require"socket.http"
    local ltn12 = require"ltn12"
    local sink  = ltn12.sink.table(rbody)
    if reqData.file and "incoming" == reqData.file.direction then
        local fd, err = io.open(reqData.file.path, "w")
        if not fd then
            return false, err
        end
        sink = ltn12.sink.file(fd)
    end
    local rs, rcode, rheaders, rscode = http.request{
         url     = URLData.target
        ,method  = method
        ,headers = reqData.headers
        ,sink    = sink
        ,source  = reqData.content and ltn12.source.string(reqData.content)
        ,create  = reqSocket.tcp
        ,proxy   = reqData.proxy,
    }

    if not rs then
        return false, rcode
    end
    
    -- collecting result data
    return {
         code    = rcode
        ,scode   = rscode
        ,headers = rheaders
        ,body    = table.concat(rbody)
    }
end

return socketBackend
