local string, tostring, select, pcall, pairs, unpack = string, tostring, select, pcall, pairs, unpack
local debug, type, table, tonumber = debug, type, table, tonumber
local util = require"core.pbxutils"
local s = require"syslog"
local ini_parser = require"inilazy"

local math = require"math"
local os = require"os"

module (... or "logger")

local function send_msg(syslog_priority, syslog_ident, debug_stack_level, context, msg, ...)

	-- prepare formating message
	local util, tlogs = util, {}
	local gi = debug.getinfo(debug_stack_level, "nl")
	local name, line = gi["name"] or "self", gi["currentline"] or ""
	local msg = "boolean" == type(msg) and "boolean: " .. tostring(msg) or msg
	msg = msg or "nil"
	local jsonlog = nil
	-- set marker for log level
	local marker = "<"..(syslog_priority:match("LOG_(.*)") or "")..">"
	local args_in_len = 0

	for i, _ in pairs({...}) do args_in_len = i end

	if type(msg) == "table" then
		local tbl_name = select(1, ...) or "table"
		tlogs[#tlogs+1] = string.format("%s %s:%s: %s = {", marker, name, line, tbl_name)
		local t = util.split(util.tprint(msg, 2, nil, true), "\n")
		for i=1, #t do
 			tlogs[#tlogs+1] = t[i]
 		end
 		tlogs[#tlogs+1] = "}"
 		if not ("send_aps_configuration" == name and "result" == tbl_name) then  -- WMS-5086: do not send devices provisioning parameters to kibana
   		local st
   		st, jsonlog = pcall(util.tojson, msg)
   		if not st then
   			jsonlog = nil
   			s.openlog("lua", s.LOG_PID + s.LOG_ODELAY, "LOG_USER")
	   		s.syslog(syslog_priority, string.format("An error occurred while serializing in json: %s", tostring(jsonlog)))
	   		s.closelog()
	   	end
		end
 	elseif string.find(msg, ">> Start") then	-- used for devide logs of different devices
 		tlogs[1] = "\n"
 		tlogs[2] = msg
	elseif 0 == args_in_len then
		tlogs[#tlogs+1] = string.format("%s %s:%s: %s - %s", marker, name, line, msg, util.tojson(context))
	else
		local args_out = {}
		local args_in = {...}
		for i = 1, args_in_len do
			if not args_in[i] then
				table.insert(args_out, "nil")
			elseif "table" == type(args_in[i]) then
				table.insert(args_out, util.tprint(args_in[i], 0, nil, true))
			elseif "string" == type(args_in[i]) then
				table.insert(args_out, args_in[i])
			else
				table.insert(args_out, tostring(args_in[i]))
			end
		end

		_, p_num = string.gsub(msg, "%%", "%%")
		if args_in_len > 0 and p_num > 0 then
			for i = #args_out, p_num  do
				table.insert(args_out, "nil")
			end
			local ok, res = pcall(function() return string.format(tostring(msg), unpack(args_out)) end)
			if ok then
				msg = res
			else
				for k, arg in pairs(args_out) do
					args_out[k] = tostring(arg)
				end
				msg = tostring(msg) .. " " .. table.concat(args_out, " ")
			end
		else
			msg = tostring(msg) .. " " .. table.concat(args_out, " ")
		end

		tlogs[#tlogs+1] = string.format("%s %s:%s: %s - %s", marker, name, line, msg, util.tojson(context))
	end

	if "[WMS 3.0 Dialplan]" == syslog_ident or "[WMS 3.0]" == syslog_ident then
		syslog_ident = "lua"
	end

	-- send message to syslog
 	local len_tlogs = #tlogs
 	if len_tlogs == 1 then -- simple string
 		s.openlog(syslog_ident, s.LOG_PID + s.LOG_ODELAY, "LOG_USER")
 		s.syslog(syslog_priority, tlogs[1])
 		s.closelog()
 	else
 		if jsonlog then
   		s.openlog("luatable", s.LOG_PID + s.LOG_ODELAY, "LOG_USER")
   		s.syslog(syslog_priority, jsonlog)
   		s.closelog()
 		end
 		s.openlog("lualocal", s.LOG_PID + s.LOG_ODELAY, "LOG_USER")
 		for i=1,len_tlogs do
 			s.syslog(syslog_priority, tlogs[i])
 		end
 		s.closelog()
 	end

	return true
end

local logger = {
    config = {},

    error = function(self, msg, ...)
        if self.config.log_level > 0 then
        	send_msg("LOG_ERR", self.config.ident, self.config.debug_stack_level, self.config.context, msg, ...)
        end
    end;

    warning = function(self, msg, ...)
        if self.config.log_level > 1 then
        	send_msg("LOG_WARNING", self.config.ident, self.config.debug_stack_level, self.config.context, msg, ...)
        end
    end;

    info = function(self, msg, ...)
        if self.config.log_level > 2 then
        	send_msg("LOG_INFO", self.config.ident, self.config.debug_stack_level, self.config.context, msg, ...)
        end
    end;

    debug = function(self, msg, ...)
        if self.config.log_level > 3 then
        	send_msg("LOG_DEBUG", self.config.ident, self.config.debug_stack_level, self.config.context, msg, ...)
        end
    end;
}

-- ident A identificator for messages at syslog
-- logLevel A level of the logs sending to syslog
function init(self, ident, log_level, context, debug_stack_level)
	context = context or {}
	local default_context = {}
	local env_id = os.getenv('REQUEST_ID')
	if env_id then
		default_context.requestId = env_id
	else
		math.randomseed(os.time())
		default_context.requestId = "rid_"..math.random(1000000, 2000000)
	end
	context = util.table_merge(default_context, context)

	-- Check logger ini file
	local logger_ini, value = "/dev/shm/logger.ini"
	local conf = ini_parser.get(logger_ini)
	if conf and "table" == type(conf) then
		local section = {}
		if "table" == type(conf[ident]) then
			section = conf[ident]
		elseif "table" == type(conf["all"]) then
			section = conf["all"]
		end
		value = tonumber(section["log_level"])
		if value then log_level = value end
		value = tonumber(section["debug_stack_level"])
		if value then debug_stack_level = value end
	end
	local _logger = util.tcopy(logger)
	_logger.config.ident = ident
	_logger.config.log_level = log_level or 1
	_logger.config.debug_stack_level = debug_stack_level or 3
	_logger.config.context = context
	return _logger
end

--- Example of the ini file "/dev/shm/logger.ini"
-- [[WMS 3.0]] 		-- ident "[WMS 3.0]"
-- log_level = 4
-- debug_stack_level = 4
--
-- [cdrconnection]  -- ident "cdrconnection"
-- log_level = 3
--
-- [all]
-- log_level = 4
--
