module(... or "pbxutils", package.seeall)

-- Used for move functions from utils.lua, helper.lua, core.lua !!!!!!!! ---

----------------------------------------------------------------
-- Output functions
----------------------------------------------------------------
--- Split string with patterns
function split (str, pat)
	local str = str or ""

	local t = {}  -- NOTE: use {n = 0} in Lua-5.0
	local fpat = "(.-)" .. pat
	local last_end = 1
	local s, e, cap = tostring(str):find(fpat, 1)
	while s do
		if s ~= 1 or cap ~= "" then
			table.insert(t,cap)
		end
		last_end = e+1
		s, e, cap = str:find(fpat, last_end)
	end
	if last_end <= #str then
		cap = str:sub(last_end)
		table.insert(t, cap)
	end
	return t
end

--- Formatting print table
local dstring
function tprint (tt, indent, done, log)
	if tt == nil then return "" end
	if done == nil then dstring = "" end

	dstring = dstring or ""
	done = done or {}
	indent = indent or 0
	if type(tt) == "table" then
		for key, value in pairs (tt) do
			dstring = dstring .. string.rep (" ", indent) -- indent it
			if type (value) == "table" and not done [value] then
				done [value] = true
				dstring = dstring .. string.format("[\"%s\"] = {\n", tostring (key));
				tprint (value, indent + 7, done)
				dstring = dstring .. string.rep (" ", indent+4) -- indent it
				dstring = dstring .. "},\n";
			else
				if log then
					dstring = dstring .. string.format("[\"%s\"] = [\"%s\"],\n", tostring (key), tostring(value))
				else
					dstring = dstring .. string.format("[\"%s\"] = [=====[%s]=====],\n", tostring (key), tostring(value))
				end
			end
		end
	else
		dstring = dstring .. tt .. "\n"
	end
	return dstring
end


function tojson (object)
	local json = require('dkjson')
	local lookup_table = {}
	local function _copy(object)
		if type(object) ~= "table" then
			return tostring(object)
		elseif lookup_table[object] then
			return lookup_table[object]
		end
		local new_table = {}
		lookup_table[object] = new_table
		for index, value in pairs(object) do
			new_table[_copy(index)] = _copy(value)
		end
		return setmetatable(new_table, getmetatable(object))
	end
	return json.encode(_copy(object))
end


--- Deep copy of a given table
-- @param object A table to be copied
-- @return object A copy of a given table
function tcopy (object)
	local lookup_table = {}
	local function _copy(object)
		if type(object) ~= "table" then
			return object
		elseif lookup_table[object] then
			return lookup_table[object]
		end
		local new_table = {}
		lookup_table[object] = new_table
		for index, value in pairs(object) do
			new_table[_copy(index)] = _copy(value)
		end
		return setmetatable(new_table, getmetatable(object))
	end
	return _copy(object)
end

--- Check value in a plain table
-- @param #string e A value to lookup
-- @param #table t Source table
-- @return #boolean result
function in_table (e, t)
	for i=1,#t do
		if t[i] == e then
			return true
		end
	end
	return false
end

--- Function for concatenation tables (add value from @list to @t
function table_concat (t, list)
	for key, value in pairs(list) do
			t[key] = value
	end
end

--- Function for recursive merging tables
-- @param #table	orig	A table used as original (often has default settings)
-- @param #table	list	A table which merged with @t
-- @return result merged table
function table_merge (orig, list)
	local t = tcopy(orig)
	for key, value in pairs(list) do
		if type(value) == "table" and t[key] and type(t[key]) == "table" then
			t[key] = table_merge(t[key], value)
		else
			t[key] = value
		end
	end
	return t
end

----------------------------------------------------------------
-- Filesystem functions
----------------------------------------------------------------
--- Get users directory
-- @return users directory
function get_users_settings_dir ()
	local userdir = '/rw2/etc/user.d'
	if require"core.fsutils".is_readable(userdir) then
		return userdir .. "/"
	end
	return false
end

----------------------------------------------------------------
-- System functions
----------------------------------------------------------------
--- Run a shell command with waiting command output
-- @param command   A command to be run in shell
-- @param Mixed     values to be applied to format function
-- @return command output otherwise error string
function run(command, ...)
	local rv  = nil
	if not command or "" == command then
        return rv, "Invalid command data"
    end

    local f, err = io.popen(string.format(command, ...))
    if not f then
        return rv, err
    end
    rv = f:read("*a") -- read output (only stdout)
	f:close()

    return rv, err
end

--- Execute shell command without waiting command results
-- @param #string command A command to be run in shell
-- @param #mixed Mixed values to be applied to format function
-- @return #boolean Always true
function exec(command, ...)
	os.execute(string.format(command, ...))
	return true
end

--- Send an email with attach using sendemail.php script without template
--  See mailNotifications module for details
function mailto (message, subject, recipient, attachments, com_id)
	return require"common.mailNotifications".send(message, subject, recipient, attachments, com_id)
end

-- Send notification to ejabberd via http api about roster change
function roster_notify(uid, difftable)

    local function __get_exten(uid)
        local userData  = require"core.wmsldap":get("uid", uid)
        if userData and userData[1] and userData[1][2] then
            local data = userData[1][2]
            return data.telephoneNumber
        end
    end

    local exten = __get_exten(uid)
    if not exten then
        return
    end

    -- ejabberd command validator does not work with complex data, use double encoding
    local diff_str = require"dkjson".encode({[exten] = difftable})
    local diffparam = {["roster_diff"] = diff_str}
    local jsonparam = require"dkjson".encode(diffparam)
    local hclient  = require"core.httpClient":new()
    hclient:rawpost("127.0.0.1", 5280, "/api/notify_roster", jsonparam, {["Content-Type"] = "application/json"})
end

--- Check format of IPv4 address
function get_ipv4 (address)
	if not address then return false end
	local ip_addr = ""
	local chunk = {address:match("(%d+)%.(%d+)%.(%d+)%.(%d+)")}
	if #chunk == 4 then
	    for i, v in pairs(chunk) do
	        if tonumber(v) < 0 or tonumber(v) > 255 then
	            return false
	        end
	        ip_addr = ip_addr..v..(i ~= 4 and "." or "")
	    end
	    return ip_addr
	else
	    return false
	end
end

--- Get and convert MAC address from string
-- @return #string MAC address ("08002737b78b")
function get_mac (str)
	local mac = ""
	if str and str ~= "" then
		mac = str:gsub("([0-9a-fA-F][0-9a-fA-F])([:g-zG-Z]*)","%1", 6):match("(%w+)"):lower()
		if mac:len() == 12 then
			return mac
		end
	end
	return false
end

----------------------------------------------------------------
-- PBX identity functions
----------------------------------------------------------------
--- Retrieve PBX serial number from the system
function pbxSerial()
    local serial = ""
    local file, error = io.open("/rw2/etc/sf2", "r")
    if file then
        local sf = {}
        for line in file:lines() do
            sf[#sf+1] = line
        end
        file:close()

        if #sf > 19 and sf[17] then
            serial = sf[17]:gsub("([0-9a-f]+)([:]*)","%1"):match("(%w+)"):lower()
        end
    end

    if not serial or serial == "" then
        file, error = io.open("/sys/class/net/eth0/address", "r")
        if file then
            local eth0 = file:read("*a")
            file:close()
            if eth0 then
                serial = eth0:gsub("([0-9a-f]+)([:]*)","%1"):match("(%w+)"):lower()
            end
        end
    end

    if not serial or serial == "" then
        file, error = io.open("/etc/serial", "r")
        if file then
            local generatedSerial = file:read("*a")
            file:close()

            if generatedSerial then
                serial = generatedSerial:gsub("([0-9a-f]+)([:]*)","%1"):match("(%w+)"):lower()
            end
        end
    end

    return serial or ""
end

--- Retrieve PBX Key
function pbxKey()
	local key = ""
	local file, error = io.open("/rw2/etc/sf2", "r")
	if file then
		local sf = {}
		for line in file:lines() do
			sf[#sf+1] = line
		end
		file:close()

		if #sf > 19 and sf[18] then
			key = sf[18]
		end
	end
	return key or ""
end

--- Retrieve IdToken (AWS Cognito Authentication IdToken)
function getIdToken()
    local rv = nil
    local socketUnix = require"socket.unix"
    local uds = socketUnix()
    local _, err = uds:connect("/var/run/data_engine/data_engine.sock")
    repeat
        if err then
            logger:debug("getIdToken unix domain socket connect error: '%s'", err)
            break
        end
        local _, err = uds:send("token\n")
        if err then
            logger:debug("getIdToken domain socket send error: '%s'", err)
            break
        end
        local resp,err = uds:receive("*a")
        if err then
            logger:debug("getIdToken unix domain socket receive error: '%s'", err)
            break
        end
        if "string" == type(resp) then
            rv = resp
        end
    until true
    uds:close()

    return rv
end

--- Get dialplan general settings
function getDialplanGeneralSettings(parameter)
    local file = "/etc/callweaver/extensions-variables.conf"
    local dialplanGSettings = require"common.helper".parse_simple_ini(file)
    return dialplanGSettings and dialplanGSettings[parameter] or dialplanGSettings
end

----------------------------------------------------------------
-- wPBX object methods
----------------------------------------------------------------
--- Retrieve PBX data from LDAP by search parameter & value
local function __getPbxData(sParam, sVal)
    local rv = nil
    if not sParam or "" == sParam then
        return rv
    end
    if not sVal or "" == sVal then
        return rv
    end

    local LDAPResult = require"core.wmsldap":get(sParam, sVal, "wPbx")[1]
    if LDAPResult and LDAPResult[1] then
        rv = {dn = LDAPResult[1]}
        if "table" == type(LDAPResult[2]) then
            for k,v in pairs(LDAPResult[2]) do
                rv[k] = v
            end
        end
    end

    return rv
end

--- Retrieve PBX data from LDAP by serial value
-- @see pbxSerial function
-- @see result table keys in wPbx object
function getPbxDataBySerial(serial)
    if not serial or "" == serial then
        serial = pbxSerial()
    end

    return __getPbxData("pbxSerial", serial)
end

--- Get WMS network PBX IP address by serial
function getPbxIpBySerial(serial)
    local pbxData = getPbxDataBySerial(serial)
    return pbxData and pbxData.pbxIp or ""
end

--- Getter for PBX DN (used for comparing with wPerson.owner)
function getPbxDNBySerial(serial)
    local pbxData = getPbxDataBySerial(serial)
    return pbxData and pbxData.dn
end

--- Retrieve PBX data from LDAP by identifier value
-- @see result table keys in wPbx object
function getPbxDataByID(ident)
    return __getPbxData("pbxid", ident)
end

--- Retrieve PBX data from LDAP by DN
function getPbxDataByDN(dn)
    local rv = nil
    if not dn or "" == dn then
        return rv
    end
    local ident = dn:match("pbxid=(%w+)")
    return getPbxDataByID(ident)
end

--- Getter for PBX DN (used for comparing with wPerson.owner)
function getPbxSerialByDN(dn)
    local pbxData = getPbxDataByDN(dn)
    return pbxData and pbxData.pbxSerial
end

----------------------------------------------------------------
-- Conversion functions
----------------------------------------------------------------
--- Convert chars in a string to phone button numbers
-- @param chars A string to be converted
-- @return string A converted string
function chars2number(chars)
    if not chars then
        return nil
    end
    
	local encode = function (c)
		local c = string.lower(c)
		if string.find(c,"%D") and string.find(c,"%A") then
			return "*"
		end

		for i, regexp in ipairs {"abc","def","ghi","jkl","mno","pqrs","tuv","wxyz"} do
			if string.find (c,"([".. regexp .."]+)") then
				return i + 1
			end
		end
		return c
	end

	return (function(chars)
		local pin = ""
		for uchar in string.gfind(chars, "([%z\1-\127\194-\244][\128-\191]*)") do
			pin = pin .. encode(uchar)
		end
		return pin
	end)(chars)
end


--- Get value for global env variable cached in dialplan GLOBAL variable
-- @param key A var name like "SERVICE_WIM_URI"
-- @param defaultValue A default value for the var
-- @return A cached env var value like "https://wim.wildix.com"
function getCachedEnvKey(key)
	if not channel or not key then
		return
	end
	local rv = channel.GLOBAL(key):get()
	if not rv or "" == rv then
		rv = require"common.helper".env(key)
		if rv and "" ~= rv then
			channel.GLOBAL(key):set(rv)
		end
	end

	return rv
end
