--- Processing callweaver call files via serialization to table and modifying it via rawlazy
-- 
-- @file	callfile.lua 
-- @author	Vyacheslav Khudan
-- @created 2014/09/08

package.path  = package.path .. ";/usr/local/share/lua/5.1/core/?.lua"
local require = require
local string  = string
local pairs   = pairs
local table   = table
local type    = type
local io      = io
local rlz     = require"rawlazy.rawlazy"

module( ... or "callfile")

--- load methods for processing raw data in the table
call_obj = rlz:new()

--- set default value to "" (empty string)
call_obj:setdef("")

--- Deserialize data from call file to table
-- @param file_path A path to loaded call file
-- @return boolean True or False with error table or mesage
function call_obj:readfile (file_path)
	local file, err = io.open(file_path)
	if not file then
		return false, string.format("Couldn't open call file: %s", err)
	end

	local rc = true	
	local err_msg = {}
	for line in file:lines() do
		-- deserialize data from line
		local key, val_a, del, val_b = line:match("^(.*):%s*([^=]*)(=?)(.*)$")
		if "=" ~= del then
			val_b = nil
		end
		local add_rc, err = call_obj:add(key, val_a, val_b)
		if not add_rc then
			rc = false
			if err then
				table.insert(err_msg, err)
			end
		end
		
	end
	
	file:close()
	return rc, not rc and err_msg
end

--- Serialize data table cell to call file string
-- @param data A cell of the call file table
-- @return table with formatted strings for call file
local function serialize (data)
	if not data or "table" ~= type(data) then
		return false, "Invalid callfile data"
	end
	
	local rt = {}
	for key, value in pairs(data) do
		if "table" == type(value) then
			for value_a, value_b in pairs(value) do
				rt[#rt + 1] = string.format("%s: %s=%s\n", key, value_a, value_b)
			end
		else
			rt[#rt + 1] = string.format("%s: %s\n", key, value)
		end
	end
	
	if 0 ~= #rt then
		return rt
	end
	return false, "Result data are empty"
end

--- Serialize data and save its to call file
-- @param file_path A destination file which should has result call file data
-- @param new_file  Flag for create/rewrite call file 
-- @return boolean True or False with error message
function call_obj:writefile (file_path, new_file)
	local fsutils = require"fsutils"
	
	if not new_file and not fsutils.is_readable(file_path) then
	   return false, string.format("Call file %q not found", file_path)
	end
	
	local rdata, err = serialize(call_obj.data)
	if not rdata then
		return false, err
	end
	
	local tmpfile = fsutils.tmpname("/tmp/")
	local tmpcf = io.open(tmpfile, "w")
	if not tmpcf then
		fsutils.rmfile(tmpfile)
		return false, "Couldn't open temporary file"
	end
	
	for _, str in pairs(rdata) do
		tmpcf:write(str)
	end
	tmpcf:close()
	call_obj:cleardata() -- clear data table
	
	return fsutils.mvfile(tmpfile, file_path)
end
