--- Processing raw data in the table
-- 
-- @file	rawlazy.lua 
-- @author	Vyacheslav Khudan
-- @created 2014/09/08

local type         = type
local string       = string
local pairs        = pairs
local setmetatable = setmetatable

module( ... or "rawlazy")

local data_obj = {}
data_obj.data = {} --- buffer for raw data
data_obj.___ = 0   --- default value

--- Create/update object with inheritance functions of this module
-- @param obj	An object which will be updated
-- @return object with inheritance functions
function data_obj:new (obj)
	obj = obj or {}
	setmetatable(obj, self)
	self.__index = self
	return obj
end

--- Set default value for empty values
-- @param value  A value which will be used as default
function data_obj:setdef (value)
	self.___ = value
end

--- Clear buffer with data for next using
function data_obj:cleardata ()
	self.data = {}
end

--- Update filed in data table with specified keys
-- @param key    A key which should be updated
-- @param val_a  A value if val_b is empty or otherwise new value for [key]
-- @param val_b  A new value for cell [key][val_a]
-- @return boolean True or False with error message
-- @note if value is empty, will be used default value
function data_obj:upd (key, val_a, val_b)
	if not key then
		return false, "Not found 'key' attribute for data"
	end
	
	-- check exist key
	if "table" == type(self.data[key]) then
		if not val_a then
			-- we need val_a as key for data table 
			return false, string.format("Attribute - [%s] is table", key)	
		end
		self.data[key][val_a] = val_b or self.___
	else
		if val_b then 
			if self.data[key] then -- current key is not a table
				return false, string.format("Attribute - [%s] is not a table", key)
			end
			self.data[key] = {}
			self.data[key][val_a] = val_b or self.___
		else 
			self.data[key] = val_a or self.___
		end
	end
	return true
end

--- Add new attribute to data table
-- @param key    A key which will be added
-- @param val_a  A value which will be added to attribute [key] or if val_b present it is subkey
-- @param val_b  A value which will be added to attribute [key][val_a]
-- @return boolean True or False with error message
function data_obj:add (key, val_a, val_b)
	if not key then
		return false, "Not found 'key' attribute for data"
	else
		if self.data[key] then -- check table with 'key'
			if "table" == type(self.data[key]) then
				if val_a and self.data[key][val_a] then -- check table with 'key.val_a'
        			return false, string.format("Attribute - [%s][%s] exists", key, val_a)
        		end
        	else
				return false, string.format("Attribute - [%s] exists", key)
			end
		end
    end
	return self:upd(key, val_a, val_b)
end

--- Delete attribute from data table
-- @param key       Specified attribute key
-- @param val       Specified subkey attribute
-- @param is_force  Set force removing attribute if it is table
-- @return boolean True or False with error message
function data_obj:del (key, val, is_force)
	if not key then
		return false, "Not found 'key' attribute for data"
	end
	
	if not self.data[key] then
		return false, string.format("Attribute - [%s] is not found", key)
	end
	
	if val then
		if "table" ~= type(self.data[key]) then
			return false, string.format("Attribute - [%s] is not a table", key)
		elseif not self.data[key][val] then
			return false, string.format("Attribute - [%s][%s] is not found", key, val)
		end
		
		self.data[key][val] = nil
		return true
	elseif "table" == type(self.data[key]) and not is_force then
		return false, string.format("Attribute - [%s] is a table", key)
	end
	
	self.data[key] = nil
	return true
end

-- push values from local table to module environment
for key, value in pairs(data_obj) do
	_M[key] = value
end
