--- Interface to connect with sipproxy via HTTP API
-- 
-- @file    sipproxy.lua 
-- @author  Vyacheslav Khudan
-- @created 2017/06/13
-- 
-- @note: Do not use Kamailio API to get all users to reduce loading!!! 

package.path = package.path .. ";/usr/local/share/lua/5.1/core/sipproxy/?.lua"

--- @module sipproxy
local sipproxy = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
function getRegistrationByAOR(data)
    --[[
    local limit   = 16
    local datalen = #data
    local parts   = {}

    for i=1, datalen, limit do
        local last = i + limit - 1
        parts[#parts + 1] = "/users/" .. table.concat(data, ",", i, last > datalen and datalen or last)
    end
    --]]

    local extens = "/users/" .. table.concat(data, ",")
    return {extens}
end

local _requestTypeList = {
    getRegistrationByAOR = getRegistrationByAOR

    ,getOnlineRegistrationByAOR = getRegistrationByAOR
    
    ,setActiveDevice      = function(data)
        local b64     = require"base64"
        local urlData = {
             "aor="..b64.encode(data.extension)
            ,"contact="..b64.encode(data.contact)
            ,"ua="..b64.encode(data.userAgent)
        }
        return "/activedevice/?" .. table.concat(urlData, "&")
    end
    
    ,getRegistrationByUAMAC = function()
        return {"/users/", "/trunks/dynamic/"}
    end

    ,getSipDestinationState = function(data)
        local urlData = {
            "uri="..data.uri
            ,"name="..(data.name or "")
            ,"serial="..(data.serial or "")
        }
        return "sipstate/?" .. table.concat( urlData, "&")
    end
}

--- Parsing methods for required requests
local _responseParserList = {
    getRegistrationByAOR = function(response)
        local rv = {}
        for _, data in pairs(response) do
            if data.extension and "table" == type(data.registrations) then
                for _, reg in ipairs(data.registrations) do
                    if "table" == type(reg) then
                        rv[#rv + 1]    = reg
                        rv[#rv].aor    = data.extension
                        rv[#rv].online = nil
                    end  
                end
            end
        end
        
        return rv
    end
   ,getRegistrationByUAMAC = function(response, request)
        for k,v in ipairs(request) do
            local mac = v and v:upper()
            if mac then
                request[k]   = k
                request[mac] = k
            end
        end
        
        local rv = {}
        for _, data in pairs(response) do
            if data.extension and "table" == type(data.registrations) then
                for _, reg in ipairs(data.registrations) do
                    local regMAC = reg and reg.useragent
                          regMAC = regMAC:match("^Wildix%s.+%s[0-9%.]+%a?%s([0-9A-Fa-f]+)")
                          regMAC = regMAC and regMAC:upper()
                    if request[regMAC] then
                        rv[#rv + 1]    = reg
                        rv[#rv].aor    = data.extension
                        rv[#rv].online = nil
                    end
                end
            end
        end
        return rv
    end
    ,getOnlineRegistrationByAOR = function(response)
        local rv = {}
        for _, data in pairs(response) do
            if data.extension and "table" == type(data.registrations) then
                for _, reg in ipairs(data.registrations) do
                    if "table" == type(reg) then
                        if "1" == reg.online then
                            rv[#rv + 1]    = reg
                            rv[#rv].aor    = data.extension
                            rv[#rv].online = nil
                        end
                    end  
                end
            end
        end
        
        return rv
    end
    ,getSipDestinationState = function(response)
        local rv = response
        return rv
    end
}

local function __makeResult(reqType, response, request)
    local rv = response
    if 0 ~= #response then
        rv = _responseParserList[reqType](response, request)
    end
    return rv
end

local function __collectResponse(rawResponse, collectedData)
    local rv  = false
    local err = "unknown error"
    if not rawResponse then
        return rv, "invalid data in raw response"
    end
    if "table" ~= type(collectedData) then
        return rv, "not found result table"
    end
    
    -- no registration on extension, just skip
    if 404 == rawResponse.code then
        return rv
    end
    
    if 200 ~= rawResponse.code then
        err = rawResponse.scode and rawResponse.scode:match("^HTTP/[0-9%.]+%s+(.+)")
        return rv, err
    end
    
    rv = true
    if rawResponse.body and "" ~= rawResponse.body then
        -- related to WMS-5311
        local bodylen = #rawResponse.body
        local headerlen = tonumber(rawResponse.headers["content-length"])
        if bodylen ~= headerlen then
            return false, string.format("trimmed response, length %d by header, got %d", headerlen, bodylen)
        end

        local body = require"dkjson".decode(rawResponse.body)
        if body and "table" == type(body.result) then
            for _, value in ipairs(body.result) do
                collectedData[#collectedData + 1] = value
            end
            rv = body.result
        end
    end
    return rv
end
   
local function __sendRequest(obj, reqType, ...)
    local reqType = reqType or ""
    if not _requestTypeList[reqType] and _responseParserList[reqType] then
        return false, "invalid/unknown type request: " .. reqType
    end

    local data    = {...}
          data    = "table" == type(data[1]) and data[1] or data
    local reqData = _requestTypeList[reqType](data)
    if not reqData then
        return false, "request data invalid"
    end
    reqData = "table" ~= type(reqData) and {reqData} or reqData

    local backend = nil -- auto
    local options = {}
    -- !HACK! please see WMS-6216 and WMS-5373
    if channel and app then
        if "getRegistrationByAOR" == reqType and string.find(reqData[1], ",") then
            backend = "shell"
            options.cw_hack = true
        end
    end

    local httpClient = require"core.httpClient":new(backend)

    local collectionErr     = nil
    local collectedResponse = {}
    for _, perURLReqData in ipairs(reqData) do
        local url = string.format("%s/%s/%s", obj.APIURL, obj.APIVer, perURLReqData)
        local crv, cerr = __collectResponse(httpClient:get(url, options), collectedResponse)
        collectionErr = cerr or collectionErr
    end

    local rv = __makeResult(reqType, collectedResponse, data)

    if 0 == #rv then
        local msg = "Not found any data in response"
        if collectionErr then
            msg = msg .. ", " .. collectionErr
        end
        return false, msg
    end
    
    return obj.resClass.wrap(rv), collectionErr
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local proxy = {     -- meta class
     APIURL   = "http://127.0.0.1:5060/"
    ,APIVer   = "v1"
    
    ,resClass = require"response"
} 

--- Send request to set up active device
-- @param data  A list of the data about target device, like:
--              extension, contact, userAgent.
-- @return TRUE otherwise FALSE with error message
function proxy:setActiveDevice(data)
    return __sendRequest(self, "setActiveDevice", data)
end

--- Retrieve user(s) registration data by extensions
-- @param ...  A requested extension or multiply extensions separated by comma
-- @return result object with related messages otherwise FALSE with error message
-- @see sipproxy.response
function proxy:getRegistrationByExten(...)
    return __sendRequest(self, "getRegistrationByAOR", ...)
end


--- Retrieve user(s) registration data by extensions, push excluded
-- @param ...  A requested extension or multiply extensions separated by comma
-- @return result object with related messages otherwise FALSE with error message
-- @see sipproxy.response
function proxy:getOnlineRegistrationByExten(...)
    return __sendRequest(self, "getOnlineRegistrationByAOR", ...)
end


--- Retrieve user(s) registration data by MAC address from user agent
-- @param ...  A requested MAC address or addresses separated by comma
-- @return result object with related messages otherwise FALSE with error message
-- @note: create big load on Kamailio process!!! Do not use as periodic or frequent request!
-- @note: responce can be cached on Kamailio side (around 1 minute) 
-- @see sipproxy.response
-- TODO: remove after improvement 'Use extesion in wmsbind instead mac address'
function proxy:getRegistrationByUAMAC(...)
    return __sendRequest(self, "getRegistrationByUAMAC", ...)
end

--- Retrieve sip destination state
-- @param ...  A destination sip uri, like sip:1.2.3.4 or sip:1.2.3.4:5060 or sip:1.2.3.4:443;transport=tls
-- @return result object with related messages otherwise FALSE with error message
function proxy:getSipDestinationState(...)
    return __sendRequest(self, "getSipDestinationState", ...)
end


-- Class constructor
function sipproxy:new()
    return setmetatable({}, {__index = proxy})
end

return sipproxy 