--- Class of the global space for lua state
--  Actual after chahges for callweaver by using one lua state for different channels
--
-- @file    stateGlobalSpace.lua
-- @author  Vyacheslav Khudan
-- @created 2016/11/23
--
-- @note: Object of the global space class will be a table located inside channel table,
--        otherwiser table inside global space. Table name for both cases will be '_stateSpace'.
--        Space table has two sub-tables for global variables and internal modules variables.
--        Global space library must be initialized before using by executing method 'initialize'
--        (we use special callback 'channelinit' to do this in the dialplan).
--        How to use:
--          a) you have simple use default space table for global variables, like 
--                  gSpace['test'] = 10; print(gSpace())
--          b) make module partial space initialization via method 'get4Module'.
--             So dialplan, etc. module with such processing must have table for internal variables,
--             that will be located in the global space (described module). Example:
--                  local module = {
--                          mData = nil     -- memory for internal data
--                  }
--                  function module:new() {
--                      self.mData     = gSpace and gSpace:get4Module() or {}
-- 
--                      .... initialize object ... 
--                      obj.mData.val = 1
--
--                      return obj
--                  }
--                  return module 

--- @module stateGlobalSpace
local stateGlobalSpace = {}

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local stateSpace = {    -- meta class
     mIdent   = true        -- object identificator
    ,mSpace   = true        -- object vairables space
    ,mMembers = nil         -- object private members
}

--- Initialize state space
function stateSpace:initialize(customSpace)
    local suffix = require"socket".gettime()
    if "table" == type(customSpace) then
        rawset(customSpace, "_stateSpace", {})
        self.mSpace         = rawget(customSpace, "_stateSpace")
        self.mIdent         = string.format("C-%s-%s", tostring(customSpace), suffix)
        if channel and channel.CHANNEL then
            self.mIdent = self.mIdent .. " - " .. channel.CHANNEL:get("name") 
        end
    else
        _G["_stateSpace"]   = {}
        self.mSpace         = _G["_stateSpace"]
        self.mIdent         = string.format("G-%s-%s", tostring(_G), suffix)
    end
    self.mSpace["_Global"]  = {}
    self.mSpace["_Modules"] = {}
    
    return self
end

--- Retrieve state space identificator
function stateSpace:getIdent()
    return "string" == type(self.mIdent) and self.mIdent or ""
end

--- Return module space for internal variables
function stateSpace:get4Module(modObj)
    local modObj = "table" == type(modObj) and modObj or {}
    
    if _G["_stateSpace"] ~= self.mSpace and "boolean" ~= type(self.mSpace) then
        if "table" ~= type(self.mSpace["_Modules"][modObj]) then
            local mt = {
                __metatable = false;    -- protected metatable
                __call     = function(t, a)
                    return self.mSpace["_Modules"][t]
                end;
                __index    = function(t, k)
                    if self.mSpace["_Modules"][t] then
                        return rawget(t, "_"..k)
                    end
                end;
                __newindex = function(t, k, v)
                    if self.mSpace["_Modules"][t] then
                        rawset(t, "_"..k, v)
                    end
                end;
            }
            self.mSpace["_Modules"][modObj] = setmetatable(modObj, mt)
        end
        return self.mSpace["_Modules"][modObj]
    end

    return modObj
end

--- Setter for variables to state space
function stateSpace:setVar(name, val)
    if not name or not val then
        return false
    end
    
    self.mSpace["_Global"][name] = val
    return true
end

--- Getter for variables from state space
function stateSpace:getVar(name)
    return self.mSpace["_Global"][name]
end


-- Object constructor
function stateGlobalSpace:new()
    local obj = {}
    
    -- preload initialization
    obj.mMembers = {}
    for k, v in pairs(stateSpace) do
        if "function" ~= type(v) then
            obj.mMembers[k] = true
        end
        obj[k] = nil
    end
    
    -- loading metatable
    setmetatable(obj, { 
        __index = function(tbl, key)
            return stateSpace[key] or stateSpace.getVar(tbl, key)
        end;
        __newindex = function(tbl, key, val)
            local rv = false
            if tbl.mMembers[key] then
                rv = rawset(tbl, key, val)
            elseif not stateSpace[key] then
                rv = stateSpace.setVar(tbl, key, val)
            end
            return rv
        end
    })
    
    return obj
end

return stateGlobalSpace
