--- Module for time date converting
--
-- @file    timedate.lua
-- @author  Vyacheslav Khudan
-- @created 2016/07/15

local tzcache     = require"luatz.tzcache"
local tzparse     = require"luatz.parse"
local tztimetable = require"luatz.timetable"
local tzinfo      = require"luatz.tzinfo"

--- @module timedate
local timedate = {}

--- check and convert time zone to tzdata format
local function __checkTZ(timeZone)
	local sign, hour = string.match(timeZone or "", "^([%+%-])(%d%d)%d%d$")                                 -- get the "+01" part of the "+0100" string
	if "+0100" == timeZone then                                                                             -- convert old +0100 format to zoneinfo timezone
	    timeZone = "Europe/Berlin"
	elseif "+0200" == timeZone then
	    timeZone = "Europe/Kiev"                                                                             --convert old +0200 timezone format to zoneinfo timezone
	elseif sign and hour then
		sign     = "-" == sign and "+" or "+" == sign and "-"				                                -- zone names beginning with "Etc/GMT" have their sign reversed	
		timeZone = "Etc/GMT"..sign..tonumber(hour)                                                          -- convert the time zone to tzdata format "Etc/GMT+01"
	end
	timeZone = string.match(timeZone or "", "[%a%-_]+/[%w%+%-_]+/?[%w%+%-_]+")                              -- matches time zone "Europe/Rome", "US/Central", "Etc/GMT+01" etc.
	timeZone = require"core.fsutils".is_readable("/usr/share/zoneinfo/"..tostring(timeZone)) and timeZone or nil -- check zoneinfo path
    
    return timeZone
end

local function __adaptDateFormat(formatStr, formatType)
    rv = formatStr
    if not formatStr or not formatType then
        return rv
    end

    if "string" == type(formatStr) and "" ~= formatStr and "php" == formatType then
        local t = {     --translation table from php date format to C strftime format
                 l="A"  -- Full weekday name   Sunday
                ,F="B"  -- Full month name March
                ,M="b"  -- Abbreviated month name  Mar
                ,A="p"  -- AM or PM designation    PM
                ,h="I"  -- Hour in 12h format (01-12)  05
                ,i="M"  -- Minute (00-59)  55
                ,s="S"  -- Second (00-61)  02
                ,T="Z"  -- Timezone name or abbreviation   CDT
        }
        rv = string.gsub(formatStr,"(%a)",t)
    end
    
    return rv
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local tbl = {	-- meta Class
	 unix = nil     -- unix time format data
	,tt   = nil     -- time table object
    ,raw  = nil     -- raw data created by object constructor
}

--- Set date-time format
-- @param   A date format string like "d/m/y" 
-- @param   A format type - php, lua ... 
-- @param   A time format - 24h or 12h AM/PM
-- @return true or false 
function tbl:setDateTimeFormat(date, type, is24h)
    if not date or "" == date then
        return false
    end
    
    self.dateFormat = __adaptDateFormat(date, type)
    self.is24h      = is24h
    
    return true
end

--- Print time by date pattern
-- @param ptrn  A date/time formatting string 
-- @return formatted date/time string 
function tbl:format(ptrn)
    if not ptrn or "" == ptrn then
        return ""
    end
    
    return self.tt:strftime(ptrn)
end

--- Print time by template
-- @param tmpl  A date/time formatting template 
-- @return formatted date/time string 
function tbl:formatBySpecOpt(trObj,options)
    if not trObj or not options or "table" ~= type(options) then
        return ""
    end
    
    local function __fixElement(element,delimiter,trObj,options)
        if options.hideYear and (element == "Y" or element == "y") then
            element   = ""
            delimiter = ""
        end
        if element == "b" or element == "B" then
            element = trObj:_t(self:format("month.%b"))
        elseif element and "" ~= element then
            element = self:format("%"..element)
        else
            element   = ""
            delimiter = ""
        end
        
        return element, delimiter
    end
    
    local time = "1" == self.is24h and self:format("%H:%M") or self:format("%I:%M %p")
    local today = self.raw.timestamp > os.time({year=os.date("!%Y"),month=os.date("!%m"),day=os.date("!%d"),hour=0})
    local delim1 = ""
    local elem1, delim2, elem2, delim3, elem3 = string.match(self.dateFormat or "", "^(%a)(.)(%a)(.)(%a)$")
    local timeDelim = " "
    if today then
        return time
    end
    if options.hideTime then
        time = ""
    end
    elem1, delim1 = __fixElement(elem1,delim1,trObj,options)
    elem2, delim2 = __fixElement(elem2,delim2,trObj,options)
    elem3, delim3 = __fixElement(elem3,delim3,trObj,options)
    delim2 = ("" == elem1) and "" or delim2 --fixes -02-06 view issue
    local date = delim1..elem1..delim2..elem2..delim3..elem3
    
    return date..timeDelim..time
end

--- Get time table
-- @return time table 
function tbl:getTimeTable()
    return self.tt
end

--- Convert time table to UTC time zone
function tbl:toUTC()
    local rv = false
    if "table" ~= type(self.raw) then
        return rv
    end
    
    self.unix = self.raw.tzObj:utctime(self.raw.timestamp)
    self.tt   = tztimetable.new_from_timestamp(self.unix)
    return self.unix and self.tt
end

--- Convert time table to specified time zone
function tbl:toTimeZone()
    if "table" ~= type(self.raw) then
        return false
    end
    
    self.unix = self.raw.tzObj:localize(self.raw.timestamp)
    self.tt   = tztimetable.new_from_timestamp(self.unix)
    return self.unix and self.tt
end

--- Find current time
-- @return the relevant tt_info object for the given UTC timestamp in the timezone.
function tbl:__findCurrent(utcTs)
    if not utcTs then
        utcTs = self.raw.timestamp
    end

	return self.raw.tzObj:find_current(utcTs)
end

function tbl:getTzOffset()
	local rv = {}
	local ttInfo = self:__findCurrent()
	local offset = ttInfo and tonumber(ttInfo.gmtoff)
	if offset then
		offset = ttInfo.isdst and (offset - 3600) or offset
        local h,m = math.modf(math.abs(offset/3600))
        rv.sign  = offset < 0 and "-" or ""
        rv.signp = offset < 0 and "-" or "+"
        rv.dec   = h.."."..m --offset in decimal format 6.5
        h        = h < 10 and "0"..h or h
        rv.hour  = h
        m        = m*60
        m        = m < 10 and "0"..m or m
        rv.min   = m
        rv.sec   = offset
	else
		rv = nil
	end
	
	return rv
end

--- Get daylight saving time data
function tbl:getDstData(timeZone)
	local rv = {}
	local dst = {}
	for _, v in pairs(self.raw.tzObj) do
    	-- get data for current year
    	if v.transition_time and self:format("%Y") == os.date("%Y", v.transition_time) then
    		dst[#dst+1] = v
    	end
    end

	if #dst ~= 2  then
		return nil, "No daylight saving time data found"
	end
	
	rv.offsetmin  = (tonumber(dst[1].info.gmtoff) - tonumber(dst[2].info.gmtoff))/60
	rv.offsethour = (tonumber(dst[1].info.gmtoff) - tonumber(dst[2].info.gmtoff))/3600
	rv.offsetsign = rv.offsethour > 0 and "+" or "-"
	
	for k,v in pairs(dst) do
		local dt = require"core.timedate":new(v.transition_time, timeZone)
		if dt then
			dt:toTimeZone(timeZone)
			rv[k] = dt:getTimeTable()
        else
        	return nil, "Timedate error: No timedate object."
        end
	end

	return rv
end

--- Convert daylight saving time data to weeks
function tbl:getDstByWeek(dst)
    local month_days = {[3]=31,[4]=30,[5]=31,[6]=30,[7]=31,[8]=31,[9]=30,[10]=31,[11]=30}
    if not (month_days[dst[1].month] and month_days[dst[2].month]) then
        return nil
    end

    local rv = {}
    for n, t in pairs({dst[1], dst[2]}) do
        local wt = {}
        wt.month = t.month
        wt.hour  = t.hour
        -- NOTE: it's locale-specific in general
        wt.wday  = 0 ~= (t.wday - 1) and t.wday - 1 or 7
        local week = math.ceil(t.day / 7)
        if 4 == week and (month_days[t.month] - t.day) < 7 then
            week = 5 -- last week for device
        end
        wt.week = week
        rv[n]   = wt
    end

    return rv
end

--- Get timezone string
function tbl:getTZString()
    -- NOTE: this is not documented!
    return self.raw.tzObj.future
end

--- Get timestamp
-- @return unix timestamp 
function tbl:getTimestamp()
    return tonumber(self.raw.timestamp)
end

--- Object constructor
-- @param timeStr   An unix or rfc3339 time format only UTC
-- @param timeZone  A string time zone format (ex. Europe/Rome) 
-- @return time format object otherwise FALSE value
function timedate:new(timeStr, timeZone, dtSettings)
    local rv = false
    if not timeStr or "" == timeStr then
        return rv
    end
    timeZone = __checkTZ(timeZone)
    local msg = not timeZone and "Time zone not valid, will be used pbx local time zone."

	local obj = {}
	setmetatable(obj, {__index = tbl})
    
	-- initialization
    local tzObj   = tzcache.get_tz(timeZone)
    local timeObj = tzparse.rfc_3339(tostring(timeStr))
    local timestamp = timeObj and timeObj:timestamp() or timeStr
    if tzObj and timestamp then
        obj.raw = {
             tzObj     = tzObj
            ,timestamp = timestamp -- UTC time stamp
            ,format    = formatType
        }
    end
    if "table" == type(dtSettings) then
        obj:setDateTimeFormat(dtSettings.format, dtSettings.type, dtSettings.is24h)
    end

	return obj, msg
end

return timedate

