--- W-AIR firmware utils
--
-- @file    wairutil.lua
-- @author  Yuriy Zveryanskyy
-- @created 2021/06/23

local fsutils  = require"core.fsutils"
local json = require"dkjson"

local lock_dir = "/rw2/var/spool/wairdevlock"

local wairutils = {
	fw_data = nil
}

local function __getLockName(net_id)
	return lock_dir .. "/" .. net_id .. ".lock"
end

local function __versionToNumber(version)
	if not version or "string" ~= type(version) then
		return 0
	end
	local rv = 0
	rv = string.gsub(version, "%.", "")
	rv = tonumber(rv or 0)
	return rv
end

local function __loadFirmwares(path)
	if not wairutils.fw_data then
		local firmwares = nil
		if require"common.helper".is_exists (path .. "firmwares.json") then
			local fp = io.open(path .. "firmwares.json")
			firmwares = json.decode(fp:read ("*a"))
			io.close (fp)
		end
		wairutils.fw_data = firmwares
	end
	return wairutils.fw_data
end

local function __getBaseDestVersion(model)
	local ver = nil
	local firmwares = __loadFirmwares("/var/www/firmwares/wildixair/")
	if firmwares then
		local firmware = firmwares[model:lower()]
		local app = firmware["application"]
		if firmware and app then
			ver = tonumber(tostring(app[1]) .. tostring(app[2]))
		end
	end
	return ver
end

local function __getNetworkInfo(net_id)
	local netinfo = {}
	local connection = require"common.core".connection
	local sql = "SELECT value FROM autoprov WHERE param='device' AND mac IN (SELECT mac FROM wair_gateways WHERE network_id='%s')"
	for devdata in connection:rows(sql, net_id) do
		local devtbl = json.decode(devdata)
		if devtbl and devtbl["fw:version"] then
			netinfo[#netinfo + 1] = {
				["fw"]      = __versionToNumber(devtbl["fw:version"]),
				["mac"]     = devtbl["mac"],
				["address"] = devtbl["address"]
			}
		end
	end
	return netinfo
end

--- Retreive available firmware info for upgrade from current version.
-- Can be used transitional firmware if needed.
-- @param agent A table with general device informaion
-- @return table with upgrade firmware info
function wairutils.getUpgradeFirmwareInfo(agent)
	if "table" ~= type(agent) or not agent["fw:version"] or "" == agent["fw:version"] then
		return false
	end

	local dir = string.format("/var/www/firmwares/%s/", agent["vendor"]:lower())
	local firmwares = __loadFirmwares(dir)

	local upgradeFirmwareInfo = nil
	if firmwares and "table" == type(firmwares) then
		upgradeFirmwareInfo = firmwares[agent["vendor:additional"]:lower()]
	end
	if upgradeFirmwareInfo and "table" == type(upgradeFirmwareInfo) then
		if upgradeFirmwareInfo["transitional"] and upgradeFirmwareInfo["transitional"][1] then
			table.sort(upgradeFirmwareInfo["transitional"], function(a, b)
				return __versionToNumber(a["version"]) < __versionToNumber(b["version"])
			end)
			for k, v in pairs(upgradeFirmwareInfo["transitional"]) do
				if __versionToNumber(agent["fw:version"]) < __versionToNumber(upgradeFirmwareInfo["transitional"][k]["version"]) then
					upgradeFirmwareInfo = upgradeFirmwareInfo["transitional"][k]
					break
				end
			end
		if upgradeFirmwareInfo["transitional"] then
			upgradeFirmwareInfo["transitional"] = nil -- remove transitional section in case no need transitional upgrade
		end
		end
		if __versionToNumber(agent["fw:version"]) ~= __versionToNumber(upgradeFirmwareInfo["version"]) then
			upgradeFirmwareInfo["needUpgrade"] = true
		end
	end

	return upgradeFirmwareInfo
end

function wairutils.lockNeededVer(net_id, model)
	local destver = __getBaseDestVersion(model)
	if not destver then
		return nil, "no destination firmware version"
	end

	local netinfo = __getNetworkInfo(net_id)
	if #netinfo < 2 then
		return nil, string.format("%d base(s) found in the database", #netinfo)
	end

	for _, device in pairs(netinfo) do
		if destver == device["fw"] then
			return nil, "version from the pbx found on wair network base(s)"
		end
	end

	return destver, json.encode(netinfo)
end

function wairutils.checkUpgraded(net_id, destver)
	local netinfo = __getNetworkInfo(net_id)

	for _, device in pairs(netinfo) do
		if destver ~= device["fw"] then
			return false, netinfo
		end
	end

	return true, netinfo
end

function wairutils.devicesLock(net_id, mac, destver)
	if not fsutils.is_dir(lock_dir) then
		fsutils.mkdir(lock_dir)
	end

	local lockname = __getLockName(net_id)
	local tmpname = lockname .. "." .. mac
	local lockdata = {
		["initiator"]   = mac,
		["target"] = destver,
		["timestamp"]   = os.time()
	}
	local fp = io.open(tmpname, "w")
	fp:write(json.encode(lockdata))
	fp:close()
	os.rename(tmpname, lockname)
end

function wairutils.devicesUnlock(net_id)
	os.remove(__getLockName(net_id))
end

function wairutils.getLocksList()
	local rv = {}
	local nets = fsutils.readdir(lock_dir, "^(.*)%.lock") or {}
	for _, net in pairs(nets) do
		local lockname = __getLockName(net)
		local fp = io.open(lockname)
		if fp then
			local content = fp:read("*a")
			fp:close()
			local lockdata = json.decode(content)
			lockdata["net_id"] = net
			rv[#rv + 1] =  lockdata
		end
	end
	return rv
end

function wairutils.isLocked(net_id)
	local lockname = __getLockName(net_id)
	local fp = io.open(lockname)
	if fp then
		local rv = fp:read("*a")
		fp:close()
		return rv
	end
end

function wairutils.proceedUnlock(net, netinfo)
    wairutils.devicesUnlock(net)

    local addrlist = {}
    local maclist = {}
    for _, base in pairs(netinfo) do
        addrlist[#addrlist + 1] = base["address"]
        maclist[#maclist + 1] = base["mac"]
    end
    local provtmp = "nohup wmsbind -o 'action=provision' %s %s >/dev/null 2>&1 &"
    local provcmd = string.format(provtmp, table.concat(addrlist, ","), table.concat(maclist, ","))
    os.execute(provcmd)
end

return wairutils
