--- Wrapper for LDAP connector
--
-- @file    wmsldap.lua
-- @author  Vyacheslav Khudan
-- @created 2016/12/06

--- @module wmsldap
wmsldap          = { -- huck for wildix using as global variable
    data = nil      -- module space for internal data
}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
--- Escape the LDAP search filter to prevent LDAP injection attacks
local function __escapeString(value)
   value = string.gsub(value, "%\\", "\\5c")
   value = string.gsub(value, "%*", "\\2a")
   value = string.gsub(value, "%(", "\\28")
   value = string.gsub(value, "%)", "\\29")
   value = string.gsub(value, "NUL", "\\00")
   value = string.gsub(value, "%/", "\\2F")
   return value
end

--- Retrieve configuration for connector
local function __getConfig()
    local cfg     = {}
    local cfgFile = "/rw2/etc/pbx/ldap_connect.conf"

    local file = io.open(cfgFile)
    if file then
        for line in file:lines() do
            local k, v = line:match("(.*)=\"(.*)\"")
            if k and v ~= "" then
                cfg[k] = tostring(v)
            end
        end
        file:close()
    end

    local wms_net_conf = require"common.helper".parse_simple_ini("/rw2/etc/pbx/net.conf")
    if wms_net_conf and wms_net_conf["STARTUP"] == "1" and tonumber(wms_net_conf["SRV_HTTPS_PORT"]) then
        if cfg.host_rw then
            cfg.host_rw = string.format("%s:%s", cfg.host_rw, wms_net_conf["SRV_HTTPS_PORT"])
        end
    end

    return {
         host_rw = cfg.host_rw or "127.0.0.1"
        ,host   = cfg.host     or "127.0.0.1"
        ,userdn = cfg.binddn   or "cn=admin,dc=wildix"
        ,passw  = cfg.bindpw   or "wildix"
        ,basedn = cfg.basedn   or "dc=wildix"
    }
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local wldap = { -- meta class
    cfg = nil    -- config for connector
}

function wldap:getConection ()
    -- lazy ldap connection
    local ldap_conn, ldap_err = require"lualdap".open_simple(self.cfg.host, self.cfg.userdn, self.cfg.passw)
    if not ldap_conn then
        return nil, ldap_err and "LDAP error: " .. ldap_err
    end

    return ldap_conn, self.cfg.basedn
end

function wldap:__get(connection, variable, value, object_class, ends_with)
    local object_class = object_class or "wPerson"
    local ends_with    = ends_with or false
    local rv           = {}

    if value and variable then
        if value ~= "*" then
            value = __escapeString(value)
        end
        if ends_with then
            value = "*" .. value
        end

        variable = __escapeString(variable)
        local filter = "(&(objectClass=" .. object_class .. ")(".. variable .. "=" .. value .. "))"
        local search = nil
        if variable ~= "dn" then
            search = { base = self.cfg.basedn, scope = "subtree", filter = filter }
        else
            search = { base = value, scope = "base" }
        end

        for dn, attrs in connection:search(search) do
            rv[#rv+1] = {dn, attrs}
        end
    end

    return rv
end

function wldap:get(variable, value, object_class, ends_with)
    local ldap_conn, ldap_err = require"lualdap".open_simple(self.cfg.host, self.cfg.userdn, self.cfg.passw)
    if not ldap_conn then
        return {}, ldap_err and "LDAP error: " .. ldap_err
    end

    local rv = self:__get(ldap_conn, variable, value, object_class, ends_with)

    ldap_conn:close()

    return rv
end

function wldap:get_lazy(variable, value, object_class, ends_with)
    if not self.connection then
        local ldap_conn, ldap_err = require"lualdap".open_simple(self.cfg.host, self.cfg.userdn, self.cfg.passw)
        if not ldap_conn then
            return rv, ldap_err and "LDAP error: " .. ldap_err
        end
        self.connection = ldap_conn
    end

    return self:__get(self.connection, variable, value, object_class, ends_with)
end

function wldap:modify(dn, variable, value, act)
    local rv  = false
    local err = nil

    repeat
        -- lazy ldap connection
        local lua_conn, ldap_err = require"lualdap".open_simple(self.cfg.host_rw, self.cfg.userdn, self.cfg.passw)
        if not lua_conn then
            err = ldap_err and "LDAP error: " .. ldap_err
            break
        end

        local action = (act == "set" and "=") or (act == "add" and "+") or (act == "del" and "-")
        if not action then
            err = "LDAP error: Invalid action for updating LDAP entry"
            break
        end

        if not dn or not variable or not value then
            err = "LDAP error: Wrong arguments for updating LDAP entry"
            break
        end

        local result, ldap_err = lua_conn:modify(dn, { action , [variable] = value })()
        lua_conn:close()
        if not result then
            err = ldap_err
            break
        end

        rv = true
    until true

    return rv, err
end

function wldap:close()
    if self.connection then
        self.connection:close()
        self.connection = nil
    end
end

--- Object constructor
function wmsldap:new()
    local obj = self.data and self.data["obj"]
    if not obj then
        -- create data space
        self.data = gSpace and gSpace:get4Module() or {}

        -- initialize object
        obj           = setmetatable({}, {__index = wldap})
        self.data.obj = obj

        -- setup object
        obj.cfg = __getConfig()
    end

    return obj
end

--- Proxy method. Get connection to ldap database
-- @note Required to close connection manually
function wmsldap:getConection()
    return self:new():getConection()
end

--- Proxy method. Get entries from ldap database by filter
-- @param variable      A string with a search field
-- @param value         A string with value of search field, to get multiply results use "*"
-- @param object_class  A string with a classObject to be look up
-- @returnA table with entries or empty #table followed by an error string.
function wmsldap:get(variable, value, object_class, ends_with)
    return self:new():get(variable, value, object_class, ends_with)
end

--- Proxy method. Modify value of a field in LDAP databse
-- @param dn        A string with dn value
-- @param variable  A string with a field to be updated
-- @param value     A string with value
-- @param act       A indicate how modify (set, add, remove)
function wmsldap:modify(dn, variable, value, act)
    return self:new():modify(dn, variable, value, act)
end

--- Proxy method. Move user in ldap database from one pbx to another
-- @param userDn    A domain of specify user
-- @param oldPbxDn  A domain of old pbx
-- @param newPbxDn  A domain of new pbx
function wmsldap:moveuser(userDn, oldPbxDn, newPbxDn)
    return self:new():modify(userDn, "pbxowner", newPbxDn, "set")
end

function wmsldap:get_lazy(variable, value, object_class, ends_with)
    return self:get_lazy(variable, value, object_class, ends_with)
end

function wmsldap:close()
    self:close()
end

return wmsldap
