--- Connector to sqlite databases with static database configuration
-- 
-- @file	wsqlite.lua 
-- @author	Vyacheslav Khudan
-- @created 2016/12/09

local wsqlite = {
    dbList = {
         ["wmsdb"]      = "/rw2/var/lib/wmsdb.d/wmsdb"
		,["devdb"]		= "/rw2/var/lib/wmsdb.d/devdb"
		,["wnpdb"]      = "/rw2/var/lib/wnp/wnpdb"
        ,["phonebook"]  = "/mnt/cticonnect/cticonnectdb"
        ,["callback"]   = "/dev/shm/wmsdb.d/callbackdb"
    }
}

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local sqlite = {   -- meta-class
     env      = nil
    ,conn     = nil
    ,database = nil
    ,err      = nil
}

function sqlite:__setUpError(msg, ...)
    if not self.err and msg then
        if "table" == type(msg) then
            self.err = msg
        else
            self.err = string.format(msg, ...)
        end
    end
end

--- Retrieve internal error
function sqlite:getErrorMsg()
    if "table" == type(self.err) then
        return self.err, "sql-error"
    else
        return self.err or ""
    end
end

--- Execute database connection by creating database file if empty
function sqlite:connectForce()
    local err = nil
    if not self.conn then -- lazy database connection
        self.err      = nil
        local luasql  = require("luasql.sqlite3")
        self.env, err = luasql.sqlite3()
        if self.env then
            self.conn, err = self.env:connect(self.database)
        end
    end
    
    if err then
        self:__setUpError("%s (%s)", err, self.database)
    end
    
    return self.env and self.conn
end

--- Execute database connection with db file checking
-- @return sqlite connection object
function sqlite:connect()
    if not self.conn then
        local dbFile = self:getDBPath()
        if 0 == require"core.fsutils".filesize(dbFile) then
            self:__setUpError("Database %q is empty/not found", dbFile)
            return false
        end
        return self:connectForce()
    end
    
    return self.conn
end

--- Execute an SQL request
-- @param sqlStat    A string with sql statment
-- @param mixed      Values to be passed to string.format funciton as query pars
-- @return execution boolean state with error message
function sqlite:execute(sqlStat, ...)
    local rv  = false
    if not sqlStat then
        self:__setUpError("Invalid sqlite statement for execute command")
        return rv
    end
    
    local params = {...}
    local ok, query = pcall(function() return string.format(sqlStat, unpack(params)) end)
    if not ok then
        error(query .. ", " .. sqlStat)
    end
    
    local connState = self:connect()
    if not connState then
        return rv
    end
    
    local maxAttempts = 30   -- max attempts
    local counter     = 0    -- attempts counter
    local divider     = channel and 1000 or 100 -- request from dialplan or not
    local err         = nil
    repeat
        counter              = counter + 1
        local sqlRes, sqlErr = self.conn:execute(query)
        
        -- processing result data
        local resType = type(sqlRes)
        if "userdata" == resType or "number" == resType then
            rv = sqlRes
            break
        end
        
        -- error processing
        err = sqlErr
        if sqlErr and   -- retry exceptions
           (-- no such table
           sqlErr:find("%:%s*no such table%s*%:")
           or -- no such column
           sqlErr:find("%:%s*no such column%s*%:")
           ) then
            err = err .. " (break retries)"
            break
        end
        
        -- make sleep some random time to retry
        local sleepTime = math.random(1,5)/divider
        require"socket".sleep(sleepTime)
        
    until counter >= maxAttempts
    
    if not rv then
        rv = self.conn:close()
        
        -- building error message
        local errMsg = {}
        errMsg[#errMsg+1] = "database: " .. self.database
		errMsg[#errMsg+1] =  "error: " .. err
        errMsg[#errMsg+1] =  "query: " .. query
        
        self:__setUpError(errMsg)
        local logger = require"core.logger":init("luasqlite", 3)
        logger:error(errMsg)
    end
    
    return rv, err
end

--- Execute an SQL query by processing multi responce
-- @return callback for result processing otherwise error message
function sqlite:rows(sqlStat, ...)
    local cursor = self:execute(sqlStat, ...)
    if "userdata" == type(cursor) then
        return function()
            local rv = {cursor:fetch()}
            if not rv[1] then
                cursor:close()
            end
            
            return unpack(rv)
        end
    else
        self:__setUpError("Invalid cursor object")
        return function()
            return nil
        end
    end
end
   
--- Close connector
function sqlite:close()
    if self.conn then
        self.conn:close()
        self.conn = nil
    end
    if self.env then
        self.env:close()
        self.env = nil
    end
end

--- Retrieve full path to database file
-- @return database path 
function sqlite:getDBPath()
    local rv = nil
    if self.database then
        rv = wsqlite.dbList[self.database] or self.database
    end
    
    return rv
end


--- Object constructor
function wsqlite:new(db)
    db = self.dbList[db] or db
    if not db then
        return nil, "Invalid database"
    end
    
    -- object initialization
    local obj = {}
    obj = setmetatable(obj, {__index = sqlite})
    obj.database = db
    
    return obj
end

return wsqlite
