--- Module for XML content serialization from table
-- 
-- @file	wxml.lua 
-- @author	Vyacheslav Khudan
-- @created 2015/04/03

local type 		   = type
local string 	   = string
local table        = table
local pairs 	   = pairs
local setmetatable = setmetatable
local genx 		   = require"genx"

module(... or "wxml")

----------------------------------------------------------------
-- XML parser implementation details
----------------------------------------------------------------
local xml_tab  = "    "		-- XML tabulation character
local xml_buf  = "" 		-- XML result buffer
local xml_lact = nil
local xml_ntab = -1

--- XML content builder
local function __xml_cont_build(str)
	if not str or "" == str then
		return false
	end
	
	if str:find(".*_%d+$") then -- normalize multi element names
		if not str:find("fxo%d+_%d+$") then
			str = str:match("(.*)_%d")
		end
	end

	local delim  = ""
	if "<" == str then
		if "eElm" ~= xml_lact then
			xml_ntab = xml_ntab + 1
		end
		xml_lact = "sElm"
		delim 	 = "\n" .. string.rep(xml_tab, xml_ntab)
	elseif "</" == str then
		if "sElm" ~= xml_lact then
			xml_ntab = 0 < xml_ntab and xml_ntab - 1
			delim    = "\n" .. string.rep(xml_tab, xml_ntab)
		end
		xml_lact = "eElm"
	end
	
	xml_buf = xml_buf .. delim .. str
end

--- Push table values to XML document
local function __xml_table_parse(d, t)
	if not d or not t then
		return false
	end
	if "table" ~= type(t) then
		return false
	end
	
	-- make sorting keys table, elements with ! tag need to be before other elements
	local keys = {}
	for key in pairs(t) do
		table.insert(keys, key)
	end
	
	table.sort(keys, function(a, b)
		local rv = false
		if a and b then
			local anum = a:match("%W(.*)")
			local bnum = b:match("%W(.*)")
			if a:find("^!") and not b:find("^!") then
				rv = true
			elseif not a:find("^!") and b:find("^!") then
				rv = false			
			elseif anum and bnum then -- string "DirectoryEntry_2" < "DirectoryEntry_10"
				if anum:len() < bnum:len() then
					rv = true
				elseif anum:len() > bnum:len() then 
					rv = false
				else
					rv = anum < bnum
				end
			else
				rv = a < b
			end
		end
		return rv --if return true "a" will be the first element of the array else "b" will be the first element of the array
	end)

	-- parsing XML table	
	for _, key in pairs(keys) do
		local value = t[key]
		if value then
    		if "table" == type(value) then	-- element block
    			d:startElement(key)
    			__xml_table_parse(d, value) -- parsing element block
    			d:endElement()
    		elseif key and "" ~= key then 
    			local tag, text = key:match("(%W)(.*)")
    			if "!" == tag and text then		-- attribute
    				d:attribute(text, value)
    			elseif "#" == tag and text then 			-- commentaries
    				--TODO: procedure for comments processing
    				-- d:comment(value)
    			else							-- text value
    				d:text(value)
    			end
    		end
		end
	end
	return true
end

--- Generate XML string from table
local function __xml_generate(xml_t)
	if not xml_t or "table" ~= type(xml_t) then
		return false
	end
	
	local d = genx.new(__xml_cont_build)
	__xml_table_parse(d, xml_t)
	d:close()
	
	return xml_buf
end

----------------------------------------------------------------
-- XML table element object implementation details
----------------------------------------------------------------

--- Search available name in keys list
local function __next_key(keys, key_name, key_prefix)
	if "table" ~= type(keys) or not key_name or "" == key_name then
		return false
	end
	
	key_name = key_prefix and key_prefix .. key_name or key_name
	if not keys[key_name] then
		return key_name -- first key name, like conf["provision"]
	end
	
	-- try find free name from list of the key names
	for i = 1,1000 do
		local key_name = key_name .. "_" .. i
		if not keys[key_name] then
			return key_name
		end
	end
end

local __elm = {} 		-- element namespace

-- Creating element object
function __elm:new()
	local e_obj = {}
	setmetatable(e_obj, {__index = self})
	
	return e_obj
end

-- Element method for adding internal element
function __elm:add_elm(name)
	if not name or "" == name then
		return false
	end
	
	local elm_name = __next_key(self, name)
	if not elm_name then
		return false
	end
	
	self[elm_name] = self:new() 
	return self[elm_name]
end

-- Adding element attributes
function __elm:add_attr(attrs)
	if "table" ~= type(attrs) then
		return false
	end
	
	local elm_prefix = "!"
	for name, value in pairs(attrs) do
		if name and "" ~= name and value and "table" ~= type(value) then
			name = __next_key(self, name, elm_prefix)
			self[name] = value
		end
	end
	return true
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local wxml = {}	-- Meta class

--- Give XML table element object
-- @return element object for XML table 
function wxml:elm_tbl()
	return __elm:new()
end

--- Serealize XML data from XML table
-- @param xml_t		A table with some formatting structure
-- @param base_str	A base string for XML document, like XML header
-- @return string with content in XML format ortherwiser FALSE value
-- @note xml_t must be:
-- 1. !<text>  - attribute
-- 2. #<text>  - commentaries
-- 3. <text>   - simple text
-- 4. <elem>_X - for elements with the same names
function wxml:generate(xml_t, base_str)
	if not xml_t then
		return false
	end
	
	xml_buf = base_str or ""
	
	return __xml_generate(xml_t)
end

--- Create object with inheritance functions of meta class
-- @return object with related methods
function new(_M)
	local obj = {}
	setmetatable(obj, {__index = wxml})
	
	return obj
end

